#include "arm.h"
#include "Module.h"
#include <cmath>
#include <algorithm>
using namespace std;

int vreg_count;
int vsreg_count;

float2int f2i;
bool opt=false;

int ori_func_stack=0;

// int offset=0;

static Map<ValuePtr, MOperand> value_map; 
// store the map if array pointer address is calculated actually
// address is calculated when passing an array argument in call instruction
static Map<ValuePtr, MOperand> array_ptr_map;
static Map<MOperand, MOperand> imm_map; 
static Map<ValuePtr, int> ptr_val_map; 
static map<MOperand, int> int_val_map;
static map<string,Binary_Op_Type> Binary_ir2asm;
static map<ValuePtr,ValuePtr> my_arr_base;
static map<ValuePtr,int> stack_val_map;
ValuePtr cur_arr_base;
MOperand make_imm(int32 constant)       { auto imm = MOperand(IMM, constant); int_val_map[imm] = constant; return imm; }
MOperand make_reg(uint8 reg)            { return MOperand(REG, reg); }
MOperand make_simm(float constant)      { return MOperand(SIMM, constant); }
MOperand make_vreg(int32 vreg_index)    { return MOperand(VREG, vreg_index); }
MOperand make_sreg(uint8 sreg)          { return MOperand(SREG, sreg); }
MOperand make_vsreg(int32 vsreg_index)  { return MOperand(VSREG, vsreg_index); }
MOperand make_operand(ValuePtr v, Machine_Block *mb, bool no_imm);

MOperand make_legal_imm(int32 constant, Machine_Block *mb) {
    if(!can_be_imm12(constant)) {
        auto vreg = make_vreg(vreg_count++);
        int_val_map[vreg] = constant;
        auto ldr = emit_load_of_constant(vreg, constant, mb);
        return vreg;
    }
    else {
        return make_imm(constant);
    }
}

//this is used to put the later args(e.g. last four) on the stack into the func
// @TODO: load of float
void emit_load_of_later_arg(ValuePtr a, MOperand vreg, Machine_Block *mb, int cum_offset, int arg_size, bool is_float) {
    // assert(idx >= 4);
    auto offset_relative_to_fp = cum_offset;
    if (is_float) {
        auto ldr = new MI_VLoad;
        ldr->mem_tag = MEM_LOAD_ARG;
        ldr->reg = vreg;
        ldr->base = make_reg(sp);
        ldr->offset = make_imm(offset_relative_to_fp);
        ldr->base.size = 8;
        if (arg_size == 8) {
            ldr->is_rv64 = true;
            ldr->reg.size = 8;
        }

        if (mb->control_transfer_inst) {
            insert(ldr, mb->control_transfer_inst);
        } else {
            push(ldr, mb);
        }
    }
    else {
        auto ldr = new MI_Load;
        ldr->mem_tag = MEM_LOAD_ARG;
        ldr->reg = vreg;
        ldr->base = make_reg(sp);
        ldr->offset = make_imm(offset_relative_to_fp);
        ldr->base.size = 8;
        if (arg_size == 8) {
            ldr->is_rv64 = true;
            ldr->reg.size = 8;
        }

        if (mb->control_transfer_inst) {
            insert(ldr, mb->control_transfer_inst);
        } else {
            push(ldr, mb);
        }
    }
}

// this is the function to get the stack value
// because we put all the int into the stack
MOperand get_stack_val(ValuePtr v, Machine_Block* mb,Func_Asm *func_asm)
{
    int val=0;
    if(ori_func_stack!=0)
        val=ori_func_stack-stack_val_map[v];
    else 
        val=func_asm->stack_size-stack_val_map[v];
    auto bi_I = new MI_Binary;
    auto nreg=make_vreg(vreg_count++);
	bi_I->op = BINARY_ADD;
	bi_I->lhs = make_reg(sp);
	bi_I->rhs = make_legal_imm(val, mb);
	bi_I->dst = nreg;
    // operates 64-bit reg
    bi_I->lhs.size = 8;
    bi_I->is_rv64 = true;
	push((MI *)bi_I, mb);
    return bi_I->dst;

}

// insert the first MI before the second
void insert(MI *mi, MI *before) {
    if (before->prev == NULL) {
        before->mb->inst = mi;
    } else {
        before->prev->next = mi;
    }
    mi->mb = before->mb;
    mi->prev = before->prev;
    mi->next = before;
    before->prev = mi;
}

// this is the function to load a global variable
void emit_load_of_global_ref(Func_Asm *func_asm, ValuePtr v, MOperand vreg,  Machine_Block *mb) {
    const char *addr = new char[1024];
	addr = v->name.c_str();
    MOperand adr_global(ADR_GLOBAL, addr);
    auto ldr = new MI_Load;
	ldr->reg = vreg;
	ldr->base = adr_global;
	ldr->mem_tag = MEM_LOAD_GLOBAL_REF;
	push((MI *)ldr, mb);
}

void push(MI *mi, Machine_Block *mb) {
    // assert(mb->last_inst == NULL || (mb->last_inst->tag != MI_BRANCH && mb->last_inst->tag != MI_RETURN));
    // assert(mi->next == NULL);
    mi->mb = mb;
    if (mb->last_inst) {
        mb->last_inst->next = mi;
        mi->prev = mb->last_inst;
    } else {
        mb->inst = mi;
    }
    mb->last_inst = mi;
}


Branch_Condition binary_op_to_branch_cond(Binary_Op_Type op_type) {
    switch(op_type) {
        case BINARY_LESS_THAN:                { return LESS_THAN; } break;
        case BINARY_GREATER_THAN:             { return GREATER_THAN; } break;
        case BINARY_LESS_THAN_OR_EQUAL_TO:    { return LESS_THAN_OR_EQUAL; } break;
        case BINARY_GREATER_THAN_OR_EQUAL_TO: { return GREATER_THAN_OR_EQUAL; } break;
        case BINARY_NOT_EQUAL_TO:             { return NOT_EQUAL; } break;
        case BINARY_EQUAL_TO:                 { return EQUAL; } break;
        case FLOAT_GE:                        { return FLOAT; } break;
        default: { exit(31); report_error("Internal Compiler Error: converting unknown binary op type to branch type."); }
    }
    return NO_CONDITION;
}

// handle relation expression
void emit_binary_icmp(IcmpInstruction *icmp, Machine_Block *mb) {
    Binary_Op_Type cmp_type = Binary_ir2asm[icmp->op];
    auto lhs = make_operand(icmp->a, mb, true);
    auto rhs = make_operand(icmp->b, mb, false);
    auto dst = make_operand(icmp->reg, mb, true);
    MOperand tmp = make_vreg(vreg_count++);
    // cout << "icmp type: " << icmp->op << " " << cmp_type << endl;
    assert(cmp_type >= BINARY_NOT_EQUAL_TO && cmp_type <= BINARY_GREATER_THAN);
    if(cmp_type == BINARY_GREATER_THAN_OR_EQUAL_TO || cmp_type == BINARY_LESS_THAN_OR_EQUAL_TO) {
        assert(false && "do not support single '>=' or '<=' Icmp");
    }

    MI_Binary *bi_sub = new MI_Binary(BINARY_SUBTRACT, tmp, lhs, rhs);
    push(bi_sub, mb);
    // mb->control_transfer_inst = bi_sub;
    MI_Move *mv_zero = new MI_Move(dst, make_reg(zero));
    push(mv_zero, mb);

    // handle equal/not equal
    if(cmp_type == BINARY_EQUAL_TO || cmp_type == BINARY_LESS_THAN_OR_EQUAL_TO || cmp_type == BINARY_GREATER_THAN_OR_EQUAL_TO) {
        MI_Slt *set_eq = new MI_Slt(dst, tmp, make_imm(1), SLT_IU);
        push(set_eq, mb);
    }
    else if(cmp_type == BINARY_NOT_EQUAL_TO) {
        MI_Slt *set_ne = new MI_Slt(dst, make_reg(zero), tmp, SLT_U);
        push(set_ne, mb);
    }

    // handle greater/less
    if(cmp_type == BINARY_GREATER_THAN || cmp_type == BINARY_GREATER_THAN_OR_EQUAL_TO) {
        MI_Slt *set_greater = new MI_Slt(dst, make_reg(zero), tmp, SLT);
        push(set_greater, mb);
    }
    else if(cmp_type == BINARY_LESS_THAN || cmp_type == BINARY_LESS_THAN_OR_EQUAL_TO) {
        MI_Slt *set_less = new MI_Slt(dst, tmp, make_reg(zero), SLT);
        push(set_less, mb);
    }
}

// this is the function to handle a branch 
void emit_branch_asm(Func_Asm *func_asm, Branch_Condition cond, BrInstruction* br_I, 
    Machine_Block *mb, MOperand lhs=make_imm(0), MOperand rhs=make_imm(0), bool is_float=false) {
    if(is_float) {
        auto vcmp_dst = make_vreg(vreg_count++);
        auto fcmp = new MI_VCmp_Set(vcmp_dst, lhs, rhs, cond);
        push(fcmp, mb);

        // if ((lhs cond rhs) != 0), jump to true
        lhs = vcmp_dst;
        rhs = make_reg(zero);
        cond = NOT_EQUAL;
    }
    
    auto br = new MI_Branch();

    br->cond = cond;
    if (cond != NO_CONDITION) {
        // br->cond_reg = make_operand(br_I->exp, mb, true);
        br->lhs = lhs;
        br->rhs = rhs;
    }
        
    auto l1=br_I->label1;
    auto l2=br_I->label2;

    int idx1=0;
    int idx2=0;
    for(int i=0;i<func_asm->mbs.size();i++)
    {
        if(func_asm->mbs[i]->label->name==l1->name)
        {
            idx1=i;
        }
        if(l2!=NULL&&func_asm->mbs[i]->label->name==l2->name)
        {
            idx2=i;
        }
    }
    br->true_target = func_asm->mbs[idx1];
    br->false_target = func_asm->mbs[idx2];

    push((MI *)br, mb);
    mb->control_transfer_inst = (MI *)br;
}

// this is a copy of branch, use to hack the false case
void emit_branch_asm_2(Func_Asm *func_asm, Branch_Condition cond, BrInstruction* br_I, 
    Machine_Block *mb, MOperand lhs=make_imm(0), MOperand rhs=make_imm(0), bool is_float=false) {
    if(is_float) {
        auto vcmp_dst = make_vreg(vreg_count++);
        auto fcmp = new MI_VCmp_Set(vcmp_dst, lhs, rhs, cond);
        push(fcmp, mb);

        // if ((lhs cond rhs) == 0), jump to false
        lhs = vcmp_dst;
        rhs = make_reg(zero);
        cond = EQUAL;
    }
    
    auto br = new MI_Branch();
    br->cond = cond;
    if (cond != NO_CONDITION) {
        // br->cond_reg = make_operand(br_I->exp, mb, true);
        br->lhs = lhs;
        br->rhs = rhs;
    }
    auto l1=br_I->label1;
    auto l2=br_I->label2;

    int idx1=0;
    int idx2=0;
    for(int i=0;i<func_asm->mbs.size();i++)
    {
        if(func_asm->mbs[i]->label->name==l1->name)
        {
            idx1=i;
        }
        if(l2!=NULL&&func_asm->mbs[i]->label->name==l2->name)
        {
            idx2=i;
        }
    }
    br->true_target = func_asm->mbs[idx2];
    br->false_target = func_asm->mbs[idx1];

    push((MI *)br, mb);
    mb->control_transfer_inst = (MI *)br;
}

// this is a function to move the value from one reg to another
MI_Move *emit_move(MOperand dst, MOperand src, Machine_Block *mb = NULL) {
    auto mv = new MI_Move(dst, src);
    if(int_val_map.count(src) && src.tag != REG)
        int_val_map[dst]=int_val_map[src];
    if (mb) {
        if (mb->control_transfer_inst) {
            insert(mv, mb->control_transfer_inst);
        } else {
            push(mv, mb);
        }
    }
    return mv;
}

MI_VMove *emit_vmove(MOperand dst, MOperand src, Machine_Block *mb = NULL) {
    if (src.tag == IMM) {
        auto new_src = make_vreg(vreg_count++);
        auto load_const = emit_load_of_constant(new_src, src.value, mb);
        src = new_src;
    }
    auto mv = new MI_VMove(dst, src, false);
    if (!is_float_moperand(src)) {
        mv->from_s32 = true;
    }
    if (mb) {
        if (mb->control_transfer_inst) {
            insert(mv, mb->control_transfer_inst);
        } else {
            push(mv, mb);
        }
    }
    return mv;
}


// this is an api to load a constant to the reg
MI_Load *emit_load_of_constant(MOperand vreg, int32 constant, Machine_Block *mb, MI *I) {
    auto ldr = new MI_Load;
    ldr->mem_tag = MEM_LOAD_FROM_LITERAL_POOL;
    ldr->reg = vreg;
    ldr->base = make_imm(constant);
    int_val_map[vreg] = constant;
    if (mb) {
        if (I)
            insert(ldr, I);
        else if (mb->control_transfer_inst)
            insert(ldr, mb->control_transfer_inst);
        else
            push(ldr, mb);
    }
    return ldr;
}

//获得操作数的api, 将你所用到的MB换为可操作的对象
// this is to make an operand 
MOperand make_operand(ValuePtr v, Machine_Block *mb, bool no_imm=false) {
    // auto s=v->getStr();
    int ID=v->type->getID();
    int tp=-1;
    if(v->isConst)
        tp=0;
    else if(v->isReg)
        tp=1;
    else if(v->isVoid)
        tp=2;
    else {
        if(v->type->isArr())
            tp=4;
        else if(v->type->isInt())
            tp=3;
        else if(v->type->isPtr())
            tp=5;
        else if(v->type->isBool())
            tp=6;
        else if(v->type->isFloat())
            tp=7;
    }

    auto opr = value_map.find(v);
    if (opr != value_map.end()) {
        return opr->second;
    }
        
    switch(tp) {
        // const
        case 0: {
            int32 constant = dynamic_cast<Const*>(v.get())->intVal;
            if (v->type->isFloat()) {
                f2i.f_value = dynamic_cast<Const*>(v.get())->floatVal;
                constant = f2i.bin_value;
                no_imm = true;
            }
            auto imm = make_imm(constant);
            if (no_imm) {
                auto vreg = make_vreg(vreg_count++);
                int_val_map[vreg] = constant;
                // value_map[v] = vreg;
                MI_Load *ldr = emit_load_of_constant(vreg, constant, mb);
                MI_VMove *fcvt = nullptr;
                if(v->type->isFloat()) {
                    auto vsreg = make_vsreg(vsreg_count++);
                    fcvt = new MI_VMove(vsreg, vreg, true);
                    if (mb) {
                        if (mb->control_transfer_inst) {
                            insert(fcvt, mb->control_transfer_inst);
                        } else {
                            push(fcvt, mb);
                        }
                    }
                    else 
                        assert(false && "can not find mb");

                    vreg = vsreg;
                }
                return vreg;
            } else {
                return imm;
            }
        } break;
        // ptr
        case 5: {
            if (no_imm) {
                auto vreg = make_vreg(vreg_count++);
                value_map[v] = vreg;
                return vreg;
            } else {
                assert(false && "can not make a ptr");
                return make_imm(0);
            }
        } break;
        // int
        case 3: {
            int32 constant = dynamic_cast<Int*>(v.get())->intVal;
            auto imm = make_imm(constant);
            if (no_imm) {
                auto vreg = make_vreg(vreg_count++);
                MI_Load *ldr = emit_load_of_constant(vreg, constant, mb);
                int_val_map[vreg] = constant;
                return vreg;
            } else {
                return imm;
            }
        } break;
        // arr
        case 4: {
            if (no_imm) {
                auto vreg = make_vreg(vreg_count++);
                value_map[v] = vreg;
                return vreg;
            } else {
                assert(false && "can not make an array");
                return make_imm(0);
            }
        } break;

        default: {
            auto opr = value_map.find(v);
            if (opr != value_map.end()) { // @TODO: always false?
                return opr->second;
            }
            MOperand new_vreg;
            if (v->type->isFloat()) {
                new_vreg = make_vsreg(vsreg_count++);
            }
            else {
                new_vreg = make_vreg(vreg_count++);
            }
            value_map[v] = new_vreg;
            return new_vreg;
        }
    }
}


// this is to emit an binary instruction
void emit_binary_value_naive(InstructionPtr I, Machine_Block *mb) {
    switch (I->type)
    {
        case Fneg: {
            auto bi_I = dynamic_cast<FnegInstruction *>(I.get());
            auto op_type = F_NEG;
            if(bi_I->a->type->isFloat())
            {
                auto bii=new MI_VBinary;
                auto name1=bi_I->a->name;
                auto lhs = make_operand(bi_I->a, mb, true);
                auto rhs = lhs;

                // auto dst = make_vsreg(vsreg_count++);
                bii->dst = make_operand(bi_I->reg, mb, true);
                bii->op = op_type;
                bii->lhs = lhs;
                // bii->rhs=rhs;
                bii->fneg = 1;
                push((MI *)bii, mb);
                value_map[bi_I->reg] = bii->dst;
                break;
            }
                
            // todo: and or binary_type???
        } break;
        case Binary: {
            auto bi_I = dynamic_cast<BinaryInstruction *>(I.get());
            string myop;
            myop+=bi_I->op;
            auto op_type = Binary_ir2asm[myop];
            if(op_type==UNARY_XOR) // not rd, rs = xori rd, rs, -1 
            {   // @TODO: assume only 0 and 1
                assert(!bi_I->a->isConst);
                auto bi_sub = new MI_Binary;
                bi_sub->op = BINARY_SUBTRACT;
                auto imm = make_imm(1);
                auto imm_reg = make_vreg(vreg_count++);
                emit_move(imm_reg,imm,mb);
                bi_sub->lhs = imm_reg;
                bi_sub->rhs = make_operand(bi_I->a, mb, true);
                bi_sub->dst = make_operand(bi_I->reg, mb);
                push((MI *)bi_sub, mb);
            }
            else{
                if(bi_I->a->type->isFloat() || bi_I->b->type->isFloat())
                {
                    auto bii=new MI_VBinary;
                    auto name1=bi_I->a->name;
                    auto name2=bi_I->b->name;
                    auto lhs = make_operand(bi_I->a, mb, true);
                    auto rhs = make_operand(bi_I->b, mb, true);
                    assert(is_float_reg(lhs) || is_float_reg(rhs));
                    if(!is_float_reg(lhs)) {
                        auto fcvt = new MI_VCvt(S32, F32);
                        auto new_lhs = make_vsreg(vsreg_count++);
                        fcvt->src = lhs;
                        fcvt->dst = new_lhs;
                        push((MI *)fcvt, mb);
                        lhs = new_lhs;
                        value_map[bi_I->a] = lhs;
                    }
                    else if(!is_float_reg(rhs)) {
                        auto fcvt = new MI_VCvt(S32, F32);
                        auto new_rhs = make_vsreg(vsreg_count++);
                        fcvt->src = rhs;
                        fcvt->dst = new_rhs;
                        push((MI *)fcvt, mb);
                        rhs = new_rhs;
                        value_map[bi_I->b] = rhs;
                    }

                    auto dst = make_vsreg(vsreg_count++);
                    value_map[bi_I->reg]=dst;
                    bii->dst=dst;
                    bii->op=op_type;
                    bii->lhs=lhs;
                    bii->rhs=rhs;
                    push((MI *)bii, mb);
                    break;
                }
                auto name1=bi_I->a->name;
                auto name2=bi_I->b->name;
                auto lhs = make_operand(bi_I->a, mb, true);
                MOperand rhs;
                if (bi_I->op == '*' || bi_I->op == '/' || bi_I->op == '%')
                    rhs = make_operand(bi_I->b, mb, /* no imm for mul,div,rem */ true);
                else
                    rhs = make_operand(bi_I->b, mb);

                auto dst = make_operand(bi_I->reg, mb, true);
                // can not be load as 12-bit imm
                if(rhs.tag == IMM && !can_be_imm12(rhs.value)) {
                    auto rhs_reg = make_vreg(vreg_count++);
                    auto mv_imm = new MI_Move(rhs_reg, rhs);
                    push((MI *)mv_imm, mb);
                    rhs = rhs_reg;
                }

                auto bi = new MI_Binary(op_type, dst, lhs, rhs);
                bi->is_rv64 = bi_I->isrv64;
                push((MI *)bi, mb);
            } // todo: and or binary_type???
        } break;

        case Icmp : { // convert [cmp a, b] to {0, 1}
            assert(false && "can not handle Icmp in emit_binary_value_naive!");
            auto icmp_I = dynamic_cast<IcmpInstruction *>(I.get());
            auto lhs = make_operand(icmp_I->a, mb, true);
            auto rhs = make_operand(icmp_I->b, mb);
            auto cmp = new MI_Compare(lhs, rhs);
            push(cmp, mb);

            auto result_vreg = make_operand(icmp_I->reg, mb);

            auto false_mv = new MI_Move(result_vreg, make_imm(0));
            auto true_mv  = new MI_Move(result_vreg, make_imm(1));
            true_mv->cond = binary_op_to_branch_cond(Binary_ir2asm[icmp_I->op]);

            push(false_mv, mb);
            push(true_mv, mb);
        } break;

        case Fcmp : {
            auto fcmp_I = dynamic_cast<FcmpInstruction *>(I.get());
            auto lhs = make_operand(fcmp_I->a, mb, true);
            auto rhs = make_operand(fcmp_I->b, mb, true);
            // assert(is_float_reg(lhs) && is_float_reg(rhs));
            assert(is_float_reg(lhs) || is_float_reg(rhs));
            if(!is_float_reg(lhs)) {
                auto fcvt = new MI_VCvt(S32, F32);
                auto new_lhs = make_vsreg(vsreg_count++);
                fcvt->src = lhs;
                fcvt->dst = new_lhs;
                push((MI *)fcvt, mb);
                lhs = new_lhs;
                value_map[fcmp_I->a] = lhs;
            }
            else if(!is_float_reg(rhs)) {
                auto fcvt = new MI_VCvt(S32, F32);
                auto new_rhs = make_vsreg(vsreg_count++);
                fcvt->src = rhs;
                fcvt->dst = new_rhs;
                push((MI *)fcvt, mb);
                rhs = new_rhs;
                value_map[fcmp_I->b] = rhs;
            }
            auto result_vreg = make_operand(fcmp_I->reg, mb, true);
            auto cmp = new MI_VCmp_Set(result_vreg, lhs, rhs, binary_op_to_branch_cond(Binary_ir2asm[fcmp_I->op]));
            push(cmp, mb);
        } break;
        
        default:{
            auto bi_I = dynamic_cast<BinaryInstruction *>(I.get());
            string myop;
            myop+=bi_I->op;
            auto op_type = Binary_ir2asm[myop];
            // auto op_type = Binary_ir2asm[string(bi_I->op,1)];
            auto lhs = make_operand(bi_I->a, mb, true);
            MOperand rhs;
            // @TODO: unreachable?
            if (bi_I->op == '*' || bi_I->op == '/')
                rhs = make_operand(bi_I->b, mb, /* no imm for mul,div */ true);
            else
                rhs = make_operand(bi_I->b, mb);

            auto dst = make_operand(bi_I->reg, mb);
            auto bi = new MI_Binary(op_type, dst, lhs, rhs);
            push((MI *)bi, mb);
        } break;
    }
}

// this is to handle the branch instruction or the Binary instruction
void emit_legacy(Func_Asm* func_asm, InstructionPtr I, BasicBlockPtr bb, Machine_Block *mb, int32 &i) {
    switch(I->type) {
        case Br: {
            auto br_I = dynamic_cast<BrInstruction*>(I.get());
            auto tp=br_I->exp;
            auto cond=NO_CONDITION;
            if(tp!=NULL)
            {
                if(tp->getStr()=="true")
                {
                    emit_branch_asm(func_asm, cond, br_I, mb);
                    break;
                }
                else if(tp->getStr()=="false")
                {
                    emit_branch_asm_2(func_asm, cond, br_I, mb);
                    break;
                }

                // assert(false && "can not handle conditional branch without operands");
                cond=binary_op_to_branch_cond(BINARY_NOT_EQUAL_TO);
                emit_branch_asm(func_asm, NOT_EQUAL, br_I, mb, make_operand(br_I->exp, mb, true), make_reg(zero));
            }
            else 
            {
                emit_branch_asm(func_asm, cond, br_I, mb);
            }
                
        } break;
        case Fneg:
        case Binary : {
            auto next_inst = bb->instructions[i + 1]; // todo : maybe get some wrong answers now
            if(next_inst->type != Br){
                emit_binary_value_naive(I, mb);
            }else{
                emit_binary_value_naive(I, mb);
            }
        } break;

        case Icmp : {
            assert(i + 1 < bb->instructions.size());
            auto next_inst = bb->instructions[i + 1]; // todo : maybe get some wrong answers now
            if(next_inst->type == Br) { // assume only branch instr after Icmp
                auto icmp_I = dynamic_cast<IcmpInstruction*>(I.get());
                auto lhs = make_operand(icmp_I->a, mb, true);
                auto rhs = make_operand(icmp_I->b, mb, true);

                i++;
                auto br_I = (BrInstruction *)(next_inst.get());
                emit_branch_asm(func_asm, binary_op_to_branch_cond(Binary_ir2asm[icmp_I->op]), br_I, mb, lhs, rhs);
            }
            else {
                emit_binary_icmp(dynamic_cast<IcmpInstruction*>(I.get()), mb);
            }

        } break;
        
        case Fcmp : {
            auto next_inst = bb->instructions[i + 1]; // todo : maybe get some wrong answers now
            if(next_inst->type != Br){
                emit_binary_value_naive(I, mb);
            }else{
                auto fcmp_I = dynamic_cast<FcmpInstruction *>(I.get());
                auto lhs = make_operand(fcmp_I->a, mb, true);
                auto rhs = make_operand(fcmp_I->b, mb, true);
                assert(is_float_reg(lhs) || is_float_reg(rhs));
                if(!is_float_reg(lhs)) {
                    auto new_lhs = make_vsreg(vsreg_count++);
                    MI_VCvt *fcvt = new MI_VCvt(S32, F32);
                    fcvt->src = lhs;
                    fcvt->dst = new_lhs;
                    push((MI *)fcvt, mb);
                    lhs = new_lhs;
                    value_map[fcmp_I->a] = new_lhs;
                }
                else if(!is_float_reg(rhs)) {
                    auto new_rhs = make_vsreg(vsreg_count++);
                    MI_VCvt *fcvt = new MI_VCvt(S32, F32);
                    fcvt->src = rhs;
                    fcvt->dst = new_rhs;
                    push((MI *)fcvt, mb);
                    rhs = new_rhs;
                    value_map[fcmp_I->b] = new_rhs;
                }

                i++;
                auto br_I = (BrInstruction *)(next_inst.get());
                emit_branch_asm(func_asm, binary_op_to_branch_cond(Binary_ir2asm[fcmp_I->op]), br_I, mb, lhs, rhs, true);
            }
            
        } break;
    }
}

// this is to handle the asm of a function
Func_Asm* emit_function_asm(FunctionPtr func, int idx, Module program_module, vector<VariablePtr> &globalValues)
{
    vreg_count = 0;
    vsreg_count = 0;
    value_map.clear();
    int_val_map.clear();
    ptr_val_map.clear();

    auto func_asm = new Func_Asm;
    func_asm->index = idx;
    func_asm->name = func->name;
    // func_asm->stack_size = 100;
    func_asm->stack_size = 0;

    auto mybb = func->block;
    int ii = 0;
    int n = mybb->basicBlocks.size();

    for(auto bb : mybb->basicBlocks) {
        func_asm->mbs.push(new Machine_Block());
        func_asm->mbs.back()->i = ii;
        func_asm->mbs.back()->label = bb->label;
        ii++;
        func_asm->mbs.back()->loop_depth = bb->loopDepth;
    }

    int ii2=0;

    for(auto bb: mybb->basicBlocks)
    {
        for(auto x: bb->succBasicBlocks)
        {
            int pidx=0;
            for(auto y : mybb->basicBlocks)
            {
                if(x==y)
                    break;
                pidx++;
            }
            func_asm->mbs[ii2]->succs.push(func_asm->mbs[pidx]);
        }
        for(auto x: bb->predBasicBlocks)
        {
            int pidx=0;
            for(auto y : mybb->basicBlocks)
            {
                if(x==y)
                    break;
                pidx++;
            }
            func_asm->mbs[ii2]->preds.push(func_asm->mbs[pidx]);
        }
        ii2++;
    }

    for (int i = 0; i < 4 && i < func->formArguments.size(); i++) {
        auto arg = func->formArguments[i];
        if(arg->type->isFloat())
        {
            auto vsreg= make_vsreg(vsreg_count++);
            value_map[arg]=vsreg;
            emit_vmove(vsreg, make_sreg(i + fa0), func_asm->mbs[0]);
            continue;
        }
        // @TODO: use 'else' instead of 'continue'
        auto vreg = make_vreg(vreg_count++);
        value_map[arg] = vreg;
        emit_move(vreg, make_reg(i + a0), func_asm->mbs[0]);
    }

    // want to load more args to the function
    // @ todo make it correct
    int cum_offset = 0;
    for (int i = 4; i < func->formArguments.size(); i++) {
        auto arg = func->formArguments[i];
        int arg_size;
        bool is_float = arg->type->isFloat();
        auto a = is_float ? make_vsreg(vsreg_count++) : make_vreg(vreg_count++);
        // load pointer(64 bits) or int/float(32 bits)
        if (arg->type->ID == ArrID || arg->type->ID == PtrID || arg->type->ID == Int64ID)
            arg_size = 8;
        else
            arg_size = 4;
       
        emit_load_of_later_arg(arg, a, func_asm->mbs[0], cum_offset, arg_size, is_float);
        cum_offset += arg_size;
        
        value_map[arg] = a;
    }

    int bbidx=0;
    for(auto bb : mybb->basicBlocks) {
        Machine_Block *mb = func_asm->mbs[bbidx];
        bbidx++;
        for(int i = 0; i < bb->instructions.size(); i++) {
            auto I = bb->instructions[i];
            switch(I->type) {
                case Br: {
                    if (!opt) {
                        emit_legacy(func_asm,I, bb, mb, i);
                        break;
                    }
                } break;

                
                case Fneg:
                case Binary: {
                    if (!opt) {
                        emit_legacy(func_asm, I, bb, mb, i);
                        break;
                    }
                }

                case Icmp: {
                    if (!opt){
                        emit_legacy(func_asm, I, bb, mb, i);
                        break;
                    }
                }

                case Fcmp: {
                    if (!opt){
                        emit_legacy(func_asm, I, bb, mb, i);
                        break;
                    }
                }

                case Sitofp:
                {   
                    auto mi = new MI_VCvt;
                    auto conv_I = dynamic_cast<SitofpInstruction*>(I.get());
                    auto src = make_operand(conv_I->from, mb, true);
                    auto dst = make_operand(conv_I->reg, mb, true);
                    value_map[conv_I->reg]=dst;
                    mi->to_type = F32;
                    mi->from_type = S32;

                    mi->dst = dst;
                    mi->src = src;
                    push((MI *)mi,mb);
                }break;


                case Fptosi:
                {   
                    auto mi = new MI_VCvt;
                    auto conv_I = dynamic_cast<FptosiInstruction*>(I.get());
                    auto src = make_operand(conv_I->from, mb, true);
                    auto dst = make_operand(conv_I->reg, mb, true);
                    value_map[conv_I->reg] = dst;
                    mi->to_type = S32;
                    mi->from_type = F32;

                    mi->dst = dst;
                    mi->src = src;
                    push((MI *)mi,mb);
                }break;

                // this is to handle the get pointer instruction
                case GEP:
                {
                    auto ptr_I = dynamic_cast<GetElementPtrInstruction *>(I.get());
                    int flag2 = 0;
                    auto sname = ptr_I->from->name;
                    int isarrele = 0;
                    // value_map[ptr_I->reg] = make_vreg(vreg_count++);
                    // handle loop-combine GEP
                    if(ptr_I->loopGEP != nullptr) {
                        if (my_arr_base.find(ptr_I->from) == my_arr_base.end())
                            my_arr_base[ptr_I->reg] = ptr_I->from;
                        else
                            my_arr_base[ptr_I->reg] = my_arr_base[ptr_I->from];
                        int step_offset = ptr_I->loopGEP->getStepOffset() * 4;
                        auto step_offset_opr = make_imm(step_offset);
                        auto base = make_operand(ptr_I->reg, mb, true);
                        if(!can_be_imm12(step_offset)) {
                            step_offset_opr = make_vreg(vreg_count++);
                            emit_load_of_constant(step_offset_opr, step_offset, mb);
                        }
                        auto cal_new_base = new MI_Binary(BINARY_ADD, base, base, step_offset_opr);
                        cal_new_base->is_rv64 = true;
                        push(cal_new_base, mb);
                        // value_map[ptr_I->reg] = base;
                        break;
                    }

                    value_map[ptr_I->reg] = make_vreg(vreg_count++);
                    if (ptr_I->from->name.find("arrayidx") == ptr_I->from->name.npos)
                    {
                        if (my_arr_base.find(ptr_I->from) == my_arr_base.end())
                            my_arr_base[ptr_I->reg] = ptr_I->from;
                        else
                            my_arr_base[ptr_I->reg] = my_arr_base[ptr_I->from];
                        int flag = 0;
                        for (auto x : globalValues)
                        {
                            if (ptr_I->from == x)
                            {
                                flag = 1;
                                break;
                            }
                        }
                        if (flag)
                        {
                            auto nvreg = make_vreg(vreg_count++);
                            emit_load_of_global_ref(func_asm, ptr_I->from, nvreg, mb);
                            value_map[ptr_I->from] = nvreg;
                        }
                    }
                    else
                    {
                        my_arr_base[ptr_I->reg] = my_arr_base[ptr_I->from];
                    }

                    if (ptr_I->reg->name.find("arrayinit") != ptr_I->reg->name.npos) {
                        isarrele = 1;
                    }

                    TypePtr cur_arr = NULL;
                    auto reg4 = value_map[ptr_I->from];
                    int cff = 0;
                    int offset = 0;
                    int ff = 0;
                    // has constant offset before
                    if (ptr_val_map.find(ptr_I->from) != ptr_val_map.end())
                    {
                        // offset = ptr_val_map.find(ptr_I->from)->second;
                        cff = 1;
                    }

                    // index number is 1
                    if (value_map.find(ptr_I->from) != value_map.end() && ptr_I->from->name.find("arrayinit") != ptr_I->from->name.npos && !cff)
                    // if (value_map.find(ptr_I->from) != value_map.end() && ptr_I->from->name.find("arrayinit") != ptr_I->from->name.npos)
                    {
                        ff = 1;
                        reg4 = value_map[ptr_I->from];
                    }

                    const int max_dims = 40;
                    MOperand mul_regs[max_dims + 1];
                    vector<MI *> mul_instr(max_dims + 1, nullptr);

                    int cnt = 0;
                    for (auto x : ptr_I->index)
                    {
                        if (!x->isConst)
                            ff = 1;
                    }
                    if (isarrele)
                    {
                        for (int i = 0; i < ptr_I->index.size(); i++)
                        {
                            cnt++;
                            auto x = ptr_I->index[i];
                            int const_val = 0;

                            if (x->isConst)
                                const_val = dynamic_cast<Const *>(x.get())->intVal;
                            else
                            {
                                ff = 1;
                            }

                            int const_size = 1;
                            if (i == 0)
                            {
                                if(ptr_I->from->type->ID == PtrID) {
                                    auto tmp = dynamic_cast<PtrType *>(ptr_I->from->type.get());
                                    if (tmp->inner->isArr())
                                        const_size = dynamic_cast<ArrType *>(tmp->inner.get())->getSize();
                                }
                                else if(ptr_I->from->type->ID == IntID || ptr_I->from->type->ID == FLoatID) {
                                    const_size = 1;
                                }
                                else {
                                    assert(false && "can not handle array type");
                                }
                            }
                            if (i == 1)
                            {
                                assert(ptr_I->from->type->ID == ArrID);
                                auto tmp = dynamic_cast<ArrType *>(ptr_I->from->type.get());
                                if (tmp->inner->isArr())
                                    const_size = dynamic_cast<ArrType *>(tmp->inner.get())->getSize();
                                cur_arr = tmp->inner;
                            }
                            else if (i > 1)
                            {
                                assert(cur_arr->ID == ArrID);
                                auto nc = dynamic_cast<ArrType *>(cur_arr.get());
                                if (nc->inner->isArr())
                                    const_size = dynamic_cast<ArrType *>(nc->inner.get())->getSize();
                                cur_arr = nc->inner;
                            }
                            if (!ff)
                                offset += const_size * const_val;
                            else
                            {
                                if (x != ptr_I->index.back() || const_size != 1)
                                {
                                    ValuePtr imptr = ValuePtr(new Int("asdf", false, false, const_size));
                                    auto bi_mul = new MI_Binary;
                                    bi_mul->op = BINARY_MULTIPLY;
                                    bi_mul->lhs = make_operand(x, mb, true);
                                    bi_mul->rhs = make_operand(imptr, mb, true);
                                    assert(cnt > 0 && cnt < max_dims);
                                    mul_regs[cnt] = make_vreg(vreg_count++);
                                    bi_mul->dst = mul_regs[cnt];
                                    // push((MI *)bi_mul, mb);
                                    mul_instr[cnt] = bi_mul;
                                }
                                else
                                {
                                    assert(cnt > 0 && cnt < max_dims);
                                    mul_regs[cnt] = make_operand(x, mb, true);
                                }
                            }
                        }
                    }
                    else
                    {
                        for (int i = 1; i < ptr_I->index.size(); i++)
                        {
                            auto x = ptr_I->index[i];
                            int const_val = 0;
                            cnt++;

                            if (x->isConst)
                                const_val = dynamic_cast<Const *>(x.get())->intVal;
                            else
                            {
                                ff = 1;
                            }

                            int const_size = 1;
                            if (i == 1)
                            {
                                auto tmp = dynamic_cast<ArrType *>(ptr_I->from->type.get());
                                if (tmp->inner->isArr())
                                    const_size = dynamic_cast<ArrType *>(tmp->inner.get())->getSize();
                                cur_arr = tmp->inner;
                            }
                            else
                            {
                                auto nc = dynamic_cast<ArrType *>(cur_arr.get());
                                if (nc->inner->isArr())
                                    const_size = dynamic_cast<ArrType *>(nc->inner.get())->getSize();
                                cur_arr = nc->inner;
                            }
                            if (!ff)
                                offset += const_size * const_val;
                            else
                            {

                                if (x != ptr_I->index.back())
                                {
                                    ValuePtr imptr = ValuePtr(new Int("asdf", false, false, const_size));
                                    auto bi_mul = new MI_Binary;
                                    bi_mul->op = BINARY_MULTIPLY;
                                    bi_mul->lhs = make_operand(x, mb, true);
                                    bi_mul->rhs = make_operand(imptr, mb, true);
                                    assert(cnt > 0 && cnt < max_dims);
                                    mul_regs[cnt] = make_vreg(vreg_count++);
                                    bi_mul->dst = mul_regs[cnt];
                                    // push((MI *)bi_mul, mb);
                                    mul_instr[cnt] = bi_mul;
                                }
                                else
                                {
                                    assert(cnt > 0 && cnt < max_dims);
                                    mul_regs[cnt] = make_operand(x, mb, true);
                                }
                            }
                        }
                    }
                    if (!ff)
                    {
                        ptr_val_map[ptr_I->reg] = offset;
                        if(ptr_val_map.count(ptr_I->from) != 0) {
                            ptr_val_map[ptr_I->reg] += ptr_val_map[ptr_I->from];
                        }
                        value_map[ptr_I->reg] = value_map[ptr_I->from];
                        ptr_I->isConstIdx = true;
                    }
                    // calculate accurate address for new base
                    else
                    {
                        if (cnt == 1)
                        {
                            auto old_base = value_map[ptr_I->from];
                            if(ptr_val_map.count(ptr_I->from) != 0) {
                                auto old_offset = make_legal_imm(ptr_val_map[ptr_I->from] * 4, mb);
                                auto tmp_base = make_vreg(vreg_count++);
                                auto bi_add = new MI_Binary(BINARY_ADD, tmp_base, old_base, old_offset);
                                bi_add->is_rv64 = true;
                                push(bi_add, mb);
                                old_base = tmp_base;
                            }
                            if(mul_instr[1] != nullptr) {
                                push(mul_instr[1], mb);
                            }
                            // auto lsl_offset = make_vreg(vreg_count++);
                            // auto bi_lsl = new MI_Binary(BINARY_LSL, lsl_offset, mul_regs[1], make_imm(2));
                            // auto bi_add = new MI_Binary(BINARY_ADD, make_operand(ptr_I->reg, mb, true), old_base, lsl_offset);
                            // bi_add->is_rv64 = true;
                            // push(bi_lsl, mb);
                            // push(bi_add, mb);

                            auto bi_sh2add = new MI_Binary(BINARY_SH2ADD, make_operand(ptr_I->reg, mb, true), mul_regs[1], old_base);
                            push(bi_sh2add, mb);
                        }
                        else if (cnt > 1) 
                        {
                            MI_Binary *bi_add;
                            assert(cnt < max_dims);
                            int i;
                            auto last_reg = mul_regs[1];
                            if(mul_instr[1] != nullptr) {
                                push(mul_instr[1], mb);
                            }
                            for (i = 1; i < cnt; i ++) {
                                if(mul_instr[i + 1] != nullptr) {
                                    push(mul_instr[i + 1], mb);
                                }
                                MOperand dst = make_vreg(vreg_count++);
                                bi_add = new MI_Binary(BINARY_ADD, dst, last_reg, mul_regs[i + 1]);
                                bi_add->is_rv64 = true;
                                push((MI *)bi_add, mb);
                                last_reg = bi_add->dst;
                            }

                            // calculate pointer address
                            auto offset = bi_add->dst;
                            auto lsl_offset = make_vreg(vreg_count++);
                            auto old_base = make_operand(ptr_I->from, mb);
                            auto new_base = make_operand(ptr_I->reg, mb, true);
                            // clear former constant offset
                            if(ptr_val_map.count(ptr_I->from) != 0) {
                                auto old_offset = make_legal_imm(ptr_val_map[ptr_I->from] * 4, mb);
                                auto tmp_base = make_vreg(vreg_count++);
                                bi_add = new MI_Binary(BINARY_ADD, tmp_base, old_base, old_offset);
                                bi_add->is_rv64 = true;
                                push(bi_add, mb);
                                old_base = tmp_base;
                            }
                            // @TODO: can be optimized by sh2add
                            // auto bi_lsl = new MI_Binary(BINARY_LSL, lsl_offset, offset, make_imm(2));
                            // bi_add = new MI_Binary(BINARY_ADD, new_base, old_base, lsl_offset);
                            // bi_add->is_rv64 = true;
                            // push(bi_lsl, mb);
                            // push(bi_add, mb);

                            auto bi_sh2add = new MI_Binary(BINARY_SH2ADD, new_base, offset, old_base);
                            push(bi_sh2add, mb);
                        }
                    }
                } break;

                // this is to handle the return value
                case Return: {
                    auto ret_I = dynamic_cast<ReturnInstruction*>(I.get());
                    auto ret = new MI_Return;
                    auto retv=ret_I->retValue;
                    auto myrev=0;
                    if(retv->type->isFloat())
                    {
                        auto ret_value = make_operand(ret_I->retValue, mb);
                        value_map[ret_I->retValue] = ret_value;
                        emit_vmove(make_sreg(fa0), ret_value, mb);
                        push((MI *)ret, mb);

                        mb->control_transfer_inst = (MI *)ret;
                        break;
                    }
                    if(retv->isConst) {
                        myrev = dynamic_cast<Const*>(retv.get())->intVal;
                        emit_move(make_reg(a0), make_legal_imm(myrev, mb), mb);
                    }
                    else if(!retv->isVoid) {
                        auto ret_value = make_operand(retv, mb);
                        emit_move(make_reg(a0), ret_value, mb);
                    }
                    push((MI *)ret, mb);

                    mb->control_transfer_inst = (MI *)ret;
                } break;

                // deal with the ext instruction
                case Ext:
                {
                    auto ext_I=dynamic_cast<ExtInstruction*>(I.get());
                    auto oreg=make_operand(ext_I->from,mb,true);
                    value_map[ext_I->reg]=oreg;

                }break;
                
                // deal with the call instruction
                // @todo: fix the function call of many args
                case Call: {
                    auto func_I = dynamic_cast<CallInstruction*>(I.get());
                    auto func_name=func_I->func->name;
                    if(func_name.size() > 15 && func_name.find("memset") != func_name.npos)
                    {
                        auto call = new MI_Func_Call("memset");
                        call->arg_count = func_I->func->formArguments.size()-1;
                        for(int i = 0; i < func_I->func->formArguments.size()-1; i++) {
                            auto arg_value = func_I->argv[i];

                            MOperand arg_operand;
                            arg_operand = make_reg(i + a0);
                            emit_move(arg_operand, make_operand(arg_value, mb), mb);
                        }

                        push((MI *)call, mb);

                        ValuePtr ret = func_I->func->retVal;

                        if (!ret->isVoid) {
                            auto result = make_operand(func_I->reg, mb,true);
                            if (ret->type->isFloat()) {
                                emit_vmove(result, make_reg(fa0), mb);
                            }
                            else {
                                emit_move(result, make_reg(a0), mb);
                            }
                            //如果有返回值必须存放到r0的寄存器
                        }
                        break;
                    }
                    //我觉得是把普通参数和浮点数参数分开
                    //分别是(int 数组) 和 (float) 二者都是用4这个范围区分开
                    //和原来的区别就是浮点数寄存器的特殊指令，以及类型转换，其他可以复用
                    auto call = new MI_Func_Call(func_I->func->name.c_str());
                    call->arg_count = func_I->func->formArguments.size();
                    int myff=0;
                    int offset = 0;
                    for(int i = 0; i < func_I->func->formArguments.size(); i++) {
                        auto arg_value = func_I->argv[i];
                        MOperand arg_operand;
                        if (i < 4) {
                            if(stack_val_map.find(arg_value)!=stack_val_map.end())
                            {
                                arg_operand = make_reg(i + a0);
                                auto ldr = new MI_Load;
                                ldr->reg = make_vreg(vreg_count++);
                                ldr->base = get_stack_val(arg_value,mb,func_asm);
                                push((MI *)ldr, mb);
                                emit_move(arg_operand, ldr->reg, mb);
                            }
                            else
                            {
                                arg_operand = make_reg(i + a0);
                                if(arg_value->type->isFloat())
                                {
                                    
                                    auto arg_opr = make_operand(arg_value, mb, true);
                                    // float array(pointer) was calculated, avoid redundant address calculation
                                    if(array_ptr_map.find(arg_value) != array_ptr_map.end())
                                    {
                                        arg_opr = array_ptr_map[arg_value];
                                        emit_move(arg_operand, arg_opr, mb);
                                    }
                                    // float array(pointer) calculation
                                    else if(my_arr_base.find(arg_value) != my_arr_base.end())
                                    {
                                        // GEP with constant offset
                                        if(ptr_val_map.find(arg_value) != ptr_val_map.end()) {
                                            auto opr = ptr_val_map.find(arg_value);
                                            int offset = opr->second * 4;
                                            auto cur_arr_base_opr = make_operand(arg_value, mb);
                                            auto offset_opr = make_imm(offset);
                                            if(!can_be_imm12(offset)) {
                                                offset_opr = make_vreg(vreg_count++);
                                                emit_load_of_constant(offset_opr, offset, mb);
                                            }
                                            arg_opr = make_vreg(vreg_count++);
                                            auto bi_add = new MI_Binary(BINARY_ADD, arg_opr, cur_arr_base_opr, offset_opr);
                                            bi_add->is_rv64 = true;
                                            push(bi_add, mb);
                                        }
                                        // GEP with offset variable
                                        else {
                                            arg_opr = make_operand(arg_value, mb, true);
                                        }

                                        // add map for calculated float array(pointer) address
                                        array_ptr_map[arg_value] = arg_opr;
                                        emit_move(arg_operand, arg_opr, mb);
                                    }
                                    else 
                                    {
                                        arg_operand = make_sreg(i + fa0);
                                        emit_vmove(arg_operand, arg_opr, mb);
                                    }
                                    continue;
                                }
                                auto arg_opr = make_operand(arg_value, mb);
                                // int array(pointer) was calculated
                                if(array_ptr_map.find(arg_value) != array_ptr_map.end())
                                {
                                    arg_opr = array_ptr_map[arg_value];
                                }
                                // int array(pointer) calculation
                                else if(my_arr_base.find(arg_value)!=my_arr_base.end())
                                {
                                    // GEP with constant offset
                                    if(ptr_val_map.find(arg_value) != ptr_val_map.end()) {
                                        auto opr = ptr_val_map.find(arg_value);
                                        int offset = opr->second * 4;
                                        auto cur_arr_base_opr = make_operand(arg_value, mb);
                                        auto offset_opr = make_imm(offset);
                                        if(!can_be_imm12(offset)) {
                                            offset_opr = make_vreg(vreg_count++);
                                            emit_load_of_constant(offset_opr, offset, mb);
                                        }
                                        arg_opr = make_vreg(vreg_count++);
                                        auto bi_add = new MI_Binary(BINARY_ADD, arg_opr, cur_arr_base_opr, offset_opr);
                                        bi_add->is_rv64 = true;
                                        push(bi_add, mb);
                                    }
                                    // GEP with offset variable
                                    else {
                                        arg_opr = make_operand(arg_value, mb, true);
                                    }

                                    // add map for calculated float array(pointer) address
                                    array_ptr_map[arg_value] = arg_opr;
                                    emit_move(arg_operand, arg_opr, mb);
                                    // add map for calculated int array(pointer) address
                                    array_ptr_map[arg_value] = arg_opr;
                                }
                                emit_move(arg_operand, arg_opr, mb);
                            }
                            
                        } else {
                            // for cases i >= 4
                            if(ptr_val_map.find(arg_value)!=ptr_val_map.end())
                            {
                                cur_arr_base = my_arr_base[arg_value];
                                auto opr = ptr_val_map.find(arg_value);
                                auto bi_add = new MI_Binary;
                                bi_add->op = BINARY_ADD;
                                bi_add->lhs = make_operand(cur_arr_base, mb);
                                bi_add->rhs = make_legal_imm(opr->second*4, mb);
                                bi_add->dst = make_vreg(vreg_count++);
                                bi_add->is_rv64 = true;
                                push((MI *)bi_add, mb);
                                auto str = new MI_Store;
                                // can be optimized if 'offset = imm12 + imm12'
                                if(!can_be_imm12(offset)) {
                                    MOperand offset_opr = make_vreg(vreg_count++);
                                    emit_load_of_constant(offset_opr, offset, mb);
                                    auto new_base_add = new MI_Binary(BINARY_ADD, make_vreg(vreg_count++), make_reg(sp), offset_opr);
                                    new_base_add->is_rv64 = true;
                                    push(new_base_add, mb);
                                    str->base = new_base_add->dst;
                                    str->offset = make_imm(0);
                                }
                                else {
                                    str->base = make_reg(sp);
                                    str->offset = make_imm(offset);
                                }
                                str->mem_tag = MEM_PREP_ARG;
                                str->reg = bi_add->dst;
                                str->base.size = 8;
                                str->is_rv64 = true;
                                push((MI *)str, mb);
                                offset += 8;
                            }
                            // move argument to the stack
                            else
                            {
                                if (arg_value->type->isFloat()) 
                                {
                                    auto str = new MI_VStore;
                                    if(!can_be_imm12(offset)) {
                                        MOperand offset_opr = make_vreg(vreg_count++);
                                        emit_load_of_constant(offset_opr, offset, mb);
                                        auto new_base_add = new MI_Binary(BINARY_ADD, make_vreg(vreg_count++), make_reg(sp), offset_opr);
                                        new_base_add->is_rv64 = true;
                                        push(new_base_add, mb);
                                        str->base = new_base_add->dst;
                                        str->offset = make_imm(0);
                                    }
                                    else {
                                        str->base = make_reg(sp);
                                        str->offset = make_imm(offset);
                                    }
                                    str->mem_tag = MEM_PREP_ARG;
                                    str->reg = make_operand(arg_value, mb, true);
                                    str->base.size = 8;
                                    if (arg_value->type->ID == ArrID || arg_value->type->ID == PtrID || arg_value->type->ID == Int64ID) {
                                        str->is_rv64 = true;
                                        offset += 8;
                                    }
                                    else offset += 4;
                                    push((MI *)str, mb);
                                }
                                else 
                                {
                                    auto str = new MI_Store;
                                    if(!can_be_imm12(offset)) {
                                        MOperand offset_opr = make_vreg(vreg_count++);
                                        emit_load_of_constant(offset_opr, offset, mb);
                                        auto new_base_add = new MI_Binary(BINARY_ADD, make_vreg(vreg_count++), make_reg(sp), offset_opr);
                                        new_base_add->is_rv64 = true;
                                        push(new_base_add, mb);
                                        str->base = new_base_add->dst;
                                        str->offset = make_imm(0);
                                    }
                                    else {
                                        str->base = make_reg(sp);
                                        str->offset = make_imm(offset);
                                    }
                                    str->mem_tag = MEM_PREP_ARG;
                                    str->reg = make_operand(arg_value, mb, true);
                                    str->base.size = 8;
                                    if (arg_value->type->ID == ArrID || arg_value->type->ID == PtrID || arg_value->type->ID == Int64ID) {
                                        str->is_rv64 = true;
                                        offset += 8;
                                    }
                                    else offset += 4;
                                    push((MI *)str, mb);
                                }
                            }
                        }
                    }
                    // if (offset % 16 != 0) {
                    //     offset += 16 - (offset % 16);
                    // }
                    call->arg_stack_size = offset;
                    if(!myff)
                        push((MI *)call, mb);

                    ValuePtr ret = func_I->func->retVal;
                    if (!ret->isVoid) {
                        if(ret->type->isFloat()){
                            auto nsreg=make_vsreg(vsreg_count++);
                            value_map[func_I->reg]=nsreg;
                            emit_vmove(nsreg,make_sreg(fa0), mb);
                        }else{
                            auto result = make_operand(func_I->reg, mb,true);
                            emit_move(result, make_reg(a0), mb);
                        }
                    }
                } break;

                // this is to deal with the allocate instruction
				case Alloca: {
                    // cout<<"Alloca"<<endl;
					auto alloc = dynamic_cast<AllocaInstruction*>(I.get());
                    auto tmp=dynamic_cast<Variable*>(alloc->des.get());
                    // auto mytp=tmp->type;
                    int tp=tmp->type->getID();
                    auto const_value=0;
                    // cout<<"Alloca type: "<<tp<<endl;
                    switch(tp)
                    {
                        case IntID:
                        {
                            auto intptr=dynamic_cast<Int *>(tmp);
                            const_value=4;
                            func_asm->stack_size += const_value;
                        }break;

                        case ArrID:
                        {
                            auto mytype = dynamic_cast<ArrType *>(tmp->type.get());
                            auto myname = tmp->name;
                            auto ptr = mytype->inner;

                            const_value = mytype->getSize() * 4;
                            func_asm->stack_size += const_value;
                        }break;

                        case PtrID:
                        {
                            const_value = 4;
                            func_asm->stack_size += const_value;
                        }break;

                        default: {
                            printf("Alloca initial type: %d\n", tp);
                            assert(false && "unknown initial value type");
                        }
                    }
                    if(tp == IntID)
                        break;

                    // stack align for 16B
                    if (func_asm->stack_size % 16 != 0) 
                    {
                        func_asm->stack_size += 16 - (func_asm->stack_size % 16);
                    }
					auto bi_I = new MI_Binary;
					bi_I->op = BINARY_SUBTRACT;
					bi_I->lhs = make_reg(sp);
                    // will be fixup after register allocation
					bi_I->rhs = make_imm(func_asm->stack_size);
					bi_I->dst = make_operand(alloc->des, mb, true);
                    bi_I->lhs.size = 8;
                    bi_I->is_rv64 = true;
                    func_asm->local_array_bases.push(bi_I);
					push((MI *)bi_I, mb);
 

				}	break;


                // bitcast instruction: no need to modify the reg
                case Bitcast:
                {
                    auto bi_I = dynamic_cast<BitCastInstruction *>(I.get());
                    auto oreg = make_operand(bi_I->from, mb);
                    value_map[bi_I->reg] = oreg;
                } break;


                // load instruction
                // 1: array type, has const offset
                // 2: array type, no const offset
                // 3: reg, simple
                // 4: reg, global
                // 5: reg, store in stack
				case Load: {
					auto Read_I = dynamic_cast<LoadInstruction*>(I.get());
					auto ldr = new MI_Load;
					
					ldr->base = make_operand(Read_I->from, mb);

                    // we would not load a pointer, so get the actual type from PtrType
                    // store instruction do not need it
                    auto type = Read_I->from->type.get();
                    while(type->isPtr()) {
                        type = dynamic_cast<PtrType *>(type)->inner.get();
                    }

                    // load float type
                    if(type->isFloat())
                    {
                        auto ldr2 = new MI_VLoad;
                        // array
                        if(my_arr_base.find(Read_I->from)!=my_arr_base.end())
                        {
                            assert(Read_I->from->I && Read_I->from->I->type == GEP);
                            // array with determined offset
                            if(ptr_val_map.find(Read_I->from)!=ptr_val_map.end())
                            {
                                ldr2->reg = make_vsreg(vsreg_count++);
                                value_map[Read_I->to] = ldr2->reg;
                                auto opr = ptr_val_map.find(Read_I->from);
                                int offset = opr->second * 4;
                                auto cur_arr_base_opr = make_operand(Read_I->from, mb);

                                if(!can_be_imm12(offset)) {
                                    auto offset_opr = make_vreg(vreg_count++);
                                    emit_load_of_constant(offset_opr, offset, mb);
                                    auto new_base_add = new MI_Binary(BINARY_ADD, make_vreg(vreg_count++), cur_arr_base_opr, offset_opr);
                                    new_base_add->is_rv64 = true;
                                    push(new_base_add, mb);
                                    ldr2->base = new_base_add->dst;
                                    ldr2->offset = make_imm(0);
                                }
                                else {
                                    ldr2->base = cur_arr_base_opr;
                                    ldr2->offset = make_imm(offset);
                                }
                                cur_arr_base=NULL;
                            }
                            // array with undetermined offset
                            else
                            {
                                ldr2->reg = make_vsreg(vsreg_count++);
                                value_map[Read_I->to] = ldr2->reg;
                                auto tp = make_vreg(vreg_count++);
                                auto cur_arr_base_opr = make_operand(Read_I->from, mb);
                                ldr2->base = cur_arr_base_opr;
                                cur_arr_base=NULL;
                            }
                        }
                        else
                        {
                            int flag=0;
                            for(auto v:globalValues)
                            {
                                if(v==Read_I->from)
                                {
                                    flag=1;
                                    break;
                                }
                            }
                            if(stack_val_map.find(Read_I->from)!=stack_val_map.end())
                                flag=2;
                            // common variable
                            if(flag==0)
                            {
                                ldr2->reg = make_vreg(vreg_count++);
                                value_map[Read_I->to] = ldr2->reg;
                                ldr2->base = make_operand(Read_I->from, mb);
                                emit_move(ldr2->reg,ldr2->base,mb);
                                break;
                            }
                            // global variable
                            else if (flag==1)
                            {
                                auto nvreg = make_vreg(vreg_count++);
                                emit_load_of_global_ref(func_asm,Read_I->from,nvreg,mb);
                                ldr2->base = nvreg;
                                ldr2->reg = make_operand(Read_I->to, mb);
                                value_map[Read_I->to] = ldr2->reg;
                            }
                            // stack variable
                            else
                            {
                                ldr2->reg = make_vreg(vreg_count++);
                                value_map[Read_I->to] = ldr2->reg;
                                ldr2->base = get_stack_val(Read_I->from,mb,func_asm);
                            }
                        }
					    push((MI *)ldr2, mb);
                        break;
                    }
                    // load int type
                    if(my_arr_base.find(Read_I->from)!=my_arr_base.end())
                    {
                        assert(Read_I->from->I && Read_I->from->I->type == GEP);
                        // array with determined offset
                        if(ptr_val_map.find(Read_I->from)!=ptr_val_map.end())
                        {
                            ldr->reg = make_operand(Read_I->to, mb, true);
                            auto opr = ptr_val_map.find(Read_I->from);
                            int offset = opr->second * 4;
                            auto cur_arr_base_opr = make_operand(Read_I->from, mb);

                            if(!can_be_imm12(offset)) {
                                auto offset_opr = make_vreg(vreg_count++);
                                emit_load_of_constant(offset_opr, offset, mb);
                                auto new_base_add = new MI_Binary(BINARY_ADD, make_vreg(vreg_count++), cur_arr_base_opr, offset_opr);
                                new_base_add->is_rv64 = true;
                                push(new_base_add, mb);
                                ldr->base = new_base_add->dst;
                                ldr->offset = make_imm(0);
                            }
                            else {
                                ldr->base = cur_arr_base_opr;
                                ldr->offset = make_imm(offset);
                            }
                            cur_arr_base=NULL;
                        }
                        // array with undetermined offset
                        else
                        {
                            cur_arr_base = my_arr_base[Read_I->from];
                            auto cur_arr_base_opr = make_operand(Read_I->from, mb);
                            ldr->base = cur_arr_base_opr;
                            ldr->reg = make_operand(Read_I->to, mb, true);
                            ldr->offset = make_imm(0);
                            value_map[Read_I->to] = ldr->reg;
                            cur_arr_base=NULL;
                        }
                        
                    }
                    else
                    {
                        int flag=0;
                        for(auto v:globalValues)
                        {
                            if(v==Read_I->from)
                            {
                                flag=1;
                                break;
                            }
                        }
                        if(stack_val_map.find(Read_I->from)!=stack_val_map.end())
                            flag=2;
                        // common variable
                        if(flag==0)
                        {
                            ldr->reg = make_vreg(vreg_count++);
                            value_map[Read_I->to] = ldr->reg;
                            ldr->base = make_operand(Read_I->from, mb);
                            emit_move(ldr->reg,ldr->base,mb);
                            break;
                        }
                        // global variable
                        else if (flag==1)
                        {
                            auto nvreg = make_vreg(vreg_count++);
                            emit_load_of_global_ref(func_asm,Read_I->from,nvreg,mb);
                            ldr->base = nvreg;
                            ldr->reg = make_operand(Read_I->to, mb);
                            value_map[Read_I->to] = ldr->reg;
                            if(int_val_map.count(ldr->base))
                                int_val_map[ldr->reg]=int_val_map[ldr->base];
                        }
                        // stack variable
                        else
                        {
                            ldr->reg = make_vreg(vreg_count++);
                            value_map[Read_I->to] = ldr->reg;
                            ldr->base = get_stack_val(Read_I->from,mb,func_asm);
                            if(int_val_map.count(ldr->base))
                                int_val_map[ldr->reg]=int_val_map[ldr->base];
                        }
                    }
					push((MI *)ldr, mb);
				} break;

                // store instruction
                // 1: array type, has const offset
                // 2: array type, no const offset
                // 3: reg, simple
                // 4: reg, global
                // 5: reg, store in stack
				case Store: {
					auto Write_I = dynamic_cast<StoreInstruction*>(I.get());
					auto str = new MI_Store;
                    // store float type
                    if(Write_I->value->type->isFloat())
                    {
                        auto vstr = new MI_VStore;
                        if(my_arr_base.find(Write_I->des)!=my_arr_base.end())
                        {
                            assert(Write_I->des->I && Write_I->des->I->type == GEP);
                            // local array with determined offset
                            if(ptr_val_map.find(Write_I->des)!=ptr_val_map.end())
                            {
                                vstr->reg = make_operand(Write_I->value,mb,true);
                                auto opr = ptr_val_map.find(Write_I->des);
                                int offset = opr->second * 4;
                                auto cur_arr_base_opr = make_operand(Write_I->des, mb);

                                if(!can_be_imm12(offset)) {
                                    auto offset_opr = make_vreg(vreg_count++);
                                    emit_load_of_constant(offset_opr, offset, mb);
                                    auto new_base_add = new MI_Binary(BINARY_ADD, make_vreg(vreg_count++), cur_arr_base_opr, offset_opr);
                                    new_base_add->is_rv64 = true;
                                    push(new_base_add, mb);
                                    vstr->base = new_base_add->dst;
                                    vstr->offset = make_imm(0);
                                }
                                else {
                                    vstr->base = cur_arr_base_opr;
                                    vstr->offset = make_imm(offset);
                                }
                                cur_arr_base=NULL;
                            }
                            // local array with undetermined offset
                            else
                            {
                                vstr->reg = make_operand(Write_I->value,mb,true);
                                auto cur_arr_base_opr = make_operand(Write_I->des, mb);
                                vstr->base = cur_arr_base_opr;
                                cur_arr_base=NULL;
                            }
                        }
                        else
                        {
                            int flag=0;
                            for(auto v:globalValues)
                            {
                                if(v==Write_I->des)
                                {
                                    flag=1;
                                    break;
                                }
                            }
                            if(stack_val_map.find(Write_I->des)!=stack_val_map.end())
                                flag=2;
                            // common variable
                            if(flag==0)
                            {
                                vstr->base = make_operand(Write_I->des,mb,true);
                                vstr->reg = make_operand(Write_I->value, mb,true);
                                emit_move(vstr->base,vstr->reg,mb);
                                break;
                            }
                            // global variable
                            else if (flag==1)
                            {
                                vstr->reg = make_operand(Write_I->value, mb,true);
                            
                                auto nvreg = make_vreg(vreg_count++);
                                emit_load_of_global_ref(func_asm,Write_I->des,nvreg,mb);
                                vstr->base = nvreg;
                                value_map[Write_I->des] = vstr->base;
                            }
                            // stack variable
                            else
                            {
                                vstr->base = get_stack_val(Write_I->des,mb,func_asm);
                                vstr->reg = make_operand(Write_I->value, mb,true);
                            }
                        }
					    push((MI *)vstr, mb);
                        break;
                    }

                    // store int type
                    if(my_arr_base.find(Write_I->des)!=my_arr_base.end())
                    {
                        assert(Write_I->des->I && Write_I->des->I->type == GEP);
                        // array with determined offset
                        if(ptr_val_map.find(Write_I->des)!=ptr_val_map.end())
                        {
                            str->reg = make_operand(Write_I->value,mb,true);
                            auto opr = ptr_val_map.find(Write_I->des);
                            int offset = opr->second * 4;
                            auto cur_arr_base_opr = make_operand(Write_I->des, mb);

                            if(!can_be_imm12(offset)) {
                                auto offset_opr = make_vreg(vreg_count++);
                                emit_load_of_constant(offset_opr, offset, mb);
                                auto new_base_add = new MI_Binary(BINARY_ADD, make_vreg(vreg_count++), cur_arr_base_opr, offset_opr);
                                new_base_add->is_rv64 = true;
                                push(new_base_add, mb);
                                str->base = new_base_add->dst;
                                str->offset = make_imm(0);
                            }
                            else {
                                str->base = cur_arr_base_opr;
                                str->offset = make_imm(offset);
                            }
                            cur_arr_base=NULL;
                        }
                        // array with undetermined offset
                        else
                        {
                            auto cur_arr_base_opr = make_operand(Write_I->des, mb);
                            str->base = cur_arr_base_opr;
                            str->reg = make_operand(Write_I->value,mb,true);
                            str->offset = make_imm(0);
                            cur_arr_base=NULL;
                        }
                    }
                    else
                    {
                        int flag=0;
                        for(auto v:globalValues)
                        {
                            if(v==Write_I->des)
                            {
                                flag=1;
                                break;
                            }
                        }
                        if(stack_val_map.find(Write_I->des)!=stack_val_map.end())
                            flag=2;
                        // common variable
                        if(flag==0)
                        {
                            str->base = make_operand(Write_I->des,mb,true);
                            str->reg = make_operand(Write_I->value, mb,true);
                            emit_move(str->base, str->reg, mb);
                            break;
                        }
                        // global vairable
                        else if (flag==1)
                        {
                            str->reg = make_operand(Write_I->value, mb,true);
                            auto nvreg = make_vreg(vreg_count++);
                            emit_load_of_global_ref(func_asm,Write_I->des,nvreg,mb);
                            str->base = nvreg;
                            value_map[Write_I->des] = str->base;
                            if(int_val_map.count(str->reg))
                                int_val_map[str->base] = int_val_map[str->reg];
                        }
                        // stack variable
                        else
                        {
                            str->base = get_stack_val(Write_I->des,mb,func_asm);
                            str->reg = make_operand(Write_I->value, mb,true);
                            if(int_val_map.count(str->reg))
                                int_val_map[str->base] = int_val_map[str->reg];
                        }
                    }
					push((MI *)str, mb);
				} break;
                case Phi:
                    break;
                case LoopGEP: {
                    // assert that GEP is defined in loop preheader
                    auto loopGEP = dynamic_cast<LoopGEPInstruction *>(I.get());
                    auto gep = loopGEP->gep;
                    int flag = 0;
                    bool isarrele = 0;
                    int matchedIdx = 0;
                    value_map[gep->reg] = make_vreg(vreg_count++);
                    if (my_arr_base.find(gep->from) == my_arr_base.end())
                        my_arr_base[gep->reg] = gep->from;
                    else
                        my_arr_base[gep->reg] = my_arr_base[gep->from];
                    for (auto x : globalValues) {
                        if (gep->from == x) {
                            flag = 1;
                            break;
                        }
                    }
                    if (flag) {
                        auto nvreg = make_vreg(vreg_count++);
                        emit_load_of_global_ref(func_asm, gep->from, nvreg, mb);
                        value_map[gep->from] = nvreg;
                    }

                    if (gep->reg->name.find("arrayinit") != gep->reg->name.npos) {
                        isarrele = 1;
                    }
                    TypePtr cur_arr = NULL;
                    auto reg4 = value_map[gep->from];
                    int cff = 0;
                    int offset = 0;
                    int ff = 0;
                    // has constant offset before
                    if (ptr_val_map.find(gep->from) != ptr_val_map.end())
                    {
                        // offset = ptr_val_map.find(gep->from)->second;
                        cff = 1;
                    }

                    // index number is 1
                    if (value_map.find(gep->from) != value_map.end() && gep->from->name.find("arrayinit") != gep->from->name.npos && !cff)
                    // if (value_map.find(gep->from) != value_map.end() && gep->from->name.find("arrayinit") != gep->from->name.npos)
                    {
                        ff = 1;
                        reg4 = value_map[gep->from];
                    }

                    const int max_dims = 40;
                    MOperand mul_regs[max_dims + 1];
                    vector<MI *> mul_instr(max_dims + 1, nullptr);

                    int cnt = 0;
                    for (auto x : gep->index)
                    {
                        if (!x->isConst)
                            ff = 1;
                    }
                    if (isarrele)
                    {
                        for (int i = 0; i < gep->index.size(); i++)
                        {
                            cnt++;
                            auto x = gep->index[i];
                            int const_val = 0;
                            if(x == loopGEP->index) {
                                matchedIdx = i;
                                x = Const::getConst(Type::getInt(), 0, "");
                            }

                            if (x->isConst)
                                const_val = dynamic_cast<Const *>(x.get())->intVal;
                            else
                            {
                                ff = 1;
                            }

                            int const_size = 1;
                            if (i == 0)
                            {
                                if(gep->from->type->ID == PtrID) {
                                    auto tmp = dynamic_cast<PtrType *>(gep->from->type.get());
                                    if (tmp->inner->isArr())
                                        const_size = dynamic_cast<ArrType *>(tmp->inner.get())->getSize();
                                }
                                else if(gep->from->type->ID == IntID || gep->from->type->ID == FLoatID) {
                                    const_size = 1;
                                }
                                else {
                                    assert(false && "can not handle array type");
                                }
                            }
                            if (i == 1)
                            {
                                assert(gep->from->type->ID == ArrID);
                                auto tmp = dynamic_cast<ArrType *>(gep->from->type.get());
                                if (tmp->inner->isArr())
                                    const_size = dynamic_cast<ArrType *>(tmp->inner.get())->getSize();
                                cur_arr = tmp->inner;
                            }
                            else if (i > 1)
                            {
                                assert(cur_arr->ID == ArrID);
                                auto nc = dynamic_cast<ArrType *>(cur_arr.get());
                                if (nc->inner->isArr())
                                    const_size = dynamic_cast<ArrType *>(nc->inner.get())->getSize();
                                cur_arr = nc->inner;
                            }
                            if (!ff)
                                offset += const_size * const_val;
                            else
                            {
                                if (x != gep->index.back() || const_size != 1)
                                {
                                    ValuePtr imptr = ValuePtr(new Int("asdf", false, false, const_size));
                                    auto bi_mul = new MI_Binary;
                                    bi_mul->op = BINARY_MULTIPLY;
                                    bi_mul->lhs = make_operand(x, mb, true);
                                    bi_mul->rhs = make_operand(imptr, mb, true);
                                    assert(cnt > 0 && cnt < max_dims);
                                    mul_regs[cnt] = make_vreg(vreg_count++);
                                    bi_mul->dst = mul_regs[cnt];
                                    // push((MI *)bi_mul, mb);
                                    mul_instr[cnt] = bi_mul;
                                }
                                else
                                {
                                    assert(cnt > 0 && cnt < max_dims);
                                    mul_regs[cnt] = make_operand(x, mb, true);
                                }
                            }
                        }
                    }
                    else
                    {
                        for (int i = 1; i < gep->index.size(); i++)
                        {
                            auto x = gep->index[i];
                            int const_val = 0;
                            cnt++;
                            if(x == loopGEP->index) {
                                matchedIdx = i;
                                x = Const::getConst(Type::getInt(), 0, "");
                            }

                            if (x->isConst)
                                const_val = dynamic_cast<Const *>(x.get())->intVal;
                            else
                            {
                                ff = 1;
                            }

                            int const_size = 1;
                            if (i == 1)
                            {
                                auto tmp = dynamic_cast<ArrType *>(gep->from->type.get());
                                if (tmp->inner->isArr())
                                    const_size = dynamic_cast<ArrType *>(tmp->inner.get())->getSize();
                                cur_arr = tmp->inner;
                            }
                            else
                            {
                                auto nc = dynamic_cast<ArrType *>(cur_arr.get());
                                if (nc->inner->isArr())
                                    const_size = dynamic_cast<ArrType *>(nc->inner.get())->getSize();
                                cur_arr = nc->inner;
                            }
                            if (!ff)
                                offset += const_size * const_val;
                            else
                            {

                                if (x != gep->index.back())
                                {
                                    ValuePtr imptr = ValuePtr(new Int("asdf", false, false, const_size));
                                    auto bi_mul = new MI_Binary;
                                    bi_mul->op = BINARY_MULTIPLY;
                                    bi_mul->lhs = make_operand(x, mb, true);
                                    bi_mul->rhs = make_operand(imptr, mb, true);
                                    assert(cnt > 0 && cnt < max_dims);
                                    mul_regs[cnt] = make_vreg(vreg_count++);
                                    bi_mul->dst = mul_regs[cnt];
                                    // push((MI *)bi_mul, mb);
                                    mul_instr[cnt] = bi_mul;
                                }
                                else
                                {
                                    assert(cnt > 0 && cnt < max_dims);
                                    mul_regs[cnt] = make_operand(x, mb, true);
                                }
                            }
                        }
                    }
                    if (!ff)
                    {
                        ptr_val_map[gep->reg] = offset;
                        if(ptr_val_map.count(gep->from) != 0) {
                            ptr_val_map[gep->reg] += ptr_val_map[gep->from];
                        }
                        value_map[gep->reg] = value_map[gep->from];
                        gep->isConstIdx = true;
                    }
                    // calculate accurate address for new base
                    else
                    {
                        if (cnt == 1)
                        {
                            // do nothing
                            value_map[gep->reg] = value_map[gep->from];
                        }
                        else if (cnt > 1) 
                        {
                            MI_Binary *bi_add;
                            assert(cnt < max_dims);
                            int i;
                            auto last_reg = mul_regs[1];
                            if(mul_instr[1] != nullptr && matchedIdx != 1) {
                                push(mul_instr[1], mb);
                            }
                            else if(matchedIdx == 1) {
                                last_reg = mul_regs[2];
                            }
                            for (i = 1; i < cnt; i ++) {
                                if(i == matchedIdx) {
                                    continue;
                                }
                                if(mul_instr[i + 1] != nullptr) {
                                    push(mul_instr[i + 1], mb);
                                }
                                MOperand dst = make_vreg(vreg_count++);
                                bi_add = new MI_Binary(BINARY_ADD, dst, last_reg, mul_regs[i + 1]);
                                bi_add->is_rv64 = true;
                                push((MI *)bi_add, mb);
                                last_reg = bi_add->dst;
                            }

                            // calculate pointer address
                            auto offset = last_reg;
                            auto lsl_offset = make_vreg(vreg_count++);
                            auto old_base = make_operand(gep->from, mb);
                            auto new_base = make_operand(gep->reg, mb, true);
                            // clear former constant offset
                            if(ptr_val_map.count(gep->from) != 0) {
                                auto old_offset = make_legal_imm(ptr_val_map[gep->from] * 4, mb);
                                auto tmp_base = make_vreg(vreg_count++);
                                bi_add = new MI_Binary(BINARY_ADD, tmp_base, old_base, old_offset);
                                bi_add->is_rv64 = true;
                                push(bi_add, mb);
                                old_base = tmp_base;
                            }
                            auto bi_sh2add = new MI_Binary(BINARY_SH2ADD, new_base, offset, old_base);
                            push(bi_sh2add, mb);
                        }
                    }

                    auto ptr = make_vreg(vreg_count++);
                    auto base = make_operand(gep->reg, mb, true);
                    int init_offset = loopGEP->getInitOffset();
                    if(ptr_val_map.count(gep->from)) {
                        init_offset += ptr_val_map[gep->from];
                    }
                    
                    init_offset = init_offset * 4;
                    auto init_offset_opr = make_imm(init_offset);
                    if(!can_be_imm12(init_offset)) {
                        init_offset_opr = make_vreg(vreg_count++);
                        emit_load_of_constant(init_offset_opr, init_offset, mb);
                    }
                    auto cal_new_base = new MI_Binary(BINARY_ADD, ptr, base, init_offset_opr);
                    cal_new_base->is_rv64 = true;
                    push(cal_new_base, mb);
                    value_map[gep->reg] = ptr;
                } break;
                default: {
                    assert(false && "translating unknown IR instructions to assembly");
                }break;
       
            
            }
        }
    }



    // cout<<"<<<<< PHI"<<endl;
    // @todo handle the phi instrutions
    ii=0;
    for(auto bb : mybb->basicBlocks) {
        auto mb = func_asm->mbs[ii];
        ii++;
        for(int i = 0; i < bb->instructions.size(); i++) {
            if (bb->instructions[i]->type == Phi) {
                auto phi = dynamic_cast<PhiInstruction *>(bb->instructions[i].get());
                if(phi->reg->type->isFloat())
                {
                    auto incoming = make_vsreg(vsreg_count++);
                    // auto resreg=make_vsreg(vsreg_count++);
                    // value_map[phi->reg]=resreg;

                    auto phi_as_operand = make_operand(phi->reg, mb);
                    auto mv = emit_vmove(phi_as_operand, incoming);
                    insert((MI *)mv, mb->inst);

                    // cout<<"loop2"<<endl;

                    for(int p = 0; p < phi->from.size(); p++) {
                        auto phi_opr = phi->from[p];
                        auto pred_bb = phi_opr.second;
                        int idx=0;
                        for(auto bb: mybb->basicBlocks)
                        {
                            if(bb==pred_bb)
                                break;
                            idx++;
                        }
                        auto pred_mb = func_asm->mbs[idx];

                        // if phi_opr is int, implicitly convert int -> float
                        auto mv = emit_vmove(incoming, make_operand(phi_opr.first, pred_mb));
                        if (pred_mb->control_transfer_inst) {
                            insert((MI *)mv, pred_mb->control_transfer_inst);
                        } else {
                            push((MI *)mv, pred_mb);
                        }
                    }
                    continue;
                }
                auto incoming = make_vreg(vreg_count++);

                auto phi_as_operand = make_operand(phi->reg, mb);
                auto mv = emit_move(phi_as_operand, incoming);
                insert((MI *)mv, mb->inst);
                // remove phinode in int_val_map
                int_val_map.erase(phi_as_operand);

                for(int p = 0; p < phi->from.size(); p++) {
                    auto phi_opr = phi->from[p];
                    auto pred_bb = phi_opr.second;
                    int idx=0;
                    for(auto bb: mybb->basicBlocks)
                    {
                        if(bb==pred_bb)
                            break;
                        idx++;
                    }
                    auto pred_mb = func_asm->mbs[idx];

                    // implicitly convert: float -> int
                    if (phi_opr.first->type->isFloat()) {
                        auto mv = new MI_VCvt(F32, S32);
                        mv->src = make_operand(phi_opr.first, pred_mb);
                        mv->dst = incoming;
                        if (pred_mb->control_transfer_inst) {
                            insert((MI *)mv, pred_mb->control_transfer_inst);
                        } else {
                            push((MI *)mv, pred_mb);
                        }
                    }
                    // int -> int
                    else {
                        auto mv = emit_move(incoming, make_operand(phi_opr.first, pred_mb));
                        if (pred_mb->control_transfer_inst) {
                            insert((MI *)mv, pred_mb->control_transfer_inst);
                        } else {
                            push((MI *)mv, pred_mb);
                        }
                    }
                }
            }
        }
    }
    map<MOperand, int> new_int_val_map;
    for(auto p: int_val_map) {
        if(p.first.tag != REG && p.first.tag != SREG)
            new_int_val_map.emplace(p.first, p.second);
    }
    func_asm->vreg_count = vreg_count;
    func_asm->vsreg_count = vsreg_count;
    func_asm->int_val_map = new_int_val_map;
    return func_asm;
}



void build_cvt_type(String_Builder* s, Vcvt_Type t){
    switch(t){
        case U32 : {
            s->append("wu");
        } break;
        case S32 : {
            s->append("w");
        } break;
        case F32 : {
            s->append("s");
        } break;
        default : {
            assert(false);
        }
    }
}

void print_operand(MOperand op) {
    String_Builder s;
    build_operand(&s, op);
    printf("%s", s.c_str());
}

void build_operand(String_Builder *s, MOperand op) {
    // cout<<op.tag<<endl;
    switch(op.tag) {
        
        case REG:   {
            if (op.value == fp) s->append("s0"); // s0 = fp
            else if (op.value == sp) s->append("sp");
            else if (op.value == ra) s->append("ra");
            else if (op.value == pc) s->append("pc");
            else if (op.value == zero) s->append("zero");
            else if (op.value == x3) s->append("gp");
            else if (op.value == x4) s->append("tp");
            // use alias
            else if (op.value >= x5 && op.value <= x7) s->append("t%d", op.value - x5);
            else if (op.value == x9) s->append("s1");
            else if (op.value >= x10 && op.value <= x17) s->append("a%d", op.value - x10);
            else if (op.value >= x18 && op.value <= x27) s->append("s%d", op.value - x18 + 2);
            else if (op.value >= x28 && op.value <= x31) s->append("t%d", op.value - x28 + 3);
            else {
                cout << "unknown register ID: " << op.value << endl;
                assert(false);
            }
        } break;
        case SREG:  {
            if (op.value >= f10 && op.value <= f17) s->append("fa%d", op.value - f10);
            else if (op.value >= f0 && op.value <= f7) s->append("ft%d", op.value);
            else if (op.value >= f8 && op.value <= f9) s->append("fs%d", op.value - f8);
            else if (op.value >= f18 && op.value <= f27) s->append("fs%d", op.value - f18 + 2);
            else if (op.value >= f28 && op.value <= f31) s->append("ft%d", op.value - f28 + 8);
            else {
                s->append("f%d\n", op.value);
            }
        } break;

        case VREG:  { s->append("vr%d", op.value); } break;
        case IMM:   { s->append("%d", op.value); } break;
		case ADR_GLOBAL: { s->append("%s", op.adr); } break;
        case VSREG: { s->append("vsr%d", op.value);} break;
        // case SIMM:  { cout<<"still error"<<endl;} break;
		case ERRORTYPE:
        default: { 
            printf("op.tag: %d\n", op.tag);
            assert(false && "unknown op tag");
        } break;

    }
	// switch(op.s_tag) {
    //     case Nothing: break;
    //     case LSL: { s->append(", lsl #%d", op.s_value); } break;
    //     case LSR: { s->append(", lsr #%d", op.s_value); } break;
    //     case ASL: { s->append(", asl #%d", op.s_value); } break;
    //     case ASR: { s->append(", asr #%d", op.s_value); } break;
    //     default: assert(false && "unknown op s_tag");
	// }
}

void print_function_asm(Func_Asm *func) {
    String_Builder s;
    build_function_asm(&s, func);
    printf("finish build_function_asm\n");
    printf("%s", s.c_str());
    fflush(stdout);
}

const char *get_branch_suffix(Branch_Condition condition) {
    switch(condition) {
        case NO_CONDITION:          { return ""; } break;
        case LESS_THAN:             { return "lt"; } break;
        case GREATER_THAN:          { return "gt"; } break;
        case LESS_THAN_OR_EQUAL:    { return "le"; } break;
        case GREATER_THAN_OR_EQUAL: { return "ge"; } break;
        case NOT_EQUAL:             { return "ne"; } break;
        case EQUAL:                 { return "eq"; } break;
        // case FLOAT:                 { return "pl"; } break;
        default: { assert(false); return ""; }
    }
}

Branch_Condition invert_branch_cond(Branch_Condition cond) {
    if (cond == NO_CONDITION || cond == FLOAT) return cond;
    if (cond >= 1 && cond <= 3) {
        return (Branch_Condition)(cond + 3);
    } else if (cond >= 4 && cond <= 6) {
        return (Branch_Condition)(cond - 3);
    }
    assert(false);
    return NO_CONDITION;
}

Branch_Condition invert_cmp_cond(Branch_Condition cond) {
    if (cond == NO_CONDITION || cond == FLOAT) return cond;
    if (cond == LESS_THAN) return GREATER_THAN;
    else if (cond == GREATER_THAN) return LESS_THAN;
    else if (cond == GREATER_THAN_OR_EQUAL) return LESS_THAN_OR_EQUAL;
    else if (cond == LESS_THAN_OR_EQUAL) return GREATER_THAN_OR_EQUAL;
    assert(false);
    return NO_CONDITION;
}

// 循环右移(RISC-V not supported)
bool can_be_imm_ror(int32 x) {
    uint32 v = x;
    for(int r = 0; r < 31; r += 2) {
        if ((v & 0xff) == v) {
            return true;
        }

        v = (v >> 2) | (v << 30);
    }
    return false;
}

void build_function_asm(String_Builder *s, Func_Asm *func, bool addSomeInstr) {
    auto ss=func->mbs.len;
	s->append("%s:\n", func->name.c_str());
    for(int i = 0; i < func->mbs.len; i++) {
        Machine_Block *next_bb = NULL;
        if (i != func->mbs.len-1) next_bb = func->mbs[i+1];
        s->append(".L%d_%d:\n", func->index, func->mbs[i]->i);
        for(auto I = func->mbs[i]->inst; I; I=I->next) {
            s->append("    ");
            const char *cond = get_branch_suffix(I->cond);
            const char *set_flags = I->update_flags ? "s" : "";
            switch(I->tag) {
                case MI_MOVE: {
                    auto mv = (MI_Move *)I;
                    if (mv->src.tag == IMM) {
                        s->append("li ");
                        build_operand(s, mv->dst);
                        s->append(",");
                        build_operand(s, mv->src);
                    } else {
                        s->append("mv ");
                        build_operand(s, mv->dst);
                        s->append(",");
                        build_operand(s, mv->src);
                    }
                } break;

                // case MI_CLZ: {
                //     auto clz = (MI_Clz *)I;
                //     s->append("clz%s%s ", set_flags, cond);
                //     build_operand(s, clz->dst);
                //     s->append(", ");
                //     build_operand(s, clz->operand);
                // } break;

                case MI_BINARY: {
                    auto bi = (MI_Binary *)I;
                    const char *instr_type = bi->is_rv64 ? "" : "w";
                    MOperand rhs = bi->rhs;

                    switch(bi->op) {
                        // @TODO: divide BINARY_ADD into BINARY_ADD_REG and BINARY_ADD_IMM
                        // @TODO: determine whether IMM can be imm_12
                        case BINARY_ADD:         { bi->rhs.tag == IMM ? s->append("addi") : s->append("add"); } break;
                        case BINARY_SUBTRACT:    { bi->rhs.tag == IMM ? rhs.value = -rhs.value, s->append("addi") : s->append("sub"); } break;
                        case BINARY_MULTIPLY:    { s->append("mul"); } break;
                        case BINARY_DIVIDE:      { s->append("div"); } break;
                        case BINARY_MOD:         { s->append("rem"); } break;
            			case BINARY_LSL:         { bi->rhs.tag == IMM ? s->append("slli") : s->append("sll"); } break;
            			case BINARY_LSR:         { bi->rhs.tag == IMM ? s->append("srli") : s->append("srl"); } break;
            			// case BINARY_ASL:         { s->append("asl");  } break;
            			case BINARY_ASR:         { bi->rhs.tag == IMM ? s->append("srai") : s->append("sra"); } break;
            			// case BINARY_RSB:         { s->append("rsb");  } break;
						case BINARY_BITWISE_AND: { bi->rhs.tag == IMM ? s->append("andi") : s->append("and"); }break;
						// case BINARY_BITWISE_OR:  { s->append("orr"); }break;
						// case BINARY_BIC:         { s->append("bic"); }break;
						// case BINARY_SMMUL:       { s->append("smmul"); }break;
                        case BINARY_XOR:         { bi->rhs.tag == IMM ? s->append("xori") : s->append("xor"); } break;
                        case BINARY_SH2ADD:      { s->append("sh2add"); } break;
                        default:                 { assert(false && "unknown binary asm instruction."); }
                    }
                    // arithmetic opeartion
                    if (bi->op >= BINARY_ADD && bi->op <= BINARY_MOD) {
                        s->append("%s", instr_type);
                    }
                    s->append(" ");
                    build_operand(s, bi->dst);
                    s->append(",");
                    build_operand(s, bi->lhs);
                    s->append(",");
                    build_operand(s, rhs);
                    if (bi->op == BINARY_ADD) s->append(" # BINARY_ADD");
                    else if (bi->op == BINARY_SUBTRACT) s->append(" # BINARY_SUBTRACT");
                } break;

                // case MI_COMPARE: {
                //     auto cmp = (MI_Compare *)I;

                //     if (cmp->neg) {
                //         s->append("cmn ");
                //     } else {
                //         s->append("cmp ");
                //     }
                //     build_operand(s, cmp->lhs);
                //     s->append(", ");
                //     build_operand(s, cmp->rhs);
                // } break;

                case MI_SLT: {
                    auto slt = (MI_Slt *)I;

                    switch(slt->slt_tag) {
                        case SLT:    { s->append("slt "); } break;
                        case SLT_U:  { s->append("sltu "); } break;
                        case SLT_I:  { s->append("slti "); } break;
                        case SLT_IU: { s->append("sltiu "); } break;
                        default:     { assert(false && "unknown MI_Slt tag"); }
                    }
                    build_operand(s, slt->dst);
                    s->append(",");
                    build_operand(s, slt->lhs);
                    s->append(",");
                    build_operand(s, slt->rhs);
                } break;

                case MI_BRANCH: {
                    auto br = (MI_Branch *)I;
                    if (next_bb && next_bb->condified) break;
                    if (br->cond == NO_CONDITION) {
                        if (br->true_target->i != i+1) {
                            s->append("j .L%d_%d", func->index, br->true_target->i);
                        }
                    }
                    else if (br->true_target->i == i+1) {
                        // invert the branch cond
                        s->append("b%s ", get_branch_suffix(invert_branch_cond(br->cond)));
                        build_operand(s, br->lhs);
                        s->append(",");
                        build_operand(s, br->rhs);
                        s->append(",.L%d_%d # jump to true\n", func->index, br->false_target->i);
                    } else if (br->false_target->i == i+1) {
                        s->append("b%s ", get_branch_suffix(br->cond));
                        build_operand(s, br->lhs);
                        s->append(",");
                        build_operand(s, br->rhs);
                        s->append(",.L%d_%d # jump to true\n", func->index, br->true_target->i);
                    } else {
                        s->append("b%s ", get_branch_suffix(br->cond));
                        build_operand(s, br->lhs);
                        s->append(",");
                        build_operand(s, br->rhs);
                        s->append(",.L%d_%d # jump to true\n", func->index, br->true_target->i);
                        s->append("    ");
                        s->append("j .L%d_%d # jump to false", func->index, br->false_target->i);
                    }
                } break;

                case MI_PUSH: {
                    auto push = (MI_Push *)I;
                    int num_reg = push->operands.size();
                    int reg_size = num_reg * 8;
                    // according to ARM, we first order the regs
                    std::sort(push->operands.begin(), push->operands.end());

                    s->append("addi sp,sp,%d # push\n", -reg_size);
                    // store highest-numbered reg into higher address
                    for (int i = num_reg - 1; i >= 0; i --) {
                        s->append("    sd ");
                        build_operand(s, push->operands[i]);
                        s->append(",%d(sp)", 8 * i);
                        if (i != 0) s->append("\n");
                    }
                } break;
                case MI_POP: {
                    auto pop = (MI_Push *)I;
                    int num_reg = pop->operands.size();
                    int reg_size = num_reg * 8;
                    // according to ARM, we first order the regs
                    std::sort(pop->operands.begin(), pop->operands.end());

                    for (int i = 0; i < num_reg; i ++) {
                        if (i > 0) s->append("    ");
                        s->append("ld ");
                        build_operand(s, pop->operands[i]);
                        s->append(",%d(sp)\n", 8 * i);
                    }
                    s->append("    addi sp,sp,%d # pop", reg_size);
                } break;

                case MI_FUNC_CALL: {
                    auto call = (MI_Func_Call *)I;
                    if(addSomeInstr && strcmp(call->func_name, "_sysy_stoptime") == 0) {
                        s->append("addi x0,x0,0 # useless instruction\n");
                        s->append("addi x0,x0,0 # useless instruction\n");
                        s->append("addi x0,x0,0 # useless instruction\n");
                        s->append("addi x0,x0,0 # useless instruction\n");
                        s->append("addi x0,x0,0 # useless instruction\n");
                        s->append("addi x0,x0,0 # useless instruction\n");
                        s->append("addi x0,x0,0 # useless instruction\n");
                        s->append("addi x0,x0,0 # useless instruction\n");
                        s->append("addi x0,x0,0 # useless instruction\n");
                        s->append("    ");
                    }
                    s->append("call %s", call->func_name);
                } break;

                case MI_LOAD:
                case MI_STORE: {
                    auto load_or_store = (MI_Load *)I;
                    const char *instr_type = load_or_store->is_rv64 ? "d" : "w";
					if (load_or_store->base.tag == ADR_GLOBAL) {
                        s->append("la ");
                        build_operand(s, load_or_store->reg);
                        s->append(",");
                        build_operand(s, load_or_store->base);
                        s->append(" # load global address");
					}
					else if (load_or_store->base.tag == IMM) {
						auto val = load_or_store->base.value;
                        s->append("li ");
                        build_operand(s, load_or_store->reg);
                        s->append(",");
                        build_operand(s, load_or_store->base);
					}
					else {
						s->append(I->tag == MI_LOAD ? "l" : "s");
                        s->append("%s ", instr_type);
						build_operand(s, load_or_store->reg);
						s->append(",");
                        if (load_or_store->offset.tag != ERRORTYPE) {
							build_operand(s, load_or_store->offset);
						}
                        s->append("(");
						build_operand(s, load_or_store->base);
						s->append(")");
					}
                } break;

                case MI_RETURN: { 
                    s->append("jr ra");
                } break;

                case MI_VMOVE : {
                    auto vmv = (MI_VMove *)I;
                    if (vmv->from_s32) {
                        s->append("fmv.w.x ");
                    }
                    else {
                        s->append("fmv.s ");
                    }
					build_operand(s, vmv->dst);
					s->append(",");
					build_operand(s, vmv->src);
                } break;

                case MI_VCVT : {
                    auto cvt = (MI_VCvt *)I;
                    s->append("fcvt.");
                    build_cvt_type(s, cvt->to_type);
                    s->append(".");
                    build_cvt_type(s, cvt->from_type);
                    s->append(" ");
                    build_operand(s, cvt->dst);
                    s->append(",");
                    build_operand(s, cvt->src);

                    // round to zero
                    if (cvt->from_type == F32) {
                        s->append(",rtz");
                    }
                } break;

                case MI_VLOAD : 
                case MI_VSTORE : {
                    auto load_or_store = (MI_VLoad *)I;
					if (load_or_store->base.tag == ADR_GLOBAL) {
                        s->append("la ");
                        build_operand(s, load_or_store->reg);
                        s->append(",");
                        build_operand(s, load_or_store->base);
                        s->append(" # load float global address");
					}
					else if (load_or_store->base.tag == SIMM) {
                        assert(false && "not support for loading SIMM");
					}
					else {
						s->append(I->tag == MI_VLOAD ? "fl" : "fs");
                        s->append(load_or_store->is_rv64 ? "d " : "w ");
						build_operand(s, load_or_store->reg);
						s->append(",");
                        if (load_or_store->offset.tag != ERRORTYPE) {
							build_operand(s, load_or_store->offset);
						}
                        s->append("(");
						build_operand(s, load_or_store->base);
						s->append(")");
					}
                } break;

                case MI_VPUSH: {
                    auto push = (MI_VPush *)I;
                    int num_reg = push->operands.size();
                    int reg_size = num_reg * 8;
                    // according to ARM, we first order the regs
                    std::sort(push->operands.begin(), push->operands.end());

                    s->append("addi sp,sp,%d # push\n", -reg_size);
                    // store highest-numbered reg into higher address
                    for (int i = num_reg - 1; i >= 0; i --) {
                        s->append("    fsd ");
                        build_operand(s, push->operands[i]);
                        s->append(",%d(sp)", 8 * i);
                        if (i != 0) s->append("\n");
                    }
                } break;
                case MI_VPOP: {
                    auto pop = (MI_VPop *)I;
                    int num_reg = pop->operands.size();
                    int reg_size = num_reg * 8;
                    // according to ARM, we first order the regs
                    std::sort(pop->operands.begin(), pop->operands.end());

                    for (int i = 0; i < num_reg; i ++) {
                        if (i > 0) s->append("    ");
                        s->append("fld ");
                        build_operand(s, pop->operands[i]);
                        s->append(",%d(sp)\n", 8 * i);
                    }
                    s->append("    addi sp,sp,%d # pop", reg_size);
                } break;
                
                case MI_VBINARY : {
                    auto bi = (MI_VBinary *)I;

                    switch(bi->op) {
                        case BINARY_ADD:         { s->append("fadd.s"); } break;
                        case BINARY_SUBTRACT:    { s->append("fsub.s"); } break;
                        case BINARY_MULTIPLY:    { s->append("fmul.s"); } break;
                        case BINARY_DIVIDE:      { s->append("fdiv.s"); } break;
                        case F_NEG:              { s->append("fneg.s"); } break;
                        default:                 { assert(false && "unknown binary asm instruction."); }
                    }
                    s->append(" ");
                    build_operand(s, bi->dst);
                    s->append(",");
                    build_operand(s, bi->lhs);
                    if(bi->fneg==1)
                    {
                        break;
                    }
                    s->append(",");
                    build_operand(s, bi->rhs);
                } break;

                // // todo : fix it for case 95
                case MI_VCMP_SET : {
                    auto Vcmp = (MI_VCmp_Set *)I;
                    MOperand lhs, rhs;
                    assert(Vcmp->dst.tag == REG || Vcmp->dst.tag == VREG);

                    // feq, flt, fle
                    if (is_fcmp_cond(Vcmp->cond)) {
                        lhs = Vcmp->lhs;
                        rhs = Vcmp->rhs;
                        s->append("f%s.s ", get_branch_suffix(Vcmp->cond));
                    }
                    // fne
                    else if (Vcmp->cond == NOT_EQUAL) {
                        s->append("feq.s ");
                        build_operand(s, Vcmp->dst);
                        s->append(",");
                        build_operand(s, Vcmp->lhs);
                        s->append(",");
                        build_operand(s, Vcmp->rhs);
                        s->append("\n    ");

                        s->append("xori ");
                        lhs = Vcmp->dst;
                        rhs = make_imm(1);
                    }
                    else {
                        lhs = Vcmp->rhs;
                        rhs = Vcmp->lhs;
                        s->append("f%s.s ", get_branch_suffix(invert_cmp_cond(Vcmp->cond)));
                    }

                    build_operand(s, Vcmp->dst);
                    s->append(",");
                    build_operand(s, lhs);
                    s->append(",");
                    build_operand(s, rhs);
                } break;
                default:
                {
                    cout << "unknow assembly instruction ID: " << I->tag << endl;
                    assert(false && "printing unknown assembly instruction.");
                }break;
                // default: { exit(54); assert(false && "printing unknown assembly instruction."); } break;
            }
            s->append("\n");
        }
    }
}

void print_instr_asm(MI *I) {
    String_Builder s;
    build_instr_asm(&s, I);
    printf("%s", s.c_str());
    fflush(stdout);
}

void build_instr_asm(String_Builder *s, MI *I) {
    const char *cond = get_branch_suffix(I->cond);
    const char *set_flags = I->update_flags ? "s" : "";
    switch(I->tag) {
        case MI_MOVE: {
            auto mv = (MI_Move *)I;
            if (mv->src.tag == IMM) {
                s->append("li ");
                build_operand(s, mv->dst);
                s->append(",");
                build_operand(s, mv->src);
            } else {
                s->append("mv ");
                build_operand(s, mv->dst);
                s->append(",");
                build_operand(s, mv->src);
            }
        } break;

        // case MI_CLZ: {
        //     auto clz = (MI_Clz *)I;
        //     s->append("clz%s%s ", set_flags, cond);
        //     build_operand(s, clz->dst);
        //     s->append(", ");
        //     build_operand(s, clz->operand);
        // } break;

        case MI_BINARY: {
            auto bi = (MI_Binary *)I;
            const char *instr_type = bi->is_rv64 ? "" : "w";
            MOperand rhs = bi->rhs;

            switch(bi->op) {
                // @TODO: divide BINARY_ADD into BINARY_ADD_REG and BINARY_ADD_IMM
                // @TODO: determine whether IMM can be imm_12
                case BINARY_ADD:         { bi->rhs.tag == IMM ? s->append("addi") : s->append("add"); } break;
                case BINARY_SUBTRACT:    { bi->rhs.tag == IMM ? rhs.value = -rhs.value, s->append("addi") : s->append("sub"); } break;
                case BINARY_MULTIPLY:    { s->append("mul"); } break;
                case BINARY_DIVIDE:      { s->append("div"); } break;
                case BINARY_MOD:         { s->append("rem"); } break;
                case BINARY_LSL:         { bi->rhs.tag == IMM ? s->append("slli") : s->append("sll"); } break;
                case BINARY_LSR:         { bi->rhs.tag == IMM ? s->append("srli") : s->append("srl"); } break;
                // case BINARY_ASL:         { s->append("asl");  } break;
                case BINARY_ASR:         { bi->rhs.tag == IMM ? s->append("srai") : s->append("sra"); } break;
                // case BINARY_RSB:         { s->append("rsb");  } break;
                case BINARY_BITWISE_AND: { bi->rhs.tag == IMM ? s->append("andi") : s->append("and"); }break;
                // case BINARY_BITWISE_OR:  { s->append("orr"); }break;
                // case BINARY_BIC:         { s->append("bic"); }break;
                // case BINARY_SMMUL:       { s->append("smmul"); }break;
                case BINARY_XOR:         { bi->rhs.tag == IMM ? s->append("xori") : s->append("xor"); } break;
                case BINARY_SH2ADD:      { s->append("sh2add"); } break;
                default:                 { assert(false && "unknown binary asm instruction."); }
            }
            // arithmetic opeartion
            if (bi->op >= BINARY_ADD && bi->op <= BINARY_MOD) {
                s->append("%s", instr_type);
            }
            s->append(" ");
            build_operand(s, bi->dst);
            s->append(",");
            build_operand(s, bi->lhs);
            s->append(",");
            build_operand(s, rhs);
            if (bi->op == BINARY_ADD) s->append(" # BINARY_ADD");
            else if (bi->op == BINARY_SUBTRACT) s->append(" # BINARY_SUBTRACT");
        } break;

        // case MI_COMPARE: {
        //     auto cmp = (MI_Compare *)I;

        //     if (cmp->neg) {
        //         s->append("cmn ");
        //     } else {
        //         s->append("cmp ");
        //     }
        //     build_operand(s, cmp->lhs);
        //     s->append(", ");
        //     build_operand(s, cmp->rhs);
        // } break;

        case MI_SLT: {
            auto slt = (MI_Slt *)I;

            switch(slt->slt_tag) {
                case SLT:    { s->append("slt "); } break;
                case SLT_U:  { s->append("sltu "); } break;
                case SLT_I:  { s->append("slti "); } break;
                case SLT_IU: { s->append("sltiu "); } break;
                default:     { assert(false && "unknown MI_Slt tag"); }
            }
            build_operand(s, slt->dst);
            s->append(",");
            build_operand(s, slt->lhs);
            s->append(",");
            build_operand(s, slt->rhs);
        } break;

        case MI_BRANCH: {
            s->append("an branch instruction");
        } break;

        case MI_PUSH: {
            auto push = (MI_Push *)I;
            int num_reg = push->operands.size();
            int reg_size = num_reg * 8;
            // according to ARM, we first order the regs
            std::sort(push->operands.begin(), push->operands.end());

            s->append("addi sp,sp,%d # push\n", -reg_size);
            // store highest-numbered reg into higher address
            for (int i = num_reg - 1; i >= 0; i --) {
                s->append("    sd ");
                build_operand(s, push->operands[i]);
                s->append(",%d(sp)", 8 * i);
                if (i != 0) s->append("\n");
            }
        } break;
        case MI_POP: {
            auto pop = (MI_Push *)I;
            int num_reg = pop->operands.size();
            int reg_size = num_reg * 8;
            // according to ARM, we first order the regs
            std::sort(pop->operands.begin(), pop->operands.end());

            for (int i = 0; i < num_reg; i ++) {
                if (i > 0) s->append("    ");
                s->append("ld ");
                build_operand(s, pop->operands[i]);
                s->append(",%d(sp)\n", 8 * i);
            }
            s->append("    addi sp,sp,%d # pop", reg_size);
        } break;

        case MI_FUNC_CALL: {
            auto call = (MI_Func_Call *)I;
            s->append("call %s", call->func_name);
        } break;

        case MI_LOAD:
        case MI_STORE: {
            auto load_or_store = (MI_Load *)I;
            const char *instr_type = load_or_store->is_rv64 ? "d" : "w";
            if (load_or_store->base.tag == ADR_GLOBAL) {
                s->append("la ");
                build_operand(s, load_or_store->reg);
                s->append(",");
                build_operand(s, load_or_store->base);
                s->append(" # load global address");
            }
            else if (load_or_store->base.tag == IMM) {
                auto val = load_or_store->base.value;
                s->append("li ");
                build_operand(s, load_or_store->reg);
                s->append(",");
                build_operand(s, load_or_store->base);
            }
            else {
                s->append(I->tag == MI_LOAD ? "l" : "s");
                s->append("%s ", instr_type);
                build_operand(s, load_or_store->reg);
                s->append(",");
                if (load_or_store->offset.tag != ERRORTYPE) {
                    build_operand(s, load_or_store->offset);
                }
                s->append("(");
                build_operand(s, load_or_store->base);
                s->append(")");
            }
        } break;

        case MI_RETURN: { 
            s->append("jr ra");
        } break;

        case MI_VMOVE : {
            auto vmv = (MI_VMove *)I;
            if (vmv->from_s32) {
                s->append("fmv.w.x ");
            }
            else {
                s->append("fmv.s ");
            }
            build_operand(s, vmv->dst);
            s->append(",");
            build_operand(s, vmv->src);
        } break;

        case MI_VCVT : {
            auto cvt = (MI_VCvt *)I;
            s->append("fcvt.");
            build_cvt_type(s, cvt->to_type);
            s->append(".");
            build_cvt_type(s, cvt->from_type);
            s->append(" ");
            build_operand(s, cvt->dst);
            s->append(",");
            build_operand(s, cvt->src);

            // round to zero
            if (cvt->from_type == F32) {
                s->append(",rtz");
            }
        } break;

        case MI_VLOAD : 
        case MI_VSTORE : {
            auto load_or_store = (MI_VLoad *)I;
            if (load_or_store->base.tag == ADR_GLOBAL) {
                s->append("la ");
                build_operand(s, load_or_store->reg);
                s->append(",");
                build_operand(s, load_or_store->base);
                s->append(" # load float global address");
            }
            else if (load_or_store->base.tag == SIMM) {
                assert(false && "not support for loading SIMM");
            }
            else {
                s->append(I->tag == MI_VLOAD ? "fl" : "fs");
                s->append(load_or_store->is_rv64 ? "d " : "w ");
                build_operand(s, load_or_store->reg);
                s->append(",");
                if (load_or_store->offset.tag != ERRORTYPE) {
                    build_operand(s, load_or_store->offset);
                }
                s->append("(");
                build_operand(s, load_or_store->base);
                s->append(")");
            }
        } break;

        case MI_VPUSH: {
            auto push = (MI_VPush *)I;
            int num_reg = push->operands.size();
            int reg_size = num_reg * 8;
            // according to ARM, we first order the regs
            std::sort(push->operands.begin(), push->operands.end());

            s->append("addi sp,sp,%d # push\n", -reg_size);
            // store highest-numbered reg into higher address
            for (int i = num_reg - 1; i >= 0; i --) {
                s->append("    fsd ");
                build_operand(s, push->operands[i]);
                s->append(",%d(sp)", 8 * i);
                if (i != 0) s->append("\n");
            }
        } break;
        case MI_VPOP: {
            auto pop = (MI_VPop *)I;
            int num_reg = pop->operands.size();
            int reg_size = num_reg * 8;
            // according to ARM, we first order the regs
            std::sort(pop->operands.begin(), pop->operands.end());

            for (int i = 0; i < num_reg; i ++) {
                if (i > 0) s->append("    ");
                s->append("fld ");
                build_operand(s, pop->operands[i]);
                s->append(",%d(sp)\n", 8 * i);
            }
            s->append("    addi sp,sp,%d # pop", reg_size);
        } break;
        
        case MI_VBINARY : {
            auto bi = (MI_VBinary *)I;

            switch(bi->op) {
                case BINARY_ADD:         { s->append("fadd.s"); } break;
                case BINARY_SUBTRACT:    { s->append("fsub.s"); } break;
                case BINARY_MULTIPLY:    { s->append("fmul.s"); } break;
                case BINARY_DIVIDE:      { s->append("fdiv.s"); } break;
                case F_NEG:              { s->append("fneg.s"); } break;
                default:                 { exit(53); assert(false && "unknown binary asm instruction."); }
            }
            s->append(" ");
            build_operand(s, bi->dst);
            s->append(",");
            build_operand(s, bi->lhs);
            if(bi->fneg==1)
            {
                break;
            }
            s->append(",");
            build_operand(s, bi->rhs);
        } break;

        // // todo : fix it for case 95
        case MI_VCMP_SET : {
            auto Vcmp = (MI_VCmp_Set *)I;
            MOperand lhs, rhs;
            assert(Vcmp->dst.tag == REG || Vcmp->dst.tag == VREG);

            // feq, flt, fle
            if (is_fcmp_cond(Vcmp->cond)) {
                lhs = Vcmp->lhs;
                rhs = Vcmp->rhs;
                s->append("f%s.s ", get_branch_suffix(Vcmp->cond));
            }
            // fne
            else if (Vcmp->cond == NOT_EQUAL) {
                s->append("feq.s ");
                build_operand(s, Vcmp->dst);
                s->append(",");
                build_operand(s, Vcmp->lhs);
                s->append(",");
                build_operand(s, Vcmp->rhs);
                s->append("\n    ");

                s->append("xori ");
                lhs = Vcmp->dst;
                rhs = make_imm(1);
            }
            else {
                lhs = Vcmp->rhs;
                rhs = Vcmp->lhs;
                s->append("f%s.s ", get_branch_suffix(invert_cmp_cond(Vcmp->cond)));
            }

            build_operand(s, Vcmp->dst);
            s->append(",");
            build_operand(s, lhs);
            s->append(",");
            build_operand(s, rhs);
        } break;
        default:
        {
            cout << "unknow assembly instruction ID: " << I->tag << endl;
            assert(false && "printing unknown assembly instruction.");
        }break;
        // default: { exit(54); assert(false && "printing unknown assembly instruction."); } break;
    }
    s->append("\n");
}

void get_init_sequence(VariablePtr globalarr, vector<Pair<string, int> > &init_inst){
    if(globalarr->type->isInt()){
        int val_use = dynamic_cast<Int *>(globalarr.get())->intVal;
        if(val_use){
            init_inst.emplace_back(make_pair(".word",val_use));
        }else{
            if(init_inst.empty()||init_inst.back().first==".word"){
                init_inst.emplace_back(make_pair(".zero",0));
            }
            init_inst.back().second+=4;
        }
        return ;
    }else if(globalarr->type->isFloat()){
        f2i.f_value = dynamic_cast<Float *>(globalarr.get())->floatVal;
        init_inst.emplace_back(make_pair(".word",(f2i.bin_value)));
        return ;
    }
    if(globalarr->zero()){
        if(init_inst.empty()||init_inst.back().first==".word"){
            init_inst.emplace_back(make_pair(".zero",0));
        }
        int zero_size = dynamic_cast<ArrType*>((globalarr->type).get())->getSize();
        init_inst.back().second += zero_size*4;
        return ;
    }else{
        auto arr_use = dynamic_cast<Arr*>(globalarr.get());
        int i = 0;
        for(; i < arr_use->inner.size(); i++){
            get_init_sequence(arr_use->inner[i],init_inst);
        }
        if(arr_use->getElementType()->isArr()){
            for(;i<arr_use->getElementLength();i++){
                get_init_sequence(arr_use->inner[i], init_inst);
            }
        }else if(arr_use->getElementType()->isInt()){
            if(i != arr_use->getElementLength()){
                if(init_inst.empty()||init_inst.back().first==".word"){
                    init_inst.emplace_back(make_pair(".zero",0));
                }
                init_inst.back().second += (arr_use->getElementLength() - i) * 4;
            }
        }else if(arr_use->getElementType()->isFloat()){
            if(i != arr_use->getElementLength()){
                if(init_inst.empty()||init_inst.back().first==".word"){
                    init_inst.emplace_back(make_pair(".zero",0));
                }
                init_inst.back().second += (arr_use->getElementLength() - i) * 4;
            }
        }
    }
}

void build_globals(String_Builder *s, vector<VariablePtr> & globals) {
    if(globals.size() == 0) return;
    s->append(".data\n.align 2\n");
    for (int i = 0; i < globals.size(); i++){
        if(globals[i]->type->isInt()){
            s->append("%s:\n", globals[i]->name.c_str());
            s->append("    ");
            int32 val = dynamic_cast<Int*>(globals[i].get())->intVal;
            s->append(".word %d\n", val);
        }else if(globals[i]->type->isFloat()){
            s->append("%s:\n", globals[i]->name.c_str());
            s->append("    ");
            f2i.f_value = dynamic_cast<Float*>(globals[i].get())->floatVal;
            s->append(".word %d\n", f2i.bin_value);
        }else if(globals[i]->type->isArr()){
            // int complete?
            vector<Pair<string, int> > init_inst;
            auto globalarr = dynamic_cast<Arr*>(globals[i].get());
            if(globalarr->zero()){
                continue;
            }else{
                get_init_sequence(globals[i],init_inst);
                s->append("%s:\n", globals[i]->name.c_str());
                for(int j = 0; j < init_inst.size(); j++){
                    s->append("    ");
                    s->append("%s  %d\n", init_inst[j].first.c_str(), init_inst[j].second);
                }
            }
        }else{
            fprintf(stderr, "globals error type\n");
        }
    }
    s->append(".bss\n");
    s->append(".align 2\n");
    for (int i = 0; i < globals.size(); i++){
        if(globals[i]->type->isArr()){
            auto globalarr = dynamic_cast<Arr*>(globals[i].get());
            if(globalarr->zero()){
                int zero_size = dynamic_cast<ArrType*>((globalarr->type).get())->getSize();
                zero_size *=4;
                s->append("    .type %s, @object\n", globals[i]->name.c_str());
                s->append("    .size %s, %d\n", globals[i]->name.c_str(), zero_size);
                s->append("%s:\n", globals[i]->name.c_str());
                // s->append("    .space %d\n", zero_size);
                // s->append("    .type %s, @object\n", globals[i]->name.c_str());
                // s->append("    .size %s, %d\n", globals[i]->name.c_str(), zero_size);
                s->append("    .zero %d\n", zero_size);
            }
        }
    }
    return;
}

void print_program_asm(Program_Asm *pro, vector<VariablePtr> & globalVariables) {
    String_Builder s;
    build_program_asm(&s, pro, globalVariables, false);
    printf("%s", s.c_str());
}

void build_program_asm(String_Builder *s, Program_Asm *pro, vector<VariablePtr> &globalVariables, bool addSomeInstr) {
	s->append("    .option nopic\n");
    s->append("    .attribute arch, \"rv64i2p1_m2p0_f2p2_d2p2_zba1p0_zbb1p0\"\n");
    s->append("    .attribute unaligned_access, 0\n");
    s->append("    .attribute stack_align, 16\n");

    build_globals(s, globalVariables);
	s->append("\n");

	s->append("    .text\n");
	s->append("    .align 1\n");

	s->append("    .globl main\n");
    s->append("    .type main, @function\n");
	for (int i = 0; i < pro->functions.len; i++) {
		build_function_asm(s, pro->functions[i], addSomeInstr);
        // cout<<"nono"<<endl;
        s->append("\n\n");
	}
    // cout<<"nono"<<endl;
}



Program_Asm *emit_asm(Module program_module) {
    auto program_asm = new Program_Asm;

    // add some map
    Binary_ir2asm["+"] = BINARY_ADD;
	Binary_ir2asm["-"] = BINARY_SUBTRACT;
	Binary_ir2asm["*"] = BINARY_MULTIPLY;
	Binary_ir2asm["/"] = BINARY_DIVIDE;
	Binary_ir2asm["%"] = BINARY_MOD;
    Binary_ir2asm["!="] = BINARY_NOT_EQUAL_TO;
    Binary_ir2asm["!"] = UNARY_XOR;
	Binary_ir2asm[">"] = BINARY_GREATER_THAN;
	Binary_ir2asm[">="] = BINARY_GREATER_THAN_OR_EQUAL_TO;
	Binary_ir2asm["<"] = BINARY_LESS_THAN;
	Binary_ir2asm["<="] = BINARY_LESS_THAN_OR_EQUAL_TO;
    Binary_ir2asm["=="] = BINARY_EQUAL_TO;
    Binary_ir2asm[","] = BINARY_LSL;
    Binary_ir2asm["."] = BINARY_ASR;
    // do only one for now
    //program_asm->functions.push(emit_function_asm(program_IR->procedures[0], 0));
    int i = 0;
    for (auto func : program_module.globalFunctions)
    {
        if(func->isLib)
            continue;
        i++;
        auto func_asm = emit_function_asm(func,i,program_module, program_module.globalVariables);
        program_asm->functions.push(func_asm);
    }
    // print_program_asm(program_asm, program_module.globalVariables);
    

    return program_asm;
}


bool is_callee_save(uint8 reg) {
    return reg == x9 || (reg >= x18 && reg <= x27) || reg == fp || reg == sp || reg == ra;
}

bool is_caller_save(uint8 reg) {
    return (reg >= x5 && reg <= x7) || (reg >= x10 && reg <= x17) || (reg >= x28 && reg <= x31);
}

bool s_is_callee_save(uint8 reg) {
    return (reg >= f8 && reg <= f9) || (reg >= f18 && reg <= f27);
}

bool s_is_caller_save(uint8 sreg) {
    return (sreg >= f0 && sreg <= f7) || (sreg >= f10 && sreg <= f17) || (sreg >= f28 && sreg <= f31);
}

// different from ARM(unsigned)
bool can_be_imm12(int32 x) {
    return (x >= -2048) && (x <= 2047);
}

bool is_fcmp_cond(Branch_Condition cond) {
    return cond == LESS_THAN || cond == LESS_THAN_OR_EQUAL || cond == EQUAL;
}

bool is_float_moperand(MOperand opr) {
    return opr.tag == SREG || opr.tag == VSREG || opr.tag == SIMM;
}

bool is_float_reg(MOperand opr) {
    return opr.tag == SREG || opr.tag == VSREG;
}

void MI::erase_from_parent() {
    if(mb != NULL) {
        if(mb->inst == this) {
            mb->inst = next;
        }
        if(mb->last_inst == this) {
            mb->last_inst = prev;
        }
        if(mb->control_transfer_inst == this) {
            mb->control_transfer_inst = NULL;
        }
        mb = NULL;
    }
    if(prev) {
        prev->next = next;
    }
    if(next) { 
        next->prev = prev;
    }
    next = prev = NULL;
}

void Machine_Block::erase_marked_values() {
    static Array<MI *> unmarked_values;
    unmarked_values.len = 0;

    for(MI *I = inst; I; I=I->next) {
        if (!I->marked) {
            unmarked_values.push(I);
        }
    }

    if (unmarked_values.len == 0) {
        inst = last_inst = control_transfer_inst = 0;
        succs = {};
    } else {
        inst = unmarked_values[0];
        last_inst = control_transfer_inst = unmarked_values[unmarked_values.len-1];


        inst->prev = NULL;
        last_inst->next = NULL;

        unmarked_values.push(NULL);
        for(int i = 0; i < unmarked_values.len-1; i++) {
            auto p = unmarked_values[i];
            auto n = unmarked_values[i+1];
            p->next = n;
            if (n) n->prev = p;
        }
    }
}

// functions of use

void MOperand::set_use_head(MI_Use *use, Func_Asm *func) {
    func->use_head_map[*this] = use;
}

void MOperand::s_set_use_head(MI_Use *use, Func_Asm *func) {
    func->s_use_head_map[*this] = use;
}

MI_Use *MOperand::get_use_head(Func_Asm *func) {
    if(tag != VREG && tag != REG) return nullptr;
    return func->use_head_map[*this];
}

MI_Use *MOperand::s_get_use_head(Func_Asm *func) {
    if(tag != VSREG && tag != SREG) return nullptr;
    return func->s_use_head_map[*this];
}

void MOperand::add_use(MI_Use *u, Func_Asm *func) {
    u->prev = nullptr;
    auto use_head = get_use_head(func);
    if(use_head){
        use_head->prev = u;
    }
    u->next = use_head;
    set_use_head(u, func);
}

void MOperand::s_add_use(MI_Use *u, Func_Asm *func) {
    u->prev = nullptr;
    auto s_use_head = s_get_use_head(func);
    if(s_use_head){
        s_use_head->prev = u;
    }
    u->next = s_use_head;
    s_set_use_head(u, func);
}

MI *MOperand::get_def_I(Func_Asm *func) {
    return func->def_I_map[*this];
}

void MOperand::set_def_I(MI *I, Func_Asm *func) {
    func->def_I_map[*this] = I;
}

void MI_Use::rm_use(Func_Asm *func) {
    if (prev == nullptr && next == nullptr) {
        val->set_use_head(nullptr, func);
    } 
    else if (prev == nullptr && next != nullptr) {
        val->set_use_head(next, func);
        next->prev = nullptr;
    } 
    else if (prev != nullptr && next == nullptr) {
        prev->next = nullptr;
    } 
    else if (prev != nullptr && next != nullptr) {
        prev->next = next;
        next->prev = prev;
    }
}

bool rm_instruction_use(MI* I, MOperand *v, Func_Asm *func) {
    auto use = v->get_use_head(func);
    while(use != nullptr){
        if(use->I == I){
            use->rm_use(func);
            return true;
        }
        use = use->next;
    }
    return false;
}

bool s_rm_instruction_use(MI* I, MOperand *v, Func_Asm *func) {
    auto use = v->s_get_use_head(func);
    while(use != nullptr){
        if(use->I == I){
            use->rm_use(func);
            return true;
        }
        use = use->next;
    }
    return false;
}

void Func_Asm::clear_use_head_map() {
    use_head_map.clear();
    s_use_head_map.clear();
}

void Func_Asm::clear_def_I_map() {
    def_I_map.clear();
}
