#include "asm_passes.h"
#include "../arm.h"

bool is_exp(int n) {
    return !(n & (n - 1));
}

std::pair<int, int> magic(int d)
{
    int p;
    unsigned ad, anc,delta, q1,r1,q2,r2,t;
    const unsigned two31 = 0x80000000;
    ad = abs(d);
    t = two31 + ((unsigned) d >> 31);
    anc = t -1 -t%ad;
    p=31;
    q1 = two31 /anc;
    r1 = two31 -q1 * anc;
    q2 = two31 /ad;
    r2 = two31 -q2 * ad;
    do
    {
        p =p +1;
        q1 = 2*q1;
        r1 = 2 * r1;
        if(r1>=anc)
        {
            q1 = q1 + 1;
            r1 = r1 - anc;
        }
        q2 = 2* q2;
        r2 = 2 * r2;
        if(r2>=ad)
        {
            q2 = q2 + 1;
            r2 = r2 - ad;
        }
        delta = ad - r2;
    } while (q1 < delta || (q1 == delta && q1 == 0));

    int M = q2 + 1;
    if(d<0) M = -M;
    int s = p - 32;
    return {M, s};
}

void algebraic_identity(Program_Asm *prog) {
    // all operands are SSA
    int erase_count = 0;
    auto zero_reg = make_reg(0);
    for(auto f : prog->functions) {
        map<MOperand, int> &int_val_map = f->int_val_map;
        for(auto mb : f->mbs) {
            for(auto I = mb->inst; I; I=I->next) {
                if (I->tag == MI_BINARY) {
                    auto binary = (MI_Binary *) I;
                    switch (binary->op) {
                        case BINARY_ADD: {
                            // 0 + x
                            if(binary->lhs.tag == IMM && binary->lhs.value == 0 || binary->lhs == zero_reg) {
                                replace_operand_by_operand(&binary->dst, &binary->rhs, f);
                                I->mark();
                            }
                            // x + 0
                            else if(binary->rhs.tag == IMM && binary->rhs.value == 0 || binary->rhs == zero_reg) {
                                replace_operand_by_operand(&binary->dst, &binary->lhs, f);
                                I->mark();
                            }
                            // x + x
                            else if(binary->lhs == binary->rhs) {
                                auto lsl = new MI_Binary(BINARY_LSL, binary->dst, binary->lhs, make_imm(1));
                                insert(lsl, I);
                                I->mark();
                            }
                            // x + imm12-constant
                            else if(int_val_map.count(binary->rhs) && can_be_imm12(int_val_map[binary->rhs])) {
                                binary->rhs = make_imm(int_val_map[binary->rhs]);
                            }
                        } break;
                        case BINARY_SUBTRACT: {
                            // x - 0
                            if(binary->rhs.tag == IMM && binary->rhs.value == 0 || binary->rhs == zero_reg) {
                                replace_operand_by_operand(&binary->dst, &binary->lhs, f);
                                I->mark();
                            }
                            // x - x
                            else if(binary->lhs == binary->rhs) {
                                replace_operand_by_operand(&binary->dst, &zero_reg, f);
                                I->mark();
                            }
                            // x - imm12-constant
                            else if(int_val_map.count(binary->rhs) && can_be_imm12(int_val_map[binary->rhs])) {
                                binary->rhs = make_imm(int_val_map[binary->rhs]);
                            }
                        } break;
                        case BINARY_MULTIPLY: {
                            // 0 * x
                            if(int_val_map.count(binary->lhs) && int_val_map[binary->lhs] == 0) {
                                replace_operand_by_operand(&binary->dst, &zero_reg, f);
                                I->mark();
                            }
                            // x * 0
                            else if(int_val_map.count(binary->rhs) && int_val_map[binary->rhs] == 0) {
                                replace_operand_by_operand(&binary->dst, &zero_reg, f);
                                I->mark();
                            }
                            // 1 * x
                            else if(int_val_map.count(binary->lhs) && int_val_map[binary->lhs] == 1) {
                                replace_operand_by_operand(&binary->dst, &binary->rhs, f);
                                I->mark();
                            }
                            // x * 1
                            else if(int_val_map.count(binary->rhs) && int_val_map[binary->rhs] == 1) {
                                replace_operand_by_operand(&binary->dst, &binary->lhs, f);
                                I->mark();
                            }

                            if(int_val_map.count(binary->rhs) == 0)
                                continue;
                            int const_rhs = int_val_map[binary->rhs];
                            if(const_rhs <= 1)
                                continue;
                            int log2v = __builtin_ctz(const_rhs);
                            int v0 = 1 << log2v;
                            // x * 2^sh
                            if(const_rhs == v0) {
                                auto shmat = int(log2(const_rhs));
                                auto lsl = new MI_Binary(BINARY_LSL, binary->dst, binary->lhs, make_imm(shmat));
                                insert(lsl, I);
                                I->mark();
                            }
                            // // x * (2^sh - 1) = x << sh - x
                            // else if(is_exp(const_rhs + 1)) {
                            //     auto shmat = int(log2(const_rhs + 1));
                            //     auto tmp = make_vreg(f->vreg_count++);
                            //     auto lsl = new MI_Binary(BINARY_LSL, tmp, binary->lhs, make_imm(shmat));
                            //     auto sub = new MI_Binary(BINARY_SUBTRACT, binary->dst, tmp, binary->lhs);
                            //     insert(lsl, I);
                            //     insert(sub, I);
                            //     I->mark();
                            // }
                            // // x * (2^sh + 1) = x << sh + x
                            // else if(is_exp(const_rhs - 1)) {
                            //     auto shmat = int(log2(const_rhs - 1));
                            //     auto tmp = make_vreg(f->vreg_count++);
                            //     auto lsl = new MI_Binary(BINARY_LSL, tmp, binary->lhs, make_imm(shmat));
                            //     auto add = new MI_Binary(BINARY_ADD, binary->dst, tmp, binary->lhs);
                            //     insert(lsl, I);
                            //     insert(add, I);
                            //     I->mark();
                            // }
                            // x * (2^sh1 + 2^sh2) = x << sh1 + x << sh2
                            else if(__builtin_popcount(const_rhs - v0) == 1) {
                                int s = __builtin_ctz(const_rhs - v0);
                                auto tmp = make_vreg(f->vreg_count++);
                                auto lsl = new MI_Binary(BINARY_LSL, tmp, binary->lhs, make_imm(s - log2v));
                                auto add = new MI_Binary(BINARY_ADD, binary->dst, tmp, binary->lhs);
                                insert(lsl, I);
                                insert(add, I);
                                if(log2v) {
                                    tmp = make_vreg(f->vreg_count++);
                                    lsl = new MI_Binary(BINARY_LSL, binary->dst, binary->dst, make_imm(log2v));
                                    insert(lsl, I);
                                }
                                I->mark();
                            }
                            // x * (2^sh1 - 2^sh2) = x << sh1 - x << sh2
                            else if(__builtin_popcount(const_rhs - v0) == 1) {
                                int s = __builtin_ctz(const_rhs - v0);
                                auto tmp = make_vreg(f->vreg_count++);
                                auto lsl = new MI_Binary(BINARY_LSL, tmp, binary->lhs, make_imm(s - log2v));
                                auto sub = new MI_Binary(BINARY_SUBTRACT, binary->dst, tmp, binary->lhs);
                                insert(lsl, I);
                                insert(sub, I);
                                if(log2v) {
                                    tmp = make_vreg(f->vreg_count++);
                                    lsl = new MI_Binary(BINARY_LSL, binary->dst, binary->dst, make_imm(log2v));
                                    insert(lsl, I);
                                }
                                I->mark();
                            }
                        } break;
                        case BINARY_DIVIDE: {
                            if(int_val_map.count(binary->rhs) == 0)
                                continue;
                            int const_rhs = int_val_map[binary->rhs];
                            if(const_rhs <= 1)
                                continue;
                            int log2v = __builtin_ctz(const_rhs);
                            int v0 = 1 << log2v;

                            // x / 2^sh
                            if(is_exp(const_rhs)) {
                                // auto lsr_lhs = make_vreg(f->vreg_count++);
                                // if(const_rhs == 2) {
                                //     auto lsr = new MI_Binary(BINARY_LSR, lsr_lhs, binary->lhs, make_imm(31));
                                //     auto add = new MI_Binary(BINARY_ADD, binary->dst, lsr_lhs, binary->lhs);
                                //     insert(lsr, I);
                                //     insert(add, I);
                                // }
                                // else {
                                //     auto asr = new MI_Binary(BINARY_ASR, binary->dst, binary->lhs, make_imm(31));
                                //     auto lsr = new MI_Binary(BINARY_LSR, lsr_lhs, binary->dst, make_imm(64 - log2v));
                                //     auto add = new MI_Binary(BINARY_ADD, binary->dst, binary->lhs, lsr_lhs);
                                //     insert(asr, I);
                                //     insert(lsr, I);
                                //     insert(add, I);
                                // }
                                // auto asr = new MI_Binary(BINARY_ASR, binary->dst, binary->dst, make_imm(log2v));
                                // insert(asr, I);
                                auto lsl_lhs = make_vreg(f->vreg_count++);
                                auto r0 = make_vreg(f->vreg_count++);
                                auto r1 = make_vreg(f->vreg_count++);
                                auto r2 = make_vreg(f->vreg_count++);
                                // auto lsl = new MI_Binary(BINARY_LSL, lsl_lhs, binary->lhs, make_imm(1));
                                // auto lsr = new MI_Binary(BINARY_LSR, lsl_lhs, lsl_lhs, make_imm(64 - log2v));
                                // auto add = new MI_Binary(BINARY_ADD, lsl_lhs, binary->lhs, lsl_lhs);
                                // auto asr = new MI_Binary(BINARY_ASR, binary->dst, lsl_lhs, make_imm(log2v));
                                auto lsl = new MI_Binary(BINARY_LSL, lsl_lhs, binary->lhs, make_imm(1));
                                auto lsr = new MI_Binary(BINARY_LSR, r0, lsl_lhs, make_imm(64 - log2v));
                                auto add = new MI_Binary(BINARY_ADD, r1, binary->lhs, r0);
                                auto asr = new MI_Binary(BINARY_ASR, binary->dst, r1, make_imm(log2v));
                                add->is_rv64 = true;
                                insert(lsl, I);
                                insert(lsr, I);
                                insert(add, I);
                                insert(asr, I);
                                I->mark();
                            }
                            else {
                                auto mag = magic(const_rhs);
                                auto M = make_imm(mag.first);
                                auto s = make_imm(mag.second);
                                if(!can_be_imm12(mag.first)) {
                                    M = make_vreg(f->vreg_count++);
                                    emit_load_of_constant(M, mag.first, mb, I);
                                }
                                auto r0 = make_vreg(f->vreg_count++);
                                auto r1 = make_vreg(f->vreg_count++);
                                auto r2 = make_vreg(f->vreg_count++);
                                auto r3 = make_vreg(f->vreg_count++);
                                auto r4 = make_vreg(f->vreg_count++);
                                auto r5 = make_vreg(f->vreg_count++);
                                auto mul1 = new MI_Binary(BINARY_MULTIPLY, r0, binary->lhs, M);
                                mul1->is_rv64 = true;
                                insert(mul1, I);

                                if(mag.first & (1LL << 31)) {
                                    s = make_imm(mag.second);
                                    auto srli1 = new MI_Binary(BINARY_LSR, r5, r0, make_imm(32));
                                    auto add2 = new MI_Binary(BINARY_ADD, r0, r5, binary->lhs);
                                    auto srli2 = new MI_Binary(BINARY_LSR, r1, r0, make_imm(31));
                                    auto srai = new MI_Binary(BINARY_ASR, r2, r0, s);
                                    insert(srli1, I);
                                    insert(add2, I);
                                    insert(srli2, I);
                                    insert(srai, I);
                                }
                                else {
                                    s = make_imm(mag.second + 32);
                                    auto srli = new MI_Binary(BINARY_LSR, r1, r0, make_imm(63));
                                    auto srai = new MI_Binary(BINARY_ASR, r2, r0, s);
                                    srli->is_rv64 = true;
                                    srai->is_rv64 = true;
                                    insert(srli, I);
                                    insert(srai, I);
                                }
                                
                                auto add = new MI_Binary(BINARY_ADD, binary->dst, r1, r2);
                                add->is_rv64 = true;
                                
                                insert(add, I);
                                I->mark();
                            }
                        } break;
                        case BINARY_MOD: {
                            if(binary->is_rv64)
                                continue;
                            if(int_val_map.count(binary->rhs) == 0)
                                continue;
                            int const_rhs = int_val_map[binary->rhs];
                            if(const_rhs <= 1)
                                continue;
                            int log2v = __builtin_ctz(const_rhs);
                            int v0 = 1 << log2v;

                            if(is_exp(const_rhs)) {
                                // auto lsr_lhs = make_vreg(f->vreg_count++);
                                // if(const_rhs == 2) {
                                //     auto lsr = new MI_Binary(BINARY_LSR, binary->dst, binary->lhs, make_imm(31));
                                //     auto add = new MI_Binary(BINARY_ADD, binary->dst, lsr_lhs, binary->lhs);
                                //     insert(lsr, I);
                                //     insert(add, I);
                                // }
                                // else {
                                //     auto asr = new MI_Binary(BINARY_ASR, binary->dst, binary->lhs, make_imm(31));
                                //     auto lsr = new MI_Binary(BINARY_LSR, lsr_lhs, binary->dst, make_imm(32 - log2v));
                                //     auto add = new MI_Binary(BINARY_ADD, binary->dst, binary->lhs, lsr_lhs);
                                //     insert(asr, I);
                                //     insert(lsr, I);
                                //     insert(add, I);
                                // }
                                // auto lsl_tmp = make_vreg(f->vreg_count++);
                                // auto lsr = new MI_Binary(BINARY_LSR, binary->dst, binary->dst, make_imm(log2v));
                                // auto lsl = new MI_Binary(BINARY_LSL, lsl_tmp, binary->dst, make_imm(log2v));
                                // auto sub = new MI_Binary(BINARY_SUBTRACT, binary->dst, binary->lhs, lsl_tmp);
                                // insert(lsr, I);
                                // insert(lsl, I);
                                // insert(sub, I);

                                auto lsl_lhs = make_vreg(f->vreg_count++);
                                auto r0 = make_vreg(f->vreg_count++);
                                auto r1 = make_vreg(f->vreg_count++);
                                auto r2 = make_vreg(f->vreg_count++);
                                auto const_opr = make_imm(-const_rhs);
                                auto lsl = new MI_Binary(BINARY_LSL, lsl_lhs, binary->lhs, make_imm(1));
                                auto lsr = new MI_Binary(BINARY_LSR, r0, lsl_lhs, make_imm(64 - log2v));
                                auto add = new MI_Binary(BINARY_ADD, r1, binary->lhs, r0);
                                add->is_rv64 = true;
                                insert(lsl, I);
                                insert(lsr, I);
                                insert(add, I);
                                if(!can_be_imm12(-const_rhs)) {
                                    const_opr = make_vreg(f->vreg_count++);
                                    emit_load_of_constant(const_opr, -const_rhs, mb, I);
                                }
                                auto and_instr = new MI_Binary(BINARY_BITWISE_AND, r2, r1, const_opr);
                                auto sub = new MI_Binary(BINARY_SUBTRACT, binary->dst, binary->lhs, r2);
                                insert(and_instr, I);
                                insert(sub, I);
                                I->mark();
                            }
                            else {
                                auto mag = magic(const_rhs);
                                auto M = make_imm(mag.first);
                                auto s = make_imm(mag.second);
                                if(!can_be_imm12(mag.first)) {
                                    M = make_vreg(f->vreg_count++);
                                    emit_load_of_constant(M, mag.first, mb, I);
                                }
                                auto r0 = make_vreg(f->vreg_count++);
                                auto r1 = make_vreg(f->vreg_count++);
                                auto r2 = make_vreg(f->vreg_count++);
                                auto r3 = make_vreg(f->vreg_count++);
                                auto r4 = make_vreg(f->vreg_count++);
                                auto r5 = make_vreg(f->vreg_count++);
                                auto mul1 = new MI_Binary(BINARY_MULTIPLY, r0, binary->lhs, M);
                                mul1->is_rv64 = true;
                                insert(mul1, I);

                                if(mag.first & (1LL << 31)) {
                                    s = make_imm(mag.second);
                                    auto srli1 = new MI_Binary(BINARY_LSR, r5, r0, make_imm(32));
                                    auto add2 = new MI_Binary(BINARY_ADD, r0, r5, binary->lhs);
                                    auto srli2 = new MI_Binary(BINARY_LSR, r1, r0, make_imm(31));
                                    auto srai = new MI_Binary(BINARY_ASR, r2, r0, s);
                                    insert(srli1, I);
                                    insert(add2, I);
                                    insert(srli2, I);
                                    insert(srai, I);
                                }
                                else {
                                    s = make_imm(mag.second + 32);
                                    auto srli = new MI_Binary(BINARY_LSR, r1, r0, make_imm(63));
                                    auto srai = new MI_Binary(BINARY_ASR, r2, r0, s);
                                    srli->is_rv64 = true;
                                    srai->is_rv64 = true;
                                    insert(srli, I);
                                    insert(srai, I);
                                }

                                auto add = new MI_Binary(BINARY_ADD, r3, r1, r2);
                                auto mul2 = new MI_Binary(BINARY_MULTIPLY, r4, r3, binary->rhs);
                                auto sub = new MI_Binary(BINARY_SUBTRACT, binary->dst, binary->lhs, r4);
                                add->is_rv64 = true;
                                
                                insert(add, I);
                                insert(mul2, I);
                                insert(sub, I);
                                I->mark();
                            }
                        } break;
                        default: break;
                    }
                    if(I->marked) {
                        // auto binary = (MI_Binary *)(I);
                        // printf("erase instruction: ");
                        // print_operand(binary->dst);
                        // printf(",");
                        // print_operand(binary->lhs);
                        // printf(",");
                        // print_operand(binary->rhs);
                        // printf("   op_type: %d\n", binary->op);
                        erase_count ++;
                    }
                }
            }
            mb->erase_marked_values();
        }
    }
    printf("[opt]: algebraic identity ereased %d instructions\n", erase_count);
}