#include "arm.h"
#include "asm_passes.h"

struct BinaryMOperands {
    MOperand lhs, rhs;
    Binary_Op_Type op;

    BinaryMOperands() {}
    BinaryMOperands(MOperand _lhs, MOperand _rhs, Binary_Op_Type _op): lhs(_lhs), rhs(_rhs), op(_op) {}

    bool operator==(const BinaryMOperands &bm) const {
        if(lhs == bm.lhs && rhs == bm.rhs && op == bm.op)
            return true;
        else
            return false;
    }

    bool operator<(const BinaryMOperands &bm) const {
        if(op != bm.op)
            return op < bm.op;
        else if(!(lhs == bm.lhs))
            return lhs < bm.lhs;
        else
            return rhs < bm.rhs;
    }
};


// address and offset
static map<MOperand, MI_Load *> load_map;
// binary reg
static map<BinaryMOperands, MI_Binary *> binary_map;

void clear_map() {
    load_map.clear();
    binary_map.clear();
}

void peephole(Program_Asm *prog) {
    int cnt = 0;
    printf("begin peephole:\n");
    for(auto f : prog->functions) {
        for(auto mb : f->mbs) {
            MI *next_instr = mb->inst;
            clear_map();
            for(auto I = mb->inst; I; I=next_instr) {
                next_instr = I->next;
                switch (I->tag)
                {
                case MI_LOAD: {
                    auto load = (MI_Load *)I;
                    if(load->mem_tag == MEM_LOAD_FROM_LITERAL_POOL 
                        || load->mem_tag == MEM_LOAD_GLOBAL_REF) {
                        if(load_map.count(load->base)) {
                            auto former_load = load_map[load->base];
                            replace_operand_by_operand(&load->reg, &former_load->reg, f);
                            cnt ++;
                        }
                        else {
                            load_map[load->base] = load;
                        }
                    }
                } break;
                
                case MI_BINARY: {
                    auto binary = (MI_Binary *)I;
                    BinaryMOperands bm{binary->lhs, binary->rhs, binary->op};
                    if(binary_map.count(bm)) {
                        auto former_binary = binary_map[bm];
                        replace_operand_by_operand(&binary->dst, &former_binary->dst, f);
                        cnt ++;
                    }
                    else {
                        binary_map[bm] = binary;
                    }
                } break;
                default:
                    break;
                }
            }
        }
    }
    printf("[opt]: peephole ereased %d instructions\n", cnt);
}
