#include <iostream>
#include <getopt.h>
#include <string>

#include "antlr4-runtime.h"
#include "SysY2022Lexer.h"
#include "MyVisitor.h"
#include "arm.h"
#include "asm_passes.h"

using namespace antlr4;

int main(int argc, char ** argv) {
  std::ifstream sourceFile(argv[4]);
  assert(sourceFile.is_open());


  ANTLRInputStream input(sourceFile);
  SysY2022Lexer lexer(&input);
  CommonTokenStream tokens(&lexer);
  tokens.fill();
  // for (auto token : tokens.getTokens()) {
  //   std::cout << token->toString() << std::endl;
  // }
  SysY2022Parser parser(&tokens);
  SysY2022Parser::CompUnitContext* tree = parser.compUnit();
  // std::cout << tree->toStringTree(&parser) << std::endl << std::endl;
  MyVisitor visitor;
  // cout << "Begin visit CompUnit" << endl;
  visitor.visitCompUnit(tree);
  // cout << "finish IR generating" << endl;
  bool addSomeInstr = false;
  visitor.opt(addSomeInstr);
  // cout << "finish frontend opt" << endl;

  visitor.print();

  auto program_asm=emit_asm(visitor.irModule);
  cout << "finish emit_asm" << endl;
  bless(program_asm,false);
  cout << "finish bless" << endl;
  String_Builder s;
  build_program_asm(&s, program_asm, visitor.irModule.globalVariables, addSomeInstr);
  s.add_terminator();

  FILE* assembly_file = fopen(argv[3], "w");
  if (assembly_file == NULL) {
      assert(false && "error opening assembly output file");
  }

  fprintf(assembly_file, "%s", s.c_str());
  fclose(assembly_file);
  return 0;
}