#include "computeUse.h"

bool hasUser(InsID id) {
    return !(id == Return || id == Br || id == Store || id == Call);
}

void computeUse(Module &ir) {
    clearGlobalUse(ir);

    for(auto func: ir.globalFunctions) {
        if(func->isLib)
            continue;
        BlockPtr block = func->block;
        // delete all use
        for(auto arg: func->formArguments) {
            arg->useHead = nullptr;
            arg->numUses = 0;
        }
        for(auto bb: block->basicBlocks) {
            for(auto instr: bb->instructions) {
                if(instr->reg) {
                    instr->reg->useHead = nullptr;
                    instr->reg->numUses = 0;
                }
            }
        }

        // create new use
        for(auto bb: block->basicBlocks) {
            bb->setBelongFunc(func);
            for(auto instr: bb->instructions) {
                instr->basicblock = bb;
                if(instr->reg)
                    instr->reg->I = instr.get();
                int numOperand = instr->getNumOperands();
                if(instr->type == Alloca)
                    continue;
                for(int i = 0; i < numOperand; i ++) {
                    if(hasUser(instr->type))
                        newUse(instr->getOperand(i).get(), instr.get(), instr->reg.get());
                    else
                        newUse(instr->getOperand(i).get(), instr.get());
                }
            }
        }
    }
}

void clearGlobalUse(Module &ir) {
    // delete all use
    for(auto var: ir.globalVariables) {
        var->useHead = nullptr;
        var->numUses = 0;
    }
}
