#include "cse.h"

bool impossibleToCSE(InsID type) {
    return (type == Br || type == Return || type == Alloca
        || type == Phi || type == Icmp || type == Fcmp);
}

bool isMemoryRelated(InsID type) {
    return type == Load || type == Store || type == Call;
}

bool maybeSameGEP(Instruction *Ii, Instruction *Ij) {
    auto gep1 = dynamic_cast<GetElementPtrInstruction *>(Ii);
    auto gep2 = dynamic_cast<GetElementPtrInstruction *>(Ij);
    // @BUG: two 'from' may be two 'GEP->reg' from the same array base
    if (gep1 && gep2) {
        // if(gep1->base == nullptr || gep2->base == nullptr || gep1->allIndex.size() != gep2->allIndex.size())
        if(gep1->allIndex.size() != gep2->allIndex.size())
            return false;
        if(gep1->base != gep2->base)
            return false;

        bool matched = true, mayMatched = false;
        for(int i = 0; i < gep1->allIndex.size(); i ++) {
            if(gep1->allIndex[i] == gep2->allIndex[i])
                continue;
            
            // same constant
            auto const1 = dynamic_cast<Const *>(gep1->allIndex[i].get());
            auto const2 = dynamic_cast<Const *>(gep2->allIndex[i].get());
            if(const1 == nullptr || const2 == nullptr) {
                mayMatched = true;
            }
            else if(const1->intVal != const2->intVal) {
                matched = false;
                break;
            }
        }
        if(!matched)
            return false;
        else if(mayMatched)
            return true;
    }
    return false;
}

void cse(BlockPtr block){
    int instrCount = 0;
    for(int bbIdx = 0; bbIdx < block->basicBlocks.size(); bbIdx ++) {
        auto bb = block->basicBlocks[bbIdx];
        unordered_map<InstructionPtr, BasicBlockPtr> del;
        vector<BasicBlockPtr> succBlocks{bb};
        unordered_set<BasicBlockPtr> changedBB;
        unordered_map<BrInstruction *, bool> canBrContinue;
        unordered_map<BasicBlockPtr, bool> canBBContinue;
        for(auto succBB: bb->succBasicBlocks) {
            succBlocks.emplace_back(succBB);
        }

        for(int i = 0; i < bb->instructions.size(); i++){
            //此时是ssa，所以不需要考虑值会不会改变
            auto Ia = bb->instructions[i];
            bool flag = false;

            // impossible to do CSE
            if(impossibleToCSE(Ia->type) || del.count(Ia))
                continue;

            for(auto succBB: succBlocks) {
                int j = (succBB == bb ? i + 1 : 0);

                // if CSE for succBB may incur error, skip it
                if(canBBContinue.find(succBB) != canBBContinue.end() && !canBBContinue[succBB])
                    continue;
    
                for( ; j < succBB->instructions.size(); j++){
                    auto Ib = succBB->instructions[j];
                    // load/store -> call
                    if((Ia->type == Load || Ia->type == Store) && Ib->type == Call) {
                        // if happened in current block, it dominates the other blocks
                        if(succBB == bb)
                            flag = true;
                        else
                            break;
                    }
                    // destination of Br
                    if(Ib->type == Br) {
                        auto br = (BrInstruction *)(Ib.get());
                        bool canContinue;
                        if(canBrContinue.find(br) != canBrContinue.end()) {
                            canContinue = canBrContinue[br];
                        }
                        else {
                            // determinate whether succBB is reached by 'bb' only
                            auto nextBB = block->LabelBBMap[br->label1];
                            canBBContinue[nextBB] = (nextBB->predBasicBlocks.size() == 1);
                            canContinue = canBBContinue[nextBB];
                            if(br->label2) {
                                nextBB = block->LabelBBMap[br->label2];
                                canBBContinue[nextBB] = (nextBB->predBasicBlocks.size() == 1);
                                canContinue = canContinue || canBBContinue[nextBB];
                            }
                            canBrContinue[br] = canContinue;
                        }
                        if(!canContinue) {
                            if(succBB == bb)
                                flag = true;
                            else
                                break;
                        }
                    }
                    if(flag) 
                        break;

                    if(del.count(Ib)) 
                        continue;

                    if(Ia->type == Ib->type){
                        if(Ia->type == Binary){
                            auto Ii = (BinaryInstruction*)(Ia.get());
                            auto Ij = (BinaryInstruction*)(Ib.get());
                            if(Ii->op!=Ij->op){
                                continue;
                            }
                            if(Ii->a==Ij->a&&Ii->b==Ij->b){
                                replaceVarByVar(Ij->reg, Ii->reg);
                                del[Ib] = succBB;
                                changedBB.insert(succBB);
                                deleteUser(Ij->reg);
                            }
                            else if(Ii->a->isConst&&Ij->a->isConst&&Ii->b==Ij->b){
                                auto newA1 = dynamic_cast<Const*>(Ii->a.get());
                                auto newA2 = dynamic_cast<Const*>(Ij->a.get());
                                if(newA1->type->isInt()&&newA2->type->isInt()){
                                    if(newA1->intVal ==  newA2->intVal){
                                        replaceVarByVar(Ij->reg, Ii->reg);
                                        del[Ib] = succBB;
                                        changedBB.insert(succBB);
                                        deleteUser(Ij->reg);
                                    }
                                }
                                else if(newA1->type->isFloat()&&newA2->type->isFloat()){
                                    if(newA1->floatVal ==  newA2->floatVal){
                                        replaceVarByVar(Ij->reg, Ii->reg);
                                        del[Ib] = succBB;
                                        changedBB.insert(succBB);
                                        deleteUser(Ij->reg);
                                    }
                                }
                                else if(newA1->type->isBool()&&newA2->type->isBool()){
                                    if(newA1->boolVal ==  newA2->boolVal){
                                        replaceVarByVar(Ij->reg, Ii->reg);
                                        del[Ib] = succBB;
                                        changedBB.insert(succBB);
                                        deleteUser(Ij->reg);
                                    }
                                }
                            }
                            else if(Ii->b->isConst&&Ij->b->isConst&&Ii->a==Ij->a){
                                auto newB1 = dynamic_cast<Const*>(Ii->b.get());
                                auto newB2 = dynamic_cast<Const*>(Ij->b.get());
                                if(newB1->type->isInt()&&newB2->type->isInt()){
                                    if(newB1->intVal ==  newB2->intVal){
                                        replaceVarByVar(Ij->reg, Ii->reg);
                                        del[Ib] = succBB;
                                        changedBB.insert(succBB);
                                        deleteUser(Ij->reg);
                                    }
                                }
                                else if(newB1->type->isFloat()&&newB2->type->isFloat()){
                                    if(newB1->floatVal ==  newB2->floatVal){
                                        replaceVarByVar(Ij->reg, Ii->reg);
                                        del[Ib] = succBB;
                                        changedBB.insert(succBB);
                                        deleteUser(Ij->reg);
                                    }
                                }
                                else if(newB1->type->isBool()&&newB2->type->isBool()){
                                    if(newB1->boolVal ==  newB2->boolVal){
                                        replaceVarByVar(Ij->reg, Ii->reg);
                                        del[Ib] = succBB;
                                        changedBB.insert(succBB);
                                        deleteUser(Ij->reg);
                                    }
                                }
                            }
                            //两个都是常量
                            else{
                            }
                        }
                        // store -> store
                        else if(Ia->type == Store) {
                            auto Ii = (StoreInstruction*)(Ia.get());
                            auto Ij = (StoreInstruction*)(Ib.get());
                            // @TODO: now we only optimize this situation in one block
                            // can be optimized with the information of dominator tree
                            // bb -> succBB(unique predecessor and successor)
                            if(succBB != bb && bb->getOutdegree() > 1) {
                                flag = true;
                                break;
                            }
                            if(Ii->des == Ij->des || maybeSameGEP(Ii->des->I, Ij->des->I)) {
                                flag = true;
                                break;
                            }
                            if(Ii->des == Ij->des) {
                                del[Ia] = bb;
                                changedBB.insert(bb);
                                rmInstructionUse(Ia, Ii->des);
                                rmInstructionUse(Ia, Ii->value);
                            }
                        }
                        // load -> load
                        else if(Ia->type == Load) {
                            auto Ii = (LoadInstruction*)(Ia.get());
                            auto Ij = (LoadInstruction*)(Ib.get());
                            if(Ii->from == Ij->from) {
                                replaceVarByVar(Ij->to, Ii->to);
                                del[Ib] = succBB;
                                changedBB.insert(succBB);
                                deleteUser(Ij->to);
                            }
                        }
                        // pointer alias
                        else if(Ia->type == GEP) {
                            auto Ii = (GetElementPtrInstruction*)(Ia.get());
                            auto Ij = (GetElementPtrInstruction*)(Ib.get());
                            if (Ii->from == Ij->from && Ii->getNumOperands() == Ij->getNumOperands()) {
                                bool matched = true;
                                for(int i = 0; i < Ii->index.size(); i ++) {
                                    // same SSA variable
                                    if(Ii->index[i] == Ij->index[i])
                                        continue;
                                    
                                    // same constant
                                    auto const0 = dynamic_cast<Const *>(Ii->index[i].get());
                                    auto const1 = dynamic_cast<Const *>(Ij->index[i].get());
                                    if(const0 == nullptr || const1 == nullptr || const0->intVal != const1->intVal) {
                                        matched = false;
                                        break;
                                    }
                                }
                                if(matched) {
                                    replaceVarByVar(Ij->reg, Ii->reg);
                                    del[Ib] = succBB;
                                    changedBB.insert(succBB);
                                    deleteUser(Ij->reg);
                                }
                            }
                        }
                        else if(Ia->type == Ext) {
                            auto Ii = (ExtInstruction*)(Ia.get());
                            auto Ij = (ExtInstruction*)(Ib.get());
                            if(Ii->from == Ij->from && Ii->to->ID == Ij->to->ID) {
                                replaceVarByVar(Ij->reg, Ii->reg);
                                del[Ib] = succBB;
                                changedBB.insert(succBB);
                                deleteUser(Ij->reg);
                            }
                        }
                    }
                    else if(Ia->type == Store || Ia->type == Load) {
                        // store -> load
                        if(Ia->type == Store && Ib->type == Load) {
                            auto Ii = (StoreInstruction*)(Ia.get());
                            auto Ij = (LoadInstruction*)(Ib.get());
                            if(Ii->des == Ij->from) {
                                replaceVarByVar(Ij->to, Ii->value);
                                del[Ib] = succBB;
                                changedBB.insert(succBB);
                                deleteUser(Ij->to);
                            }
                        }
                        // load -> store
                        if(Ia->type == Load && Ib->type == Store) {
                            auto Ii = (LoadInstruction*)(Ia.get());
                            auto Ij = (StoreInstruction*)(Ib.get());
                            if(Ii->from == Ij->des || maybeSameGEP(Ii->from->I, Ij->des->I)) {
                                if(succBB != bb)
                                    break;
                                else
                                    flag = true;
                            }
                        }
                    }
                }
            }
        }
        for(auto newBB: changedBB) {
            vector<InstructionPtr > newIns;
            for(int i = 0; i < newBB->instructions.size(); i++){
                if(del.count(newBB->instructions[i]) == 0){
                    newIns.push_back(newBB->instructions[i]);
                }
            }
            newBB->instructions = newIns;
        }
        instrCount += del.size();

        // if(del.size()) {
        //     cout << bb->label->name << " eliminated:" << endl;
        //     for(auto instrPair: del) {
        //         instrPair.first->print();
        //     }
        // }
    }
    cout << "[frontend opt]: CSE eliminated " << instrCount << " instructions" << endl;
}