#include "dce.h"


bool findDeadCode(Instruction *Ins, unordered_map<Instruction *, bool>& del){
    if(Ins == nullptr)
        return false;
    //Return,Alloca,Store,Load
    if(Ins->type==Bitcast){
        auto t = ((BitCastInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            //移除use
            rmInstructionUse(Ins,t->from);
            del[Ins] = true;
            //不是常量且未删除
            if(t->from->I!=nullptr&&del.find(t->from->I) == del.end())
                findDeadCode(t->from->I, del);
            return true;
        }
        return false;
    }
    else if(Ins->type == Ext){
        auto t = ((ExtInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            //移除use
            rmInstructionUse(Ins,t->from);
            del[Ins] = true;
            if(t->from->I!=nullptr&&del.find(t->from->I) == del.end())
                findDeadCode(t->from->I, del);
            return true;
        }
        return false;
    }
    else if(Ins->type == Sitofp){
        auto t = ((SitofpInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            //移除use
            rmInstructionUse(Ins,t->from);
            del[Ins] = true;
            if(t->from->I!=nullptr&&del.find(t->from->I) == del.end())
                findDeadCode(t->from->I, del);
            return true;
        }
        return false;
    }
    else if(Ins->type == Fptosi){
        auto t = ((FptosiInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            //移除use
            rmInstructionUse(Ins,t->from);
            del[Ins] = true;
            if(t->from->I!=nullptr&&del.find(t->from->I) == del.end())
                findDeadCode(t->from->I, del);
            return true;
        }
        return false;
    }
    else if(Ins->type == Call){
        auto t = ((CallInstruction*)(Ins));
        if(t->reg->useHead==nullptr && t->func->isReenterable && !(t->func->isLib)){
            del[Ins] = true;
            for(int i=0;i<t->argv.size();i++){
                rmInstructionUse(Ins,t->argv[i]);
                if(t->argv[i]->I!=nullptr&&del.find(t->argv[i]->I) == del.end())
                    findDeadCode(t->argv[i]->I, del);
            }
            return true;
        }
        return false;
    }
    else if(Ins->type == GEP){
        auto t = ((GetElementPtrInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            del[Ins] = true;
            //移除use
            rmInstructionUse(Ins,t->from);
            if(t->from->I!=nullptr&&del.find(t->from->I) == del.end())
                findDeadCode(t->from->I, del);
            for(int i=0;i<t->index.size();i++){
                rmInstructionUse(Ins,t->index[i]);
                if(t->index[i]->I!=nullptr&&del.find(t->index[i]->I) == del.end())
                    findDeadCode(t->index[i]->I, del);
            }
            return true;
        }
        return false;
    }
    else if(Ins->type == Binary){
        auto t = ((BinaryInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            del[Ins] = true;
            //移除use
            rmInstructionUse(Ins,t->a);
            if(t->a->I!=nullptr&&del.find(t->a->I) == del.end())
                findDeadCode(t->a->I, del);
            rmInstructionUse(Ins,t->b);
            if(t->b->I!=nullptr&&del.find(t->b->I) == del.end())
                findDeadCode(t->b->I, del);
            return true;
        }
        return false;
    }
    else if(Ins->type == Fneg){
        auto t = ((FnegInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            //移除use
            rmInstructionUse(Ins,t->a);
            del[Ins] = true;
            if(t->a->I!=nullptr&&del.find(t->a->I) == del.end())
                findDeadCode(t->a->I, del);
            return true;
        }
        return false;
    }
    else if(Ins->type == Icmp){
        auto t = ((IcmpInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            del[Ins] = true;
            //移除use
            rmInstructionUse(Ins,t->a);
            if(t->a->I!=nullptr&&del.find(t->a->I) == del.end())
                findDeadCode(t->a->I, del);
            rmInstructionUse(Ins,t->b);
            if(t->b->I!=nullptr&&del.find(t->b->I) == del.end())
                findDeadCode(t->b->I, del);
            return true;
        }
        return false;
    }
    else if(Ins->type == Fcmp){
        auto t = ((FcmpInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            del[Ins] = true;
            //移除use
            rmInstructionUse(Ins,t->a);
            if(t->a->I!=nullptr&&del.find(t->a->I) == del.end())
                findDeadCode(t->a->I, del);
            rmInstructionUse(Ins,t->b);
            if(t->b->I!=nullptr&&del.find(t->b->I) == del.end())
                findDeadCode(t->b->I, del);
            return true;
        }
        return false;
    }
    else if(Ins->type == Phi){
        auto t = ((PhiInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            del[Ins] = true;
            //移除use
            for(int i=0;i<t->from.size();i++){
                rmInstructionUse(Ins,t->from[i].first);
                if(t->from[i].first->I!=nullptr&&del.find(t->from[i].first->I) == del.end())
                    findDeadCode(t->from[i].first->I, del);
            }
            return true;
        }
        return false;
    }
    else if(Ins->type == Load){
        auto t = ((LoadInstruction*)(Ins));
        if(t->reg->useHead==nullptr){
            del[Ins] = true;
            //移除use
            rmInstructionUse(Ins,t->from);
            if(t->from->I!=nullptr&&del.find(t->from->I) == del.end())
                findDeadCode(t->from->I, del);
            return true;
        }
        return false;
    }
    return false;
}

void dce(BlockPtr block){
    unordered_map<Instruction *,bool> del;
    for(auto&bb:(block->basicBlocks)){
        for(auto& ins:(bb->instructions)){
            findDeadCode(ins.get(),del);
        }
    }
    vector<InstructionPtr> newIns;
    for(auto& bb:(block->basicBlocks)){
        newIns.clear();
        for(auto ins:(bb->instructions)){
            if(del.find(ins.get())==del.end()){
                newIns.emplace_back(ins);
            }
        }
        bb->instructions = newIns;
    }
}

void functionDCE(Module &module) {
    bool fixed;
    do {
        fixed = true;
        //构建调用图
        clearFunctionCallerAndCallee(module);
        for(auto func: module.globalFunctions){
            if(!func->isLib){
                findFunctionCallerAndCallee(func);
            }
        }
        for(auto func: module.globalFunctions) {
            if(func->isLib || func->retVal->type->isVoid() || func->name == "main")
                continue;

            bool retUsed = false;
            unordered_map<ValuePtr, PhiInstruction *> phiRetMap;
            // printf("call for function %s:\n", func->name.c_str());
            for(auto instr: func->callerIns) {
                auto call = dynamic_cast<CallInstruction *>(instr.get());
                // call->print();
                if(call->reg->useHead != nullptr) {
                    // ignore recursive calling and return directly
                    auto use = call->reg->useHead;
                    if(call->func == call->basicblock->belongfunc) {
                        if(dynamic_cast<ReturnInstruction *>(use->user)) {
                            continue;
                        }
                        else if(auto phi = dynamic_cast<PhiInstruction *>(use->user)) {
                            if(phi->reg->useHead && phi->reg->numUses == 1 && phi->reg->useHead->user->type == Return) {
                                phiRetMap[call->reg] = phi;
                                continue;
                            }
                        }
                    }
                    retUsed = true;
                }
            }
                

            if(!retUsed) {
                bool changedFunction = false;
                for(auto bb: func->block->basicBlocks) {
                    for(auto instr: bb->instructions) {
                        if(auto ret = dynamic_cast<ReturnInstruction *>(instr.get())) {
                            assert(ret->retValue != nullptr);
                            if(ret->retValue->I != nullptr) {
                                // printf("has retValue: ");
                                // ret->print();
                                rmInstructionUse(ret, ret->retValue);
                                if(ret->retValue->type->isInt()) 
                                    ret->retValue = Const::getConst(Type::getInt(), 0, "pesudoReturnVal");
                                else if(ret->retValue->type->isFloat())
                                    ret->retValue = Const::getConst(Type::getFloat(), 0, "pesudoReturnVal");
                                else
                                    assert(false && "unknow return type");
                                
                                changedFunction = true;
                            }
                        }
                    }
                }
                if(!phiRetMap.empty()) {
                    for(auto [reg, phi]: phiRetMap) {
                        for(auto iter = phi->from.begin(); iter != phi->from.end(); iter ++) {
                            if(iter->first == reg) {
                                phi->from.erase(iter);
                                break;
                            }
                        }
                        rmInstructionUse(phi, reg);
                    }
                }
                if(changedFunction) {
                    dce(func->block);
                    fixed = false;
                }
            }
        }
    } while(!fixed);
}

// types can be loop DCE
bool canLoopDCE(InstructionPtr instr) {
    if(instr == nullptr)
        return true;
    auto type = instr->type;
    return type == Phi || type == Binary || type == Ext;
}

// whether instruction in current loop or subloops
bool inCurLoop(InstructionPtr instr, LoopPtr loop) {
    return loop->contains(instr->basicblock);
} 

bool useDFSVisit(InstructionPtr instr, LoopPtr loop, unordered_map<InstructionPtr, bool> &visitdedMap, function<bool(InstructionPtr)> cond, function<void(InstructionPtr)> action) {
    if(!cond(instr))
        return visitdedMap[instr];
    action(instr);
    if(!canLoopDCE(instr) || !inCurLoop(instr, loop))
        return false;
    assert(instr->reg != nullptr);
    auto use = instr->reg->useHead;
    while(use) {
        assert(use->user != nullptr);
        if(!useDFSVisit(use->user->getSharedThis(), loop, visitdedMap, cond, action))
            return false;
        use = use->next;
    }
    return true;
}



// delete instructions whose def-use chain in loop
void loopDCE(BlockPtr block) {
    int eliminatedCnt = 0;
    for(auto loop: block->getTopLoops()) {
        unordered_set<InstructionPtr> visited;
        unordered_map<InstructionPtr, bool> visitdedMap;
        unordered_set<InstructionPtr> toDelete;
        unordered_set<BasicBlockPtr> changedBB;
        for(auto bb: loop->getLoopBasicBlocks()) {
            for(auto instr: bb->instructions) {
                vector<InstructionPtr> mayDelete;
                if(visited.count(instr))
                    continue;
                bool flag = useDFSVisit(instr, loop, visitdedMap, [&](InstructionPtr i) -> bool {
                    if(visited.count(i))
                        return false;
                    else {
                        visited.insert(i);
                        visitdedMap[i] = true;
                        return true;
                    }
                }, [&](InstructionPtr i) -> void {
                    mayDelete.push_back(i);
                });
                // marking whether a visited instruction can be DCE
                for(auto i: mayDelete) {
                    visitdedMap[i] = flag;
                }

                if(flag) {
                    // for(auto i: mayDelete) {
                    //     i->print();
                    // }
                    for(auto i: mayDelete) {
                        toDelete.insert(i);
                        changedBB.insert(i->basicblock);
                    }
                }
            }
        }
        if(!toDelete.empty()) {
            for(auto bb: changedBB) {
                vector<InstructionPtr> newInstr;
                for(auto instr: bb->instructions) {
                    if(toDelete.count(instr) == 0) {
                        newInstr.emplace_back(instr);
                    }
                }
                bb->instructions = newInstr;
            }
            eliminatedCnt += toDelete.size();
        }
    }
    // if(eliminatedCnt)
    //     printf("loopDCE eliminated %d instructions in function %s\n", eliminatedCnt, block->getEntryBlock()->belongfunc->name.c_str());
}
