#include "eliminateGlobal.h"

static void dfsUseVisit(InstructionPtr instr, function<bool(InstructionPtr)> visit) {
    if(visit(instr))
        return;

    auto reg = instr->reg;
    assert(reg !=nullptr);
    auto use = reg->useHead;
    while(use) {
        if(use->user)
            dfsUseVisit(use->user->getSharedThis(), visit);
        use = use->next;
    }
}

void eliminateGlobal(Module &ir) {
    printf("enter eliminateGlobal\n");
    fflush(stdout);
    vector<VariablePtr> newgbvar;
    for(auto globalVar: ir.globalVariables) {
        // only handle one-dimension int array
        unordered_set<InstructionPtr> relatedInstr;   // tracked store instr
        bool flag = true;

        // dfs traverse for arr uses
        auto use = globalVar->useHead;
        queue<InstructionPtr> dfsqueue;
        while(use) {
            assert(use->user != nullptr);
            dfsUseVisit(use->user->getSharedThis(), [&](InstructionPtr instr) -> bool {
                if(!flag || instr == nullptr || relatedInstr.count(instr))
                    return true;

                relatedInstr.insert(instr);
                // instr->print();
                // fflush(stdout);
                if(instr->type == Store) {
                    return true;
                }
                else if(instr->type == Load) {
                    flag = false;
                    return true;
                }
                else if(instr->type == GEP) {
                    return false;
                }
                // should not be function parameter
                else if(instr->type == Call) {
                    flag = false;
                    return true;
                }
                else {
                    return true;
                }
            });
            use = use->next;
        }
        

        // if 'arr' satisfy all conditions
        if(flag) {
            // printf("instructions related to global variable %s:\n", globalVar->getStr().c_str());
            // for(auto instr: relatedInstr)
            //     instr->print();
            // fflush(stdout);
            // cout << "here" << endl;
            for(auto func: ir.globalFunctions) {
                // cout << "entering function " << endl;
                if(func->isLib)
                    continue;
                for(auto bb: func->block->basicBlocks) {
                    // cout << "entering bb " << bb->label->getStr() << endl;
                    vector<InstructionPtr> newInstr;
                    for(auto instr: bb->instructions) {
                        if(relatedInstr.count(instr) == 0)
                            newInstr.emplace_back(instr);
                        // else
                        // {
                        //     printf("is about to erase\n");
                        //     instr->print();
                        //     fflush(stdout);
                        // }
                            
                    }
                    bb->instructions = newInstr;
                }
            }

        }
        else
            newgbvar.push_back(globalVar);
    }

    ir.globalVariables = newgbvar;

    printf("Optimization Finished Correctly\n");
    fflush(stdout);
}