#include "functionCache.h"

void functionCache(Module &ir) {
    for(auto func: ir.globalFunctions) {
        if(func->isLib)
            continue;
        bool canCached = true;
        
        // function parameter must be 2
        if(func->formArguments.size() != 2)
            canCached = false;
        // do not allow ptr
        for(auto arg: func->formArguments) {
            if(auto ptr = dynamic_cast<Ptr *>(arg.get())) {
                canCached = false;
                break;
            }
        }
        if(!canCached)
            continue;

        for(auto bb: func->block->basicBlocks) {
            for(auto instr: bb->instructions) {
                // no load/store
                if(instr->type == Load || instr->type == Store) {
                    canCached = false;
                    break;
                }
                // only call to itself
                else if(instr->type == Call) {
                    auto call = dynamic_cast<CallInstruction *>(instr.get());
                    if(call->func != func) {
                        canCached = false;
                        break;
                    }
                }
                // no allocate for local array
                else if(instr->type == Alloca) {
                    canCached = false;
                    break;
                }
            }
            if(!canCached)
                break;
        }

        if(canCached) {
            auto entry = func->block->getEntryBlock();
            auto argv = vector<ValuePtr>{
                RegPtr(new Reg(TypePtr(new PtrType(Type::getInt())), "lut_addr")),
                RegPtr(new Reg(Type::getInt(), "param1")),
                RegPtr(new Reg(Type::getInt(), "param2"))};
            auto cacheCall = InstructionPtr(new CallInstruction(ir.globalFunctions[14], argv, entry));
            entry->pushFrontInstruction(vector<InstructionPtr>{cacheCall});
        }
    }
}
