#include "ifRangeCombine.h"

bool ifRangeCombine(BlockPtr block) {
    unordered_set<BasicBlockPtr> visitedBB;
    bool flag = false;
    for(auto bb: block->basicBlocks) {
        if(visitedBB.count(bb))
            continue;
        visitedBB.insert(bb);

        for(int i = 0; i < bb->instructions.size(); i ++) {
            auto instr = bb->instructions[i];
            if(instr->type == Icmp) {
                auto icmp = dynamic_cast<IcmpInstruction *>(instr.get());
            }
            else if(instr->type == Fcmp) {
                auto fcmp = dynamic_cast<FcmpInstruction *>(instr.get());
                auto lhs = fcmp->a;
                auto rhs = fcmp->b;
                auto constRhs = dynamic_cast<Const *>(rhs.get());
                if(!constRhs)
                    continue;
                // assume all constant rhs is float
                float rhsValue = constRhs->floatVal;
                
                // next instr is Br and exp is fcmp->reg
                assert(i + 1 < bb->instructions.size());
                if(bb->instructions[i + 1]->type != Br)
                    continue;

                auto nextInstr = bb->instructions[i + 1];
                auto br = dynamic_cast<BrInstruction *>(nextInstr.get());
                assert(br->exp == fcmp->reg);

                auto trueBB = block->LabelBBMap[br->label1];
                auto falseBB = block->LabelBBMap[br->label2];
                assert(trueBB && falseBB);
                
                // enter trueBB, find interset of constRhs and nextConstRhs
                if(trueBB->instructions.size() > 1 && trueBB->instructions[trueBB->instructions.size() - 2]->type == Fcmp) {
                    auto nextFcmp = dynamic_cast<FcmpInstruction *>(trueBB->instructions[trueBB->instructions.size() - 2].get());
                    auto nextBr = dynamic_cast<BrInstruction *>(trueBB->instructions.back().get());
                    auto nextLhs = nextFcmp->a;
                    auto nextRhs = nextFcmp->b;
                    auto nextConstRhs = dynamic_cast<Const *>(nextRhs.get());
                    bool canBeMerged = false;

                    if(nextConstRhs && nextBr) {
                        float nextRhsValue = nextConstRhs->floatVal;
                        if(fcmp->op == "<") {
                            // (-inf, rhsValue) and (nextRhsValue, +inf)
                            if(nextFcmp->op == ">") {
                                if(rhsValue <= nextRhsValue)
                                    canBeMerged = true;
                            }
                            // (-inf, rhsValue) and [nextRhsValue, +inf)
                            else if(nextFcmp->op == ">=") {
                                if(rhsValue < nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == "<=") {
                            // (-inf, rhsValue] and (nextRhsValue, +inf)
                            if(nextFcmp->op == ">" || nextFcmp->op == ">=") {
                                if(rhsValue < nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == ">") {
                            // (-inf, nextRhsValue) and (rhsValue, +inf)
                            if(nextFcmp->op == "<") {
                                if(rhsValue >= nextRhsValue)
                                    canBeMerged = true;
                            }
                            // (-inf, nextRhsValue] and (rhsValue, +inf)
                            else if(nextFcmp->op == "<=") {
                                if(rhsValue > nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == ">=") {
                            // (-inf, nextRhsValue) and [rhsValue, +inf)
                            if(nextFcmp->op == "<" || nextFcmp->op == "<=") {
                                if(rhsValue > nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == "==") {
                            if(nextFcmp->op == "!=") {
                                if(rhsValue == nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == "!=") {
                            if(nextFcmp->op == "==") {
                                if(rhsValue == nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                    }

                    if(canBeMerged && nextBr) {
                        br->exp = Const::getConst(Type::getBool(), 0, "if_combined_1");
                        nextBr->exp = Const::getConst(Type::getBool(), 0, "if_combined_1");
                        visitedBB.insert(trueBB);
                        flag = true;
                    }
                }

                // enter falseBB, find union set of constRhs and nextConstRhs
                if(falseBB->instructions.size() > 1 && falseBB->instructions[falseBB->instructions.size() - 2]->type == Fcmp) {
                    auto nextFcmp = dynamic_cast<FcmpInstruction *>(falseBB->instructions[falseBB->instructions.size() - 2].get());
                    auto nextBr = dynamic_cast<BrInstruction *>(falseBB->instructions.back().get());
                    auto nextLhs = nextFcmp->a;
                    auto nextRhs = nextFcmp->b;
                    auto nextConstRhs = dynamic_cast<Const *>(nextRhs.get());
                    bool canBeMerged = false;

                    if(nextConstRhs && nextBr) {
                        float nextRhsValue = nextConstRhs->floatVal;
                        if(fcmp->op == "<") {
                            // (-inf, rhsValue) and (nextRhsValue, +inf)
                            if(nextFcmp->op == ">") {
                                if(rhsValue > nextRhsValue)
                                    canBeMerged = true;
                            }
                            // (-inf, rhsValue) and [nextRhsValue, +inf)
                            else if(nextFcmp->op == ">=") {
                                if(rhsValue >= nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == "<=") {
                            // (-inf, rhsValue] and (nextRhsValue, +inf)
                            if(nextFcmp->op == ">" || nextFcmp->op == ">=") {
                                if(rhsValue >= nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == ">") {
                            // (-inf, nextRhsValue) and (rhsValue, +inf)
                            if(nextFcmp->op == "<") {
                                if(rhsValue < nextRhsValue)
                                    canBeMerged = true;
                            }
                            // (-inf, nextRhsValue] and (rhsValue, +inf)
                            else if(nextFcmp->op == "<=") {
                                if(rhsValue <= nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == ">=") {
                            // (-inf, nextRhsValue) and [rhsValue, +inf)
                            if(nextFcmp->op == "<" || nextFcmp->op == "<=") {
                                if(rhsValue <= nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == "==") {
                            if(nextFcmp->op == "!=") {
                                if(rhsValue == nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                        else if(fcmp->op == "!=") {
                            if(nextFcmp->op == "==") {
                                if(rhsValue == nextRhsValue)
                                    canBeMerged = true;
                            }
                        }
                    }

                    if(canBeMerged && nextBr) {
                        br->exp = Const::getConst(Type::getBool(), 1, "if_combined_1");
                        nextBr->exp = Const::getConst(Type::getBool(), 1, "if_combined_1");
                        visitedBB.insert(falseBB);
                        flag = true;
                    }
                }
            }
        }
    }
    return flag;
}
