#include "loopGepCombine.h"
#include "LICM.h"

static unordered_map<GetElementPtrInstruction*, pair<int,int>> gepmap;

// i32* %a = getelementptr &([1 * i32]* %x)[i32 0][i32 %i1];
// i32 %i2 = add i32 %i1, i32 1;
// i32* %b = getelementptr &([1 * i32]* %x)[i32 0][i32 %i2];
// -->
// i32* %a = getelementptr &([1 * i32]* %x)[i32 0][i32 %i1];
// i32* %b = getelementptr &([1 * i32]* %a)[i32 1];

void runGEPOnLoop(LoopPtr loop, BlockPtr block)
{
    int maxTripCount = 80;

    auto legalLoop = [&](LoopPtr loop) -> bool {
        // 1. Loop can not have break. At our simplify loop continue become 
        // to the common for loop
        if(loop->getExitingBlocks().size() != 1 || loop->getExitBlocks().size() != 1 || loop->getSubLoops().size() || loop->basicBlocks.size() > 2)
        {
            // cerr << "exiting size is " << loop->getExitingBlocks().size() << " " <<  loop->getExitBlocks().size() <<endl;
            // error("111\n");
            return false;    
        } 
                                   // It must to be 1 at any time.
        auto trip = loop->getTripCount();
        if( trip)
        {
            // std::cerr << "333" <<std::endl;
            return false;
        }

        for(auto bb : loop->getBasicBlocks())
            for(auto instr : bb->instructions)
            {
                // instr->print();
                if(instr->type == Call)
                    return false;
            }

            
        // 3. Do not have any function call.
        auto phi = loop->indPhi;

        if(!loop->hasSCEV(phi))
        {
            return false;
        }
        auto& scev = loop->getSCEV(phi);
        auto endval = loop->indEnd;
        auto initval = scev.at(0);
        auto stepval = scev.at(1);
        if(endval && initval && initval ->isConst && stepval && stepval->isConst)
        {

            auto cInitial = dynamic_cast<Const *>(initval.get());
            auto cStep = dynamic_cast<Const *>(stepval.get());
            if(cInitial->intVal == 0 && cStep->intVal == 1)
            {
                return true;
            }
        }

        

        // for (auto instr : loop->getLatchBlock()->instructions)
        //     if (instr->type != Binary)
        //     {
        //         if (instr->type != Br)
        //             return false;
        //     }

        return false;
    };


    if(!legalLoop(loop))
    {
        // error("Not legal!\n");
        return;
    }
    else
    {
        // error("Legal!\n");
    }
    // printf("found legal loop!!\n");
    // fflush(stdout);
    auto phi = loop->indPhi;
    auto& scev = loop->getSCEV(phi);
    auto endval = loop->indEnd;
    auto initval = scev.at(0);
    auto stepval = scev.at(1);
    auto cInitial = dynamic_cast<Const *>(initval.get());
    auto cStep = dynamic_cast<Const *>(stepval.get());
    auto phireg = phi->reg;
    auto ptr = phireg->useHead;
    ValuePtr extreg = nullptr;
    // printf("get all variables!!, ptr = %p\n", ptr);
    // fflush(stdout);
    while(ptr)
    {
        auto user = ptr->user;
        if(user->type == Ext)
        {
            // printf("user->type == Ext\n");
            // fflush(stdout);
            auto extins = dynamic_cast<ExtInstruction*> (user);
            // extins->print();
            // fflush(stdout);
            auto extbb = extins->basicblock;
            if(loop->contains(extbb))
                extreg = extins->reg;
            break;
        }
        ptr = ptr -> next;
    }

    if(extreg == nullptr)
        return;

    vector<GetElementPtrInstruction*> geps;

    ptr = extreg->useHead;
    // printf("found extreg:");
    // fflush(stdout);
    // printf("%s\n", extreg->getStr().c_str());
    // fflush(stdout);
    while(ptr)
    {
        auto user = ptr->user;
        if(user->type == GEP)
        {
            auto gepins = dynamic_cast<GetElementPtrInstruction*> (user);
            geps.push_back(gepins);
        }
        ptr = ptr -> next;
    }

    // printf("found GEP in loopGEPCombine:\n");
    // fflush(stdout);
    for(auto gep: geps)
    {
        gepmap[gep] = {cInitial->intVal, cStep->intVal};
        auto loopGEP = shared_ptr<LoopGEPInstruction>(new LoopGEPInstruction(loop->getPreheader(), gep, extreg, cInitial->intVal, cStep->intVal));
        // printf("preheader of loop: %s\n", loop->getPreheader()->label->getStr().c_str());
        loop->getPreheader()->pushInstruction(vector<InstructionPtr>{loopGEP});
        // printf("init: %lld, step: %lld, initOffset: %d, stepOffset: %d, index: %s, from type: %d\n", 
        //     cInitial->intVal, cStep->intVal, loopGEP->getInitOffset(), loopGEP->getStepOffset(), loopGEP->index->getStr().c_str(), loopGEP->gep->from->type->ID);
        // gep->print();
        // fflush(stdout);
    }
}

void combineGEP(BlockPtr block)
{
    queue<LoopPtr> QueWorkList;
    for(auto loop : block->getAllLoops())
    {
        if(loop->getParentLoop()!=nullptr)
            continue;
        else
            QueWorkList.push(loop);
    }
    while(!QueWorkList.empty())
    {
        auto loop = QueWorkList.front();
        QueWorkList.pop();
        runGEPOnLoop(loop, block);
        for(auto subLoop : loop->getSubLoops())
            QueWorkList.push(subLoop);
    }
    
}
