#include "loopUnroll.h"
#include "LICM.h"

int copy_num =0;

void unuseall(InstructionPtr ins)
{
    auto operands_num = ins->getNumOperands();
    for(int i=0;i<operands_num;i++)
    {
        auto operand = ins->getOperand(i);
        if(operand)
        {
            deleteUser(operand);
        }
    }
}


void error(string s)
{
    cerr << s << endl;
}

//copy个Value
static VariablePtr copyVariable(VariablePtr old){
    //arr的inner就不要了，因为在局部变量中其会转换成gep和store指令，所以可以在后续获得
    //intval floatval也都不要了，其以store、load指令存在，variable其实不应该有这几个量存在
    // error("copyVariavle\n");
    VariablePtr ret = Variable::copy(old);
    // error("copyVariavle1\n");
    //清除use，这些use会在之后copy指令的过程中重新建立
    ret->name = ret->name+".copy";
    ret->useHead = nullptr;
    ret->numUses = 0;
    return ret;
}


static ValuePtr getNewOperand(ValuePtr old, unordered_map<ValuePtr, ValuePtr>&ValueMap){
    // error("entre getNewOperand\n");
    //const use same pointer
    if(old->isConst){
        // error("const\n");
        return old;
    }
    // error("getNewOperand1\n");
    if(ValueMap.find(old) == ValueMap.end()){
        //打个补丁，phi中可能会用到
        if(!dynamic_pointer_cast<Variable>(old)){
            ValueMap[old] =  ValuePtr(new Reg(old->type,old->name+".unroll.copy"));
            return ValueMap[old];
        }
        ValueMap[old] = copyVariable(dynamic_pointer_cast<Variable>(old));
    }
    // error("getNewOperand2\n");

    return ValueMap[old];
}



static InstructionPtr copyInstruction(InstructionPtr old, unordered_map<ValuePtr,ValuePtr>&ValueMap,
                                        unordered_map<BasicBlockPtr,BasicBlockPtr>&BBMap,
                                        unordered_map<LabelPtr,LabelPtr>&LabelMap, BlockPtr block){
    
    // error("enter CopyInstruction\n");
    if(old->type == Return){
        ReturnInstruction* RI = dynamic_cast<ReturnInstruction*>(old.get());

        auto ret =  InstructionPtr(new ReturnInstruction(getNewOperand(RI->retValue,ValueMap),nullptr));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Br){
        BrInstruction* RI = dynamic_cast<BrInstruction*>(old.get());
        if(RI->exp){
            //这里的label使用label获得，其实block里有labelbbmap，但实在是太麻烦了，就算了。
            // cerr << "Copyting Branch" << endl;
            if(LabelMap[RI->label1]==nullptr)
            {
                // cerr << "Label 1 null "<< endl;
            }
            if(LabelMap[RI->label2]==nullptr)
            {
                // cerr << "Label 2 null "<< endl;
            }
            auto ret =  InstructionPtr(new BrInstruction(getNewOperand(RI->exp,ValueMap),LabelMap[RI->label1],LabelMap[RI->label2],nullptr));
            //供后续阶段使用
            ValueMap[old->reg] = ret->reg;
            return ret;
        }
        else{
            auto ret =  InstructionPtr(new BrInstruction(LabelMap[RI->label1],nullptr));
            //供后续阶段使用
            ValueMap[old->reg] = ret->reg;
            return ret;
        }
    }
    else if(old->type == Alloca){
        AllocaInstruction* RI = dynamic_cast<AllocaInstruction*>(old.get());

        auto ret =  InstructionPtr(new AllocaInstruction(getNewOperand(RI->des,ValueMap),nullptr));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Store){
        StoreInstruction* RI = dynamic_cast<StoreInstruction*>(old.get());
        if(RI->gep){
            GetElementPtrInstruction* newGEP = dynamic_cast<GetElementPtrInstruction*>((ValueMap[RI->gep->reg])->I);
            auto ret =  InstructionPtr(new StoreInstruction(dynamic_pointer_cast<GetElementPtrInstruction>(newGEP->getSharedThis()), getNewOperand(RI->value,ValueMap),nullptr));
            //供后续阶段使用
            ValueMap[old->reg] = ret->reg;
            return ret;  
        }
        else{
            auto ret =  InstructionPtr(new StoreInstruction(getNewOperand(RI->des,ValueMap),getNewOperand(RI->value,ValueMap),nullptr));
            //供后续阶段使用
            ValueMap[old->reg] = ret->reg;
            return ret;   
        }
    }
    else if(old->type == Load){
        LoadInstruction* RI = dynamic_cast<LoadInstruction*>(old.get());
        if(RI->from->type->isPtr()){
            auto nreg = block->getReg(RI->from->type);
            auto ret =  InstructionPtr(new LoadInstruction(getNewOperand(RI->from, ValueMap), getNewOperand(nreg,ValueMap),nullptr));
            // auto ret =  InstructionPtr(new LoadInstruction(getNewOperand(RI->from, ValueMap), getNewOperand(RI->to,ValueMap),nullptr));
            //供后续阶段使用
            ValueMap[old->reg] = ret->reg;
            return ret;  
        }
        else{
            auto nreg = block->getReg(RI->from->type);
            auto ret =  InstructionPtr(new LoadInstruction(getNewOperand(RI->from, ValueMap), getNewOperand(nreg,ValueMap),nullptr));
            // auto ret =  InstructionPtr(new LoadInstruction(getNewOperand(RI->from,ValueMap),getNewOperand(RI->to,ValueMap),nullptr));
            //供后续阶段使用
            ValueMap[old->reg] = ret->reg;
            return ret;   
        }
    }
    else if(old->type == Call){
        CallInstruction* RI = dynamic_cast<CallInstruction*>(old.get());
        vector<ValuePtr> newArgv;
        for(int i = 0;i<RI->argv.size();i++){
            newArgv.push_back(getNewOperand(RI->argv[i],ValueMap));
        }
        auto ret =  InstructionPtr(new CallInstruction(RI->func, newArgv, nullptr));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Bitcast){
        BitCastInstruction* RI = dynamic_cast<BitCastInstruction*>(old.get());

        auto ret =  InstructionPtr(new BitCastInstruction(getNewOperand(RI->from, ValueMap), getNewOperand(RI->reg, ValueMap),nullptr, RI->toType));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Ext){
        ExtInstruction* RI = dynamic_cast<ExtInstruction*>(old.get());

        auto ret =  InstructionPtr(new ExtInstruction(getNewOperand(RI->from, ValueMap), RI->to,RI->isign,nullptr));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Sitofp){
        SitofpInstruction* RI = dynamic_cast<SitofpInstruction*>(old.get());

        auto ret =  InstructionPtr(new SitofpInstruction(getNewOperand(RI->from, ValueMap), nullptr));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Fptosi){
        FptosiInstruction* RI = dynamic_cast<FptosiInstruction*>(old.get());

        auto ret =  InstructionPtr(new FptosiInstruction(getNewOperand(RI->from, ValueMap), nullptr));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == GEP){
        GetElementPtrInstruction* RI = dynamic_cast<GetElementPtrInstruction*>(old.get());
        vector<ValuePtr> newIndex;
        for(auto oldInd:RI->index){
            newIndex.push_back(getNewOperand(oldInd,ValueMap));
        }
        auto newFrom = getNewOperand(RI->from,ValueMap);
       
        auto ret =  InstructionPtr(new GetElementPtrInstruction(getNewOperand(RI->from,ValueMap),newIndex,nullptr));
         if(RI->from->type->isPtr()){
            ret->reg->type = dynamic_cast<PtrType*>(RI->from->type.get())->inner;
        }
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Binary){
        // error("binary \n");
        BinaryInstruction* RI = dynamic_cast<BinaryInstruction*>(old.get());


        auto ret =  InstructionPtr(new BinaryInstruction(getNewOperand(RI->a, ValueMap),getNewOperand(RI->b, ValueMap),RI->op, BasicBlockPtr(nullptr)));
        //供后续阶段使用
        // error("binary2 \n");
        ValueMap[old->reg] = ret->reg;
        // error("binary3 \n");
        return ret;
    }
    else if(old->type == Fneg){
        FnegInstruction* RI = dynamic_cast<FnegInstruction*>(old.get());

        auto ret =  InstructionPtr(new FnegInstruction(getNewOperand(RI->a, ValueMap), nullptr));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Icmp){
        IcmpInstruction* RI = dynamic_cast<IcmpInstruction*>(old.get());

        auto ret =  InstructionPtr(new IcmpInstruction(nullptr, getNewOperand(RI->a, ValueMap), getNewOperand(RI->b, ValueMap),RI->op));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Fcmp){
        FcmpInstruction* RI = dynamic_cast<FcmpInstruction*>(old.get());

        auto ret =  InstructionPtr(new FcmpInstruction(nullptr, getNewOperand(RI->a, ValueMap), getNewOperand(RI->b, ValueMap),RI->op));
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else if(old->type == Phi){
        // error("Phi\n");
        PhiInstruction* RI = dynamic_cast<PhiInstruction*>(old.get());

        auto ret =  InstructionPtr(new PhiInstruction(nullptr, getNewOperand(RI->val, ValueMap)));
        //第二项参数实际上不需要了，后面直接能找到对应项，第二项参数留着会导致一个bug，即如果该phi是由后面用于替代return产生的，那么这个val就是call
        //而call不是变量，copy时出现错误
        // auto ret =  InstructionPtr(new PhiInstruction(nullptr, RI->val));
        // error("Phi4\n");
        //供后续阶段使用
        ValueMap[old->reg] = ret->reg;
        return ret;
    }
    else{
        assert(false&&"unknown Inst type");
    }
}



LoopPtr getLoopCopy(LoopPtr loop, std::unordered_map<ValuePtr, ValuePtr>& LoopValueCpy, unordered_map<ValuePtr, ValuePtr>& CurrentIncomingValue, BasicBlockPtr exitblock, BlockPtr block)
{
    // cerr << "Entering Loop copy" << endl;
    // auto headerCpy = BasicBlock::Create(loop->getHeader()->getName() + ".unrollcpy", loop->getParent());
    string now_num = to_string(copy_num++);
    LabelPtr BBLabel = LabelPtr(new Label(loop->getHeader()->label->name+".copy"+now_num));

    auto headerCpy = BasicBlockPtr(new BasicBlock(BBLabel));

    LabelPtr latchLabel = LabelPtr(new Label(loop->latchBlock->label->name+".copy"+now_num));

    auto latchCpy = (loop->getHeader() == loop->latchBlock) ? headerCpy : BasicBlockPtr(new BasicBlock(latchLabel));
    
    auto loopCpy = LoopPtr(new Loop(headerCpy,1));

    loopCpy->setLatchBlock(latchCpy);

    std::unordered_map<PhiInstruction*, PhiInstruction*> queuePhi;

    unordered_map<ValuePtr,ValuePtr> ValueMap;
    unordered_map<BasicBlockPtr, BasicBlockPtr> loopBBMap;
    unordered_map<LabelPtr, LabelPtr> LabelMap;
    unordered_map<InstructionPtr, InstructionPtr> insMap;

    unordered_map<ValuePtr, ValuePtr> mycopyMap;

    for(auto bb : loop->getBasicBlocks())
        if(bb == loop->getHeader())
        {
            loopBBMap[bb] = headerCpy;
            LabelMap[bb->label] = headerCpy->label;
            // LoopValueCpy[bb] = headerCpy;
        }
        else if(bb == loop->getLatchBlock())
        {
            loopBBMap[bb] = latchCpy;
            LabelMap[bb->label] = latchCpy->label;
            // LoopValueCpy[bb] = latchCpy;
            loopCpy->addBasicBlock(latchCpy);
        }
        else if(bb != loop->getHeader() && bb != loop->getLatchBlock())
        {
            LabelPtr BBLabel = LabelPtr(new Label(bb->label->name+".unrollcpy"+now_num));
            auto loopBodyCpy = BasicBlockPtr(new BasicBlock(BBLabel));
            loopBBMap[bb] = loopBodyCpy;
            LabelMap[bb->label] = BBLabel;
            // LoopValueCpy[bb] = loopBodyCpy;
            loopCpy->addBasicBlock(loopBodyCpy);

        }

    LabelMap[exitblock->label] = exitblock->label; 

    for(auto bb : loopCpy->getBasicBlocks())
    {
        // cerr << bb->label->getStr() << endl;
    }
    
    // cerr << "Loop copy stage 1" << endl;

    for(auto bb : loop->getBasicBlocks())
    {
        if(bb == loop->getHeader())
        {
            // cerr << "Loop stage 2 header" << endl;
            for(auto instr : bb->instructions)
            {
                if(instr->type == Phi)
                    continue;
                // cerr << "BB copying" << endl;
                // if(instr->reg==nullptr && instr->type == Icmp)
                // {
                //     cerr << "Null on ICMP" << endl;
                // }
                // assert(instr!=nullptr && instr->reg !=nullptr);
                // cerr << "copy instr " << instr->getName() << endl;
                auto instrCpy = copyInstruction(instr, CurrentIncomingValue, loopBBMap, LabelMap, block);
                // cerr << "copied instr" << endl;
                // LoopValueCpy[instr] = instrCpy;
                headerCpy->pushInstruction(instrCpy);
                instrCpy->basicblock = headerCpy;
            }
        }
        else if(bb == loop->getLatchBlock())
        {
            // cerr << "Loop stage 2 latch" << endl;
            for(auto instr : bb->instructions)
            {
                auto instrCpy = copyInstruction(instr, CurrentIncomingValue, loopBBMap, LabelMap, block);
                // LoopValueCpy[instr] = instrCpy;
                latchCpy->pushInstruction(instrCpy);
                instrCpy->basicblock = latchCpy;
                if(instrCpy ->type ==Binary && instr -> type ==Binary)
                {
                    auto bi = dynamic_cast<BinaryInstruction*>(instr.get());
                    auto bi_cpy = dynamic_cast<BinaryInstruction*>(instrCpy.get());
                    auto v1 = bi->a;
                    auto v2 = bi->b;
                    mycopyMap[bi->reg] = bi_cpy->reg;
                    
                }

                if(instrCpy ->type ==Load && instr -> type ==Load)
                {
                    auto ldr = dynamic_cast<LoadInstruction*>(instr.get());
                    auto ldr_cpy = dynamic_cast<LoadInstruction*>(instrCpy.get());

                    mycopyMap[ldr->reg] = ldr_cpy->reg;
                    
                }


                if(instr->type==Phi)
                {
                    auto phi = dynamic_cast<PhiInstruction*>(instr.get());
                    if(instrCpy->type==Phi)
                    {
                        auto phiCpy = dynamic_cast<PhiInstruction*>(instrCpy.get());
                        queuePhi[phi] = phiCpy;
                    }
                        
                }
                    
            }
        }
        else
        {
            // cerr << "Loop stage 2 else" << endl;
            for(auto instr : bb->instructions)
            {
                auto instrCpy = copyInstruction(instr, CurrentIncomingValue, loopBBMap, LabelMap, block);
                // LoopValueCpy[instr] = instrCpy;
                auto loopBodyCpy = loopBBMap[bb];
                loopBodyCpy->pushInstruction(instrCpy);
                instrCpy->basicblock = loopBodyCpy;
                if(instrCpy ->type ==Binary && instr -> type ==Binary)
                {
                    auto bi = dynamic_cast<BinaryInstruction*>(instr.get());
                    auto bi_cpy = dynamic_cast<BinaryInstruction*>(instr.get());
                    auto v1 = bi->a;
                    auto v2 = bi->b;
                    mycopyMap[bi->reg] = bi_cpy->reg;
                   
                }

                if(instrCpy ->type ==Load && instr -> type ==Load)
                {
                    auto ldr = dynamic_cast<LoadInstruction*>(instr.get());
                    auto ldr_cpy = dynamic_cast<LoadInstruction*>(instrCpy.get());

                    mycopyMap[ldr->reg] = ldr_cpy->reg;
                    
                }

                if(instr->type==Phi)
                {
                    auto phi = dynamic_cast<PhiInstruction*>(instr.get());
                    if(instrCpy->type ==Phi)
                    {
                        auto phiCpy = dynamic_cast<PhiInstruction*>(instrCpy.get());
                        queuePhi[phi] = phiCpy;
                    }
                        
                }
                    
            }
        }
    }

    // cerr << "Loop copy stage 2" << endl;
    // if(copy_num==2)
    // for(auto instr: latchCpy->instructions)
    // {
    //     cerr << instr->getName() << endl;
    // }

    for(auto iter : loopBBMap)
    {
        auto bb = iter.first;
        auto bbCpy = iter.second;
        // dumpBasickBlock(bbCpy);
        if(bb != loop->getHeader())
            for(auto pred : bb->predBasicBlocks)
                bbCpy->predBasicBlocks.insert(loopBBMap[pred]);
        if(bb != loop->getLatchBlock())
            for(auto succ : bb->succBasicBlocks)
                bbCpy->succBasicBlocks.insert(loopBBMap[succ]);
        // dumpBasickBlock(bbCpy);
    }

    // cerr << "Loop copy stage 3" << endl;

    for(auto iter : queuePhi)
    {
        auto phi = iter.first;
        auto phiCpy = iter.second;
        for(auto iter : phi->from)
        {
            auto val = iter.first;
            auto pred = iter.second;
            // phiCpy->replaceValue
            phiCpy->addFrom(val, loopBBMap[pred]);
        }
    }

    // cerr << "finding phis" << endl;

    std::unordered_map<PhiInstruction*, ValuePtr> phiAssignLater;
    for(auto instr : loop->getHeader()->instructions)
    {
        if(instr->type==Phi)
        {
            auto phi = dynamic_cast<PhiInstruction*>(instr.get());
            for(auto iter : phi->from)
            {
                auto val = iter.first;
                auto pred = iter.second;
                if(pred == loop->getLatchBlock())
                    phiAssignLater[phi] = val;
            }
                
        }
        else
            break;
    }

    // cerr << "Finish finding phis" << endl;

    for(auto iter : phiAssignLater)
    {
        auto phi = iter.first;
        auto val = iter.second;
        if(mycopyMap.find(val) == mycopyMap.end())
        {
            cout<< "Cannot unroll loop!!!" << endl;
        }
        assert(mycopyMap.find(val)!=mycopyMap.end());
        CurrentIncomingValue[phi->reg] = mycopyMap[val];
        // cerr << "Replace! " << phi->reg->getStr() << " -> " << mycopyMap[val]->getStr() << endl; 
    }

    // cerr << "Finish assign phis" << endl;
        
    return loopCpy;
}


void runOnLoop(LoopPtr loop, BlockPtr block)
{
    int maxTripCount = 80;
    copy_num = 0;

    auto legalLoop = [&](LoopPtr loop) -> bool {
        // 1. Loop can not have break. At our simplify loop continue become 
        // to the common for loop
        if(loop->getExitingBlocks().size() != 1 || loop->getExitBlocks().size() != 1 || loop->getSubLoops().size())
        {
            // cerr << "exiting size is " << loop->getExitingBlocks().size() << " " <<  loop->getExitBlocks().size() <<endl;
            // error("111\n");
            return false;    
        } 
                                   // It must to be 1 at any time.
        auto trip = loop->getTripCount();
        if( trip<=0)
        {
            // std::cerr << "333" <<std::endl;
            return false;
        }
            
        else if(trip >= 2000)
        {
            // std::cerr << "444" <<std::endl;
            return false;
        }
        // 3. Do not have any function call.
        for(auto bb : loop->getBasicBlocks())
            for(auto instr : bb->instructions)
                if(instr->type == Call)
                    return false;
        
        for(auto bb : loop->getBasicBlocks())
            for(auto instr : bb->instructions)
                if(instr->type == Icmp && bb != loop->header)
                    return false;

        // for (auto instr : loop->getLatchBlock()->instructions)
        //     if (instr->type != Binary)
        //     {
        //         if (instr->type != Br)
        //             return false;
        //     }

        return true;
    };


    if(!legalLoop(loop))
    {
        // error("Not legal!\n");
        return;
    }
    else
    {
        // error("Legal!\n");
    }

    // error("unpass init of header\n");
    // cerr << "unpass init of header" << endl;
        
    
    auto header = loop->getHeader();

    // error("[passed] 1 init of header\n");
    auto latch = loop->latchBlock;
    // error("[passed] 2 init of header\n");

    // std::cerr << "header is " << header->label->getStr() << endl;
    // std::cerr << "latch is " << latch->label->getStr() << endl;
    auto exit = *loop->getExitBlocks().begin();


    // prelook

    unordered_set<ValuePtr> phiset;

    bool phiflag =  false;
    for(auto instr : loop->getHeader()->instructions)
    {
        if(instr->type==Phi)
        {
            auto phi = dynamic_cast<PhiInstruction*>(instr.get());
            for(auto iter : phi->from)
            {
                auto val = iter.first;
                auto pred = iter.second;
                if(pred == loop->getLatchBlock())
                    phiset.insert(val);
            }
                
        }
        else
            break;
    }

    for(auto instr : loop->getHeader()->instructions)
    {
        if(instr->type==Phi)
        {
            auto phi = dynamic_cast<PhiInstruction*>(instr.get());
            if(phiset.count(phi->reg))
            {
                phiflag = true;
                break;
            }

                
        }
        else
            break;
    }

    if(phiflag)
    {
        cout<< "Cannot Unroll due to phi loop" << endl;
        return;
    }
        


    // PrintMsg("header");
    // dumpBasickBlock(header);
    // PrintMsg("latch");
    // dumpBasickBlock(latch);
    // PrintMsg("exit");
    // dumpBasickBlock(exit);

    //    |-----                a <- a0
    //  a <- phi(a0, a1)        header
    //  header |                  |  
    //    |    |  -->           a1 <- ..
    //  a0 <- ...               latch
    //  latch  |                  |
    //    |    |                a2 <- a1
    //  exit ---                header'
    //                            |
    //                          exit

    // cerr << "[passed] init of header" << endl;

    std::unordered_map<ValuePtr, ValuePtr> CurrentIncomingValue;
    std::vector<InstructionPtr> headerPhi;

    std::unordered_map<ValuePtr, ValuePtr> tmpmap;

    int itrtimes = 0;

    ValuePtr tracephi = nullptr;
    for(auto instr : header->instructions)
    {
        // cerr << "now time is " << itrtimes++ << endl; 
        int phitimes = 0;
        if(instr->type==Phi)
        {
            auto phi  = (PhiInstruction*)(instr.get());
            headerPhi.push_back(instr);
            // cerr << "phi is " << phi->reg->getStr() << "Use is "<< phi->reg->getNumUses() << endl;
            if(phi->reg->getStr()=="%phi2")
                tracephi = phi->reg;
            for(auto it : phi->from)
            {
                // cerr << "phy num is " << phitimes++ << endl; 
                auto val = it.first;
                auto pred = it.second;
                // cerr << val->getStr()<<endl;
                // cerr << pred->label->getStr() << endl;
                // cerr << latch->label->getStr() << endl;
                if(pred == nullptr)
                    continue;
                if(pred == latch)
                {
                    // cerr << "Yes!" << endl;
                    // CurrentIncomingValue[phi->reg] = getNewOperand(val, tmpmap);
                    CurrentIncomingValue[phi->reg] = val;
                    // cerr << "No!" << endl;
                }
                    
            }
        }
        else
            break;
    }
    
    for(auto instr: latch->instructions)
    {
        if(instr->type ==Binary)
        {
            auto bi = dynamic_cast<BinaryInstruction*>(instr.get());
            // if(bi->op!='+' || bi->op !='-')
            //     continue;
            auto reg = bi -> reg;
            if(CurrentIncomingValue.find(reg)==CurrentIncomingValue.end())
            {
                CurrentIncomingValue[reg] = reg;
            }
            if(bi->a && CurrentIncomingValue.find(bi->a)==CurrentIncomingValue.end() && !bi->a->isConst)
            {
                // cerr << "debug is " << bi->a->getStr() << endl;
                CurrentIncomingValue[bi->a] = bi->a;
            }
            if(bi->b && CurrentIncomingValue.find(bi->b)==CurrentIncomingValue.end() && !bi->b->isConst)
            {
                CurrentIncomingValue[bi->b] = bi->b;
            }
        }
        else if(instr->type == GEP)
        {
            auto gep = dynamic_cast<GetElementPtrInstruction*>(instr.get());
            auto reg = gep->reg;
            if(CurrentIncomingValue.find(reg) == CurrentIncomingValue.end())
            {
                CurrentIncomingValue[reg] = reg;
            }
            if(gep->from && CurrentIncomingValue.find(gep->from) ==CurrentIncomingValue.end())
            {
                CurrentIncomingValue[gep->from] = gep->from;
            }
            for(auto idx: gep->index)
            {
                if (idx && CurrentIncomingValue.find(idx) == CurrentIncomingValue.end())
                {
                    CurrentIncomingValue[idx] = idx;
                }
            }
        }
        else if(instr->type == Store)
        {
            auto str = dynamic_cast<StoreInstruction*>(instr.get());
            auto reg = str->des;
            if(CurrentIncomingValue.find(reg) == CurrentIncomingValue.end())
            {
                CurrentIncomingValue[reg] = reg;
            }
            if(str->value && CurrentIncomingValue.find(str->value) ==CurrentIncomingValue.end())
            {
                CurrentIncomingValue[str->value] = str->value;
            }
        }
        else if(instr->type == Load)
        {
            auto ldr = dynamic_cast<LoadInstruction*>(instr.get());
            auto reg = ldr->from;
            if(CurrentIncomingValue.find(reg) == CurrentIncomingValue.end())
            {
                CurrentIncomingValue[reg] = reg;
            }
            // if(str->value && CurrentIncomingValue.find(str->value) ==CurrentIncomingValue.end())
            // {
            //     CurrentIncomingValue[str->value] = str->value;
            // }
        }
        else if(instr->type == Ext)
        {
            auto extins =dynamic_cast<ExtInstruction*>(instr.get());
            auto reg = extins->from;
            if(CurrentIncomingValue.find(reg) == CurrentIncomingValue.end())
            {
                CurrentIncomingValue[reg] = reg;
            }

        }
        else if(instr-> type == Sitofp)
        {
            auto ins =dynamic_cast<SitofpInstruction*>(instr.get());
            auto reg = ins->from;
            if(CurrentIncomingValue.find(reg) == CurrentIncomingValue.end())
            {
                CurrentIncomingValue[reg] = reg;
            }
        }
        else if(instr-> type == Fptosi)
        {
            auto ins =dynamic_cast<FptosiInstruction*>(instr.get());
            auto reg = ins->from;
            if(CurrentIncomingValue.find(reg) == CurrentIncomingValue.end())
            {
                CurrentIncomingValue[reg] = reg;
            }
        }
        else if(instr-> type == Fneg)
        {
            auto ins =dynamic_cast<FnegInstruction*>(instr.get());
            auto reg = ins->a;
            if(CurrentIncomingValue.find(reg) == CurrentIncomingValue.end())
            {
                CurrentIncomingValue[reg] = reg;
            }
        }
        else if(instr -> type == Br)
        {

        }
        else if(instr-> type == Bitcast)
        {
            auto ins =dynamic_cast<BitCastInstruction*>(instr.get());
            auto reg = ins->from;
            if(CurrentIncomingValue.find(reg) == CurrentIncomingValue.end())
            {
                CurrentIncomingValue[reg] = reg;
            }
        }
        else{
            instr->print();
            fflush(stdout);
            assert(false && "Loop unroll not handle instr");
        }
        
    }


    auto itr = --latch->instructions.end();
    auto latchCondBr =  *itr;
    latch->instructions.erase(itr);
    latch->endInstruction = nullptr;
    unuseall(latchCondBr);
    latch->succBasicBlocks.clear();

    // cerr << "phi is " << tracephi->getStr() << "Use is "<< tracephi->getNumUses() << endl;

    std::vector<BasicBlockPtr> latchHeaderEdge;
    BasicBlockPtr preLatch = latch;
    BasicBlockPtr nowHeader = nullptr;

    auto trip = loop->getTripCount();


    // cerr << "Loop trip count is " << trip << endl;


    std::unordered_map<ValuePtr, ValuePtr> copyMap;
    latchHeaderEdge.push_back(preLatch);
    for(int i = 1; i < trip; i++)
    {
        // cerr << "now copying " << i << endl;
        auto loopCpy = getLoopCopy(loop, copyMap, CurrentIncomingValue,exit, block);

        // PrintMsg("Loop copy " + std::to_string(i));
        // for(auto bb : loopCpy->getLoopBasicBlocks())
        //     dumpBasickBlock(bb);

        auto headerCpy = loopCpy->getHeader();
        auto latchCpy = loopCpy->getLatchBlock();

        nowHeader = headerCpy;
        latchHeaderEdge.push_back(latchCpy);
        preLatch = latchCpy;
    }

    // cerr << "phi is " << tracephi->getStr() << "Use is "<< tracephi->getNumUses() << endl;

    // cerr << "[passed] at copy" << endl;


    latchHeaderEdge.push_back(exit);

    for(auto instr : exit->instructions)
    {
        if(instr->type == Binary)
        {
            auto bi = dynamic_cast<BinaryInstruction*>(instr.get());
            if(bi->a&&bi->a->I && bi->a->I->type == Phi)
            {   
                ValuePtr incomingVal = nullptr;
                auto phi = dynamic_cast<PhiInstruction*>(bi->a->I);
                bool flag = false;
                for (auto iter : phi->from)
                {
                    auto val = iter.first;
                    auto pred = iter.second;
                    if (pred == latch)
                        flag = true;
                    
                    
                }
                if(!latch || !flag)
                    continue;
                // phi->removeIncomingByBB(latch);
                // cerr << phi->reg->getStr() << " " << CurrentIncomingValue[phi->reg]->getStr() << endl;
                bi->setOperands(0,CurrentIncomingValue[phi->reg]);
                // replaceVarByVar(bi->a, CurrentIncomingValue[phi->reg]);
                // phi->addFrom(CurrentIncomingValue[phi->reg], preLatch);
            }

            if(bi->b&&bi->b->I && bi->b->I->type ==Phi)
            {   
                ValuePtr incomingVal = nullptr;
                auto phi = dynamic_cast<PhiInstruction*>(bi->b->I);
                bool flag = false;
                for (auto iter : phi->from)
                {
                    auto val = iter.first;
                    auto pred = iter.second;
                    if (pred == latch)
                        flag = true;
                }
                if(!latch || !flag)
                    continue;
                // phi->removeIncomingByBB(latch);
                // phi->addFrom(CurrentIncomingValue[phi->reg], preLatch);
                bi->setOperands(1,CurrentIncomingValue[phi->reg]);
                // replaceVarByVar(bi->b, CurrentIncomingValue[phi->reg]);
            }

            
        }
        else
            continue;
    }

    // cerr << "phi is " << tracephi->getStr() << "Use is "<< tracephi->getNumUses() << endl;

    // cerr << "[passed] at exitloop" << endl;

    for(auto instr : headerPhi)
    {
        auto phi  = (PhiInstruction*)(instr.get());
        // cerr << "instr name is " << phi->reg->getStr() << endl;
        for(auto iter : phi->from)
        {
            auto val = iter.first;
            auto pred = iter.second;
            if(pred == loop->preHeader)
            {
                // cerr << "in preheader" << endl;
                // cerr << "need to replace " << val->getStr() << endl;
                // cerr << "Use num is" << phi->reg->getNumUses() << endl;
                auto ptr = phi->reg->useHead;
                while(ptr)
                {
                    auto tmp =ptr->next;
                    auto user = ptr->user;
                    if(!loop->contains(user->basicblock))
                    {
                        if(!user->replaceValue(phi->reg,CurrentIncomingValue[phi->reg]))
                        {
                            cerr << user->reg->name << " error " << user->type << endl;
                        }
                        ptr->rmUse();
                        newUse(CurrentIncomingValue[phi->reg].get(),user);
                    }
                    ptr = tmp;
                }

                replaceVarByVar(phi->reg, val);
                // cerr << "[passed] replace var by var" << endl;
                
                unuseall(instr);
                // cerr << "[passed] unuseall" << endl;
                // auto mybb = phi -> basicblock;
                // mybb->removeInsturction(InstructionPtr(phi));
                phi->basicblock->removeInsturction(instr);
                // cerr << "[passed] remove" << endl;
                break;
            }
        }
    }

    // cerr << "[passed] at headerphi" << endl;

    // cerr << "[passed] at headerphi" << endl;

    // cerr << "[passed] at erase exit pred blocks" << endl;

    int iter_num = 0;

    auto block_iter = block->basicBlocks.begin();

    for(auto bb: block->basicBlocks)
    {
        block_iter++;
        if(bb == latch)
        {
            break;
        }
        // block_iter++;
    }

    vector<BasicBlockPtr> ins_list;

    for(auto iter : latchHeaderEdge)
    {
        iter_num ++;

        auto latch_cp = iter;
        if(iter_num==latchHeaderEdge.size())
        {
            break;
        }
        auto nextLatch = latchHeaderEdge[iter_num];
        latch_cp->succBasicBlocks.insert(nextLatch);
        nextLatch->predBasicBlocks.clear();
        nextLatch->predBasicBlocks.insert(latch_cp);
        
        auto br = InstructionPtr( new BrInstruction(nextLatch->label, latch_cp));
        latch_cp->pushInstruction(br);

        if(!loop->contains(latch_cp)&&latch_cp) {
            loop->addBasicBlock(latch_cp);
            ins_list.push_back(latch_cp);
            // block->pushBasicBlock(latch_cp);
            // auto ll = *block_iter;
            // cerr << "Now inserting" << endl;
            // cerr << ll->label->getStr() << " " << latch_cp->label->getStr() << endl;
            // block->basicBlocks.emplace(block_iter,latch_cp);
            // block_iter++;
        }
    }

    block->basicBlocks.insert(block_iter,ins_list.begin(),ins_list.end());


    itr = --header->instructions.end();
    header->instructions.erase(itr);
    header->endInstruction = nullptr;
    auto br = InstructionPtr( new BrInstruction(latch->label, header));
    header->pushInstruction(br);

    // cerr << "[passed] at latchHeaderEdge" << endl;

    for(auto iter: header->predBasicBlocks)
    {
        if(iter == latch)
        {
            header->predBasicBlocks.erase(iter);
            break;
        }
    }

    for(auto iter: exit->predBasicBlocks)
    {
        if(iter == header)
        {
            exit->predBasicBlocks.erase(iter);
            break;
        }
    }

    unordered_set<BasicBlockPtr> visited;

    for(auto bb: block->basicBlocks)
    {
        visited.insert(bb);
        // for(auto iter: bb->getPredecessor())
        // {
            
        // }
    }

    for(auto bb: block->basicBlocks)
    {
        // visited.insert(bb);
        for(auto iter: bb->getPredecessor())
        {
            if(!visited.count(iter))
                bb->predBasicBlocks.erase(iter);
        }
    }

    // cerr << "[passed] at erase header pred blocks" << endl;
}

// @FIXME: bug when compiling 09_BFS and almost useless
void unrollLoop(BlockPtr block) {

    queue<LoopPtr> QueWorkList;
    for(auto loop : block->getAllLoops())
    {
        if(loop->getParentLoop()!=nullptr)
            continue;
        else
            QueWorkList.push(loop);
    }
    while(!QueWorkList.empty())
    {
        auto loop = QueWorkList.front();
        QueWorkList.pop();
        runOnLoop(loop, block);
        for(auto subLoop : loop->getSubLoops())
            QueWorkList.push(subLoop);
    }
    
}