#include "strengthReduction.h"

bool isExp(int n) {
    if (n & 1) return false;
    int bitCount = 0;
    for(int i = 0; i < 31; i ++) 
    {
        n >>= 1;
        bitCount += (n & 1);
    }
    return (bitCount == 1);
}

void strengthReduction(BlockPtr block) {
    int regCount = 0;
    for (auto &bb : block->basicBlocks)
    {
        unordered_set<InstructionPtr> toDelete;
        vector<InstructionPtr> toSave;
        for (int i = 0; i < bb->instructions.size(); i ++)
        {
            auto &ins = bb->instructions[i];
            if (ins->type == InsID::Binary)
            {
                auto binary = dynamic_cast<BinaryInstruction *>(ins.get());
                auto lhs = binary->a;
                auto rhs = binary->b;
                if (binary->op == '*') 
                {
                    // dst = 2^shamt * rhs
                    if (lhs->isConst && rhs->type->ID == IntID) 
                    {
                        auto const_lhs = dynamic_cast<Const *>(lhs.get());
                        if (const_lhs->intVal > 0 && isExp(const_lhs->intVal))
                        {
                            auto shamt = Const::getConst(Type::getInt(),int(log2(const_lhs->intVal)), "strengthReduction%" + to_string(regCount++));
                            // op ',' indicates arithmetic left shift('<<')
                            auto new_binary = new BinaryInstruction(rhs, shamt, ',', ins->basicblock);
                            replaceVarByVar(binary->reg, new_binary->reg);
                            bb->instructions[i] = shared_ptr<Instruction>(new_binary);
                        }
                    }
                    // dst = lhs * constant
                    else if (rhs->isConst && lhs->type->ID == IntID) 
                    {
                        auto const_rhs = dynamic_cast<Const *>(rhs.get());
                        // dst = lhs * constant / constant
                        if(i + 1 < bb->instructions.size() && bb->instructions[i + 1]->type == Binary) {
                            auto nextBinary = dynamic_cast<BinaryInstruction *>(bb->instructions[i + 1].get());
                            if(nextBinary->op == '/' && nextBinary->b == rhs) {
                                replaceVarByVar(nextBinary->reg, binary->a);
                                toDelete.insert(bb->instructions[i]);
                                toDelete.insert(bb->instructions[i + 1]);
                                i ++;
                            }
                        }
                        // dst = lhs * 2^shamt
                        if (const_rhs->intVal > 0 && isExp(const_rhs->intVal))
                        {
                            auto shamt = Const::getConst(Type::getInt(),int(log2(const_rhs->intVal)), "strengthReduction%" + to_string(regCount++));
                            // op ',' indicates arithmetic left shift('<<')
                            auto new_binary = new BinaryInstruction(lhs, shamt, ',', ins->basicblock);
                            replaceVarByVar(binary->reg, new_binary->reg);
                            bb->instructions[i] = shared_ptr<Instruction>(new_binary);
                        }
                        // dst = lhs * (2^shamt1 + 2^shamt2) = (lhs << shamt1) + (lhs << shamt2)
                        else {}
                    }
                }
            }
        }
        if(toDelete.size()) {
            for(auto instr: bb->instructions)
            {
                if(toDelete.count(instr))
                    continue;
                toSave.push_back(instr);
            }

            bb->instructions = toSave;
            // for(auto instr: toDelete) {
            //     bb->removeInsturction(instr);
            // }
        }
    }
    // cout << "[frontend opt]: strengthReduction finished" << endl;
}
