#include "tailCallElimination.h"


// #define TailCallEliminationBug

#ifdef TailCallEliminationBug
#define error(...) cerr<<__VA_ARGS__
#else
#define error(...)
#endif


static bool isRecursiveCall(InstructionPtr call, FunctionPtr func){
    if(call->type != Call){
        return false;
    }
    return dynamic_cast<CallInstruction*>(call.get())->func == func;
}

//获得该BB中最后一个递归调用
static InstructionPtr getLastRecurCall(BasicBlockPtr bb, FunctionPtr func){
    for(auto it = bb->instructions.rbegin();it!=bb->instructions.rend();it++){
        if(isRecursiveCall(*it, func)){
            return *it;
        }
    }
    return nullptr;
}

static ValuePtr getIdenity(unsigned id, TypePtr type){
    if(id == Add || id == Xor){
        return Const::getConst(type,0);
    }
    else if(id == Mul){
        return Const::getConst(type, 1);
    }
    else if(id == FAdd){
        return Const::getConst(type,float(-0.0));
    }
    else if(id == FMul){
        return Const::getConst(type,float(-0.0));
    }
    assert(false&&"unknown type");
    return nullptr;
}

void eliminateCall(InstructionPtr I, FunctionPtr func, BasicBlockPtr BB){
    error("eliminateCall1\n");
    auto retIns = BB->instructions.back();

    InstructionPtr accuIns = nullptr;
    auto it = BB->instructions.begin();

    for(;it!=BB->instructions.end();it++){
        if(*it == I){
            break;
        }
    }
    error("eliminateCall2\n");
    assert(it!=BB->instructions.end()&&"call Ins not in BB");
    it++;
    for(;it != BB->instructions.end();it++){
        if(*it == BB->instructions.back()){
            break;
        }
        if(accuIns){
            return;
        }
        error("eliminateCall21\n");
        if((*it)->type==Binary&&(dynamic_cast<BinaryInstruction*>((*it).get())->getOpcode()==Add||
                                dynamic_cast<BinaryInstruction*>((*it).get())->getOpcode()==Xor||
                                // dynamic_cast<BinaryInstruction*>((*it).get())->getOpcode()==FAdd||
                                dynamic_cast<BinaryInstruction*>((*it).get())->getOpcode()==Mul)){
            accuIns = *it;
        }
        else{
            return;
        }
        error("eliminateCall22\n");
        if(((*it)->getOperand(0)!=I->reg&&(*it)->getOperand(1) != I->reg)||
            ((*it)->getOperand(0)==I->reg&&(*it)->getOperand(1) == I->reg)){
                return;
        }
        error("eliminateCall23\n");
        error(accuIns->reg->numUses<<endl);
        if(!(accuIns->reg->hasOneUse()&&accuIns->reg->useHead->user->type == Return)){
            return ;
        }
        error("eliminateCall24\n");
    }
    error("eliminateCall3\n");

    auto loopHeader = func->block->basicBlocks[0];
    auto newEntryLabel = LabelPtr(new Label("newEntry"));
    auto newEntry = BasicBlockPtr(new BasicBlock(newEntryLabel));

    func->block->basicBlocks.emplace(func->block->basicBlocks.begin(),newEntry);
    newEntry->belongfunc = func;
    vector<InstructionPtr> arguments;


    auto branchIns = InstructionPtr(new BrInstruction(loopHeader->label, newEntry));
    
    newEntry->instructions.push_back(branchIns);
    for(auto arg:func->formArguments){
        auto phiIns = InstructionPtr(new PhiInstruction(loopHeader, arg));
        loopHeader->insertInstruction(phiIns, loopHeader->instructions[0]);
        replaceVarByVar(arg, phiIns->reg);
        dynamic_cast<PhiInstruction*>(phiIns.get())->addFrom(arg, newEntry);
        arguments.push_back(phiIns);
    }
    
    
    InstructionPtr retValue = nullptr;
    InstructionPtr retValid = nullptr;
    // loopHeader->print();
    if(!(I->reg->type->isVoid())){
        retValue = InstructionPtr(new PhiInstruction(loopHeader, I->reg));
        loopHeader->insertInstruction(retValue, loopHeader->instructions[0]);
        dynamic_cast<PhiInstruction*>(retValue.get())->addFrom(Const::getConst(retValue->reg->type, 0),newEntry);
        retValid = InstructionPtr(new PhiInstruction(loopHeader, Type::getBool()));
        loopHeader->insertInstruction(retValid, loopHeader->instructions[0]);
        dynamic_cast<PhiInstruction*>(retValid.get())->addFrom(Const::getConst(Type::getBool(),false),newEntry);
    }
    // loopHeader->print();

    error("eliminateCall4\n");
    unsigned int idx = 0;
    for(auto arg:dynamic_cast<CallInstruction*>(I.get())->argv){
        dynamic_cast<PhiInstruction*>(arguments[idx++].get())->addFrom(arg, BB);
    }

    InstructionPtr accValue = nullptr;
    if(accuIns){
        accValue = InstructionPtr(new PhiInstruction(loopHeader, accuIns->reg->type));
        loopHeader->insertInstruction(accValue, loopHeader->instructions[0]);
        auto identity = getIdenity(dynamic_cast<BinaryInstruction*>(accuIns.get())->getOpcode(), accuIns->reg->type);
        dynamic_cast<PhiInstruction*>(accValue.get())->addFrom(identity, newEntry);
        I->reg->useHead->user = accValue.get();
        if(accuIns->getOperand(0)==I->reg){
            dynamic_cast<BinaryInstruction*>(accuIns.get())->setOperands(0, accValue->reg);
        }
        if(accuIns->getOperand(1)==I->reg){
            dynamic_cast<BinaryInstruction*>(accuIns.get())->setOperands(1, accValue->reg);
        }
    }
    vector<InstructionPtr> selectInsts;
    if(retValue){
        if(I->reg->useHead || accValue){
            dynamic_cast<PhiInstruction*>(retValue.get())->addFrom(retValue->reg,BB);
            dynamic_cast<PhiInstruction*>(retValid.get())->addFrom(retValid->reg,BB);
            
        }
        else{
            auto select = InstructionPtr(new SelectInstruction(BB,retValid->reg, retValue->reg, retIns->getOperand(0)));
            // select->print();
            selectInsts.push_back(select);
            BB->insertInstruction(select, BB->instructions.back());
            dynamic_cast<PhiInstruction*>(retValue.get())->addFrom(select->reg,BB);
            dynamic_cast<PhiInstruction*>(retValid.get())->addFrom(Const::getConst(Type::getBool(),true),BB);   
        }
        if(accValue){
            dynamic_cast<PhiInstruction*>(accValue.get())->addFrom(accuIns->reg,BB);
        }
    }
    error("eliminateCall5\n");
    auto branch = InstructionPtr(new BrInstruction(loopHeader->label,BB));
    deleteUser(BB->instructions.back());
    BB->instructions.pop_back();
    BB->removeInsturction(I);
    deleteUser(I);
    BB->instructions.push_back(branch);

    if(retValue){
        if(selectInsts.empty()){
            loopHeader->removeInsturction(retValue);
            loopHeader->removeInsturction(retValid);
            if(accValue){
                for(auto bb:func->block->basicBlocks){
                    auto terminator = bb->instructions.back();
                    if(terminator->type == Return){
                        auto ret = terminator->getOperand(0);
                        auto acc = InstructionPtr(new BinaryInstruction(accuIns->getOperand(0),accuIns->getOperand(1), dynamic_cast<BinaryInstruction*>(accuIns.get())->op, bb));
                        dynamic_cast<BinaryInstruction*>(acc.get())->setOperands(acc->getOperand(0)== accValue->reg,ret);
                        bb->insertInstruction(acc,terminator);
                        dynamic_cast<ReturnInstruction*>(terminator.get())->retValue = acc->reg;
                    }
                }
            }
        }
    }
    else{
        for(auto bb:func->block->basicBlocks){
            auto terminator = bb->instructions.back();
            if(terminator->type == Return){
                auto ret = terminator->getOperand(0);
                auto select = InstructionPtr(new SelectInstruction(bb, retValid->reg, retValue->reg, ret));
                selectInsts.push_back(select);
                // select->print();
                bb->insertInstruction(select,terminator);
                dynamic_cast<ReturnInstruction*>(terminator.get())->retValue = select->reg;
            }
        }
        if(accValue){
            for(auto select : selectInsts){
                auto value = select->getOperand(2);
                auto acc = InstructionPtr(new BinaryInstruction(accuIns->getOperand(0),accuIns->getOperand(1), dynamic_cast<BinaryInstruction*>(accuIns.get())->op, select->basicblock));
                dynamic_cast<BinaryInstruction*>(acc.get())->setOperands(acc->getOperand(0)== accValue->reg,value);
                select->basicblock->insertInstruction(acc,select);
                dynamic_cast<SelectInstruction*>(select.get())->val2 = acc->reg;
            
            }
        }
    }


}

void tailCallEliminateBB(BasicBlockPtr BB, FunctionPtr func){
    auto term = BB->instructions.back();
    error("tailCallEliminateBB1\n");
    if(term->type == Return){
        if(auto callI = getLastRecurCall(BB, func)){
            eliminateCall(callI, func, BB);
            return ;
        }
    }

    error("tailCallEliminateBB2\n");
    if(term->type == Br){
        // if(dynamic_cast<BrInstruction*>(term.get())->exp) return;
        auto exitBlock = func->block->LabelBBMap[dynamic_cast<BrInstruction*>(term.get())->label1];
        for(auto ins:exitBlock->instructions){
            if(ins->type == Phi){
                continue;
            }
            if(ins->type != Return){
                break;
            }
            if(auto callI = getLastRecurCall(BB, func)){
                deleteUser(BB->instructions.back());
                BB->instructions.pop_back();
                InstructionPtr newRet = InstructionPtr(new ReturnInstruction(dynamic_cast<ReturnInstruction*>(ins.get())->retValue,BB));
                BB->instructions.push_back(newRet);
                auto retVal = dynamic_cast<ReturnInstruction*>(newRet.get())->retValue;
                if(retVal->I->type==Phi){
                    auto PI = dynamic_cast<PhiInstruction*>(retVal->I);
                    for(int i = 0;i<PI->from.size();i++){
                        if(PI->from[i].second==BB){
                            newUse(PI->from[i].first.get(), newRet.get());
                            dynamic_cast<ReturnInstruction*>(newRet.get())->retValue = PI->from[i].first;
                            break;
                        }
                    }
                }

                for(auto PhiI:exitBlock->instructions){
                    if(PhiI->type == Phi){
                        auto dyPhiI = dynamic_cast<PhiInstruction*>(PhiI.get());
                        for(auto fit = dyPhiI->from.begin();fit!=dyPhiI->from.end();){
                            if((*fit).second == BB){
                                {
                                    Use *use = (*fit).first->useHead;
                                    while(use) {
                                        if(use->user == PhiI.get()) {
                                            use->rmUse();
                                            break;
                                        }
                                        use = use->next;
                                    }
                                }
                                fit = dyPhiI->from.erase(fit);
                            }
                            else{
                                fit++;
                            }
                        }
                    }
                    else{
                        break;
                    }
                }
                error("tailCallEliminateBB3\n");
                eliminateCall(callI, func, BB);
                error("tailCallEliminateBB4\n");
            }
        }
    }
}

void tailCallEliminate(FunctionPtr func){
    if(func->block->basicBlocks[0]->instructions[0]->type == Alloca){
        return;
    }

    for(int i = 0 ;i<func->block->basicBlocks.size();i++){
        error("enter tailCallEliminate\n");
        error(func->name<<endl);
        tailCallEliminateBB(func->block->basicBlocks[i],func);
        error("end tailCallEliminate");
    }
}